/* $Id: free.c 2741 2022-09-17 21:58:12Z teshields $ */

/* This file is part of the Libnew package that provides replacement
   implementations for core standard C library memory management
   functions, and for the standard C++ 'new()' and 'delete()' operators.

   Copyright (C) 2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#if defined (TRACE)
#include <stdio.h>
#endif

#include <stdlib.h>

#if defined (USE_BDWGC)
#include <gc/gc.h>
#warning "The BDWGC package must be configured with --enable-redirect-malloc \
in order to support replacement of the 'free()' function.  When so \
configured, it provides an implementation of 'free()'."
#elif defined (USE_DMALLOC)
#define DMALLOC_DISABLE
#include <dmalloc.h>
#warning "The DMALLOC package provides an implementation of 'free()'."
#else
#error "Neither USE_BDWGC nor USE_DMALLOC is defined"
#endif

void free (void* ptr)
{
#if defined (TRACE)
  fprintf (stdout, "free (%#lx)\n", ptr);
#endif
#if defined (USE_BDWGC)
  GC_FREE (ptr);
#elif defined (USE_DMALLOC)
  dmalloc_free (NULL, 0,
		ptr,
		DMALLOC_FUNC_FREE
	       );
#endif
}
