// $Id: op_new_v_nt.cc 2742 2022-09-17 21:59:20Z teshields $

// This file is part of the Libnew package that provides replacement
// implementations for core standard C library memory management
// functions, and for the standard C++ 'new()' and 'delete()' operators.
//
// Copyright (C) 2020-2022  Thomas E. Shields
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#if defined (TRACE)
#include <iostream>
#endif

#include <new>

void* operator new[] (std::size_t sz, const std::nothrow_t&)
#if __cplusplus >= 201103L // >= C++11
                                                             noexcept
#else
                                                             throw()
#endif
{
#if defined (TRACE)
  std::cerr << "operator new[] (size_t sz, nothrow_t&) noexcept"
               ": calling ... " << std::endl << "   ";
#endif

  try
  {
    return ::operator new[] (sz);
  }
  catch (...)
  {
    return
#if __cplusplus >= 201103L // >= C++11
           nullptr
#else
           NULL
#endif
                  ;
  }
}
