/* expr.Y:  Y-file for a simple expression language */
%left '+'
%left '*'
%token CONST

@attributes {long val;} s e CONST

@traversal PV

%{
#if defined (__cplusplus)
#include <cstdio>
#else
#include <stdio.h>
#endif

#if YYBTYACC && !YYBYACC
/* the `btyacc' 3.0 skeleton omits the following */
#include "parser.hh"
#endif

long sVal;

extern int yylex(void);

#if YYBTYACC && !YYBYACC
void yyerror(const char *, ...);
#else
void yyerror(const char *);
#endif

#if defined(MODNODE_TREEOUT)
char node_label[50];
#endif
%}

%%
s       :       e
        @{ @i sVal = @s.val@ = @e.val@;
           @PV {
#if defined(MODNODE_TREEOUT)
                snprintf(node_label,sizeof(node_label),
                         "s|result=%ld",@s.val@);
                agset(@e.NIO@,(char *)"label",node_label);
#endif
               }
        @}
        ;
e       :       e       '+'     e
        @{ @i @e.0.val@ = @e.1.val@ + @e.2.val@;
           @PV {
#if defined(MODNODE_TREEOUT)
                snprintf(node_label,sizeof(node_label),
                         "e|%ld+%ld=%ld",@e.1.val@,@e.2.val@,@e.0.val@);
                agset(@e.0.NIO@,(char *)"label",node_label);
#endif
               }
        @}
        ;
e       :       e       '*'     e
        @{ @i @e.0.val@ = @e.1.val@ * @e.2.val@;
           @PV {
#if defined(MODNODE_TREEOUT)
                snprintf(node_label,sizeof(node_label),
                         "e|%ld*%ld=%ld",@e.1.val@,@e.2.val@,@e.0.val@);
                agset(@e.0.NIO@,(char *)"label",node_label);
#endif
               }
        @}
        ;
e       :       '('     e       ')'
        @{ @i @e.val@ = @e.1.val@;
           @PV {
#if defined(MODNODE_TREEOUT)
                snprintf(node_label,sizeof(node_label),"e|(%ld)",@e.1.val@);
                agset(@e.0.NIO@,(char *)"label",node_label);
#endif
               }
        @}
        ;
e       :       CONST
        @{ @i @e.val@ = @CONST.val@;
           @PV {
#if defined(MODNODE_TREEOUT)
                snprintf(node_label,sizeof(node_label),"e|%ld",@e.val@);
                agset(@e.NIO@,(char *)"label",node_label);
#endif
               }
        @}
        ;
%%
#if YYBTYACC && !YYBYACC
void yyerror(const char *s, ...)
{extern int yylineno;
 fprintf(stderr,"Near line %d: %s\n",yylineno,s);
 yyyabort();
}
#else
void yyerror(const char *s)
{fprintf(stderr,"%s\n",s);
 yyyabort();
}
#endif

#ifndef YYMSTA
int yyparse(void);
#endif

#include <graphviz/cgraph.h>
int main (int argc, char *argv[])
{
#if   defined(DEFAULT_TREEOUT)
 FILE *treeout=fopen("default-treeout.gv","w");
#elif defined(MODLEAF_TREEOUT)
 FILE *treeout=fopen("modleaf-treeout.gv","w");
#elif defined(MODNODE_TREEOUT)
 FILE *treeout=fopen("modnode-treeout.gv","w");
#endif

 extern FILE *yyin;
 if (argc > 1)
   yyin = fopen (argv[argc-1], "r");
 else
   yyin = stdin;

 yyparse();
#if defined(MODLEAF_TREEOUT)
 agattr(yyyTreeVizGraph, AGRAPH, (char *)"label",
        (char *)"AG parse tree w/ modified terminal nodes");
#elif defined(MODNODE_TREEOUT)
 agattr(yyyTreeVizGraph, AGRAPH, (char *)"label",
        (char *)"AG parse tree w/ modified nonterminal & terminal nodes");
#endif
#if   defined(DEFAULT_TREEOUT) || defined(MODLEAF_TREEOUT) || defined(MODNODE_TREEOUT)
 agwrite(yyyTreeVizGraph,treeout);
 agclose(yyyTreeVizGraph); fclose(treeout);
#endif
 printf("%ld\n",sVal);
}
