/* $Id: gppl_glob.h 2379 2022-01-07 22:41:51Z teshields $ */ 
  
/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1993  Kurt M. Bischoff
   Copyright (C) 2016-2022  Thomas E. Shields
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef GPPL_GLOB_H_INCLUDED
#define GPPL_GLOB_H_INCLUDED

/* GCC (and some other compilers) define '__attribute__', but just in case
 '__attribute__' isn't defined, make a dummy. */
#if !(defined(__GNUC__) || defined(__attribute__))
#define __attribute__(p) /* nothing */
#endif

typedef unsigned short lineNumType;
typedef unsigned short labelType;
typedef unsigned short levelType;
typedef short regNumType;

enum stgClass {VARIABLEC,VALPARAMC,REFPARAMC};
enum symType  {INTT,FLOATT,BOOLT,VOIDT,ERRT};  
enum mapType  {ARITHM,LOGM}; 
enum pred     {False,True};

extern enum pred doCodeGen;

extern long stgSize(enum symType,long); 
extern void handleSemError(void); 
extern void yyyabort(void); 

union num {int i; 
           float f; 
           int b; 
          };

struct dim {long size;
            long scale;
            struct dim *next;
           };

struct sym {
            /* members used for all kinds of symbols */
            char *id; 	
            lineNumType line; 
            enum symType type; 
            levelType level;
            struct sym *next; 

            /* members used only for storage symbols */
            enum stgClass Class; 
            long addr;
            long size;
            struct dim *dimList;

            /* members used only for function symbols */
            long paramsSize;
            labelType label; 
            struct sym *paramLoc;
           }; 

struct env {struct sym *funcSym;
            struct sym *varLoc;
            struct sym *funcLoc; 
            levelType level; 
            short nRegs;
            labelType exitLabel;
            long varsSize;
            struct env *outerEnv; 
           }; 

extern struct dim *dimListSpace; 
#define lambdaDim dimListSpace 

extern struct sym *symSpace; 
#define lambdaSym symSpace 

extern struct env *envSpace; 
#define lambdaEnv envSpace 

#endif /* !GPPL_GLOB_H_INCLUDED */
