/* $Id: util.c 2379 2022-01-07 22:41:51Z teshields $ */ 
  
/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1993  Kurt M. Bischoff
   Copyright (C) 2016-2022  Thomas E. Shields
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#if defined (__cplusplus)
#include <cstdio>
#include <cstdlib>
#else
#include <stdio.h>
#include <stdlib.h>
#endif

#include "gppl_glob.h"
#include "util.h"

extern int yylineno; 

#if YYBTYACC && !YYBYACC
void yyerror(const char *cp, ...)
#else  
void yyerror(const char *cp) 
#endif 
  {fprintf(stderr,"Syntax error near line %d: %s\n",yylineno,cp); 
   yyyabort(); 
  }  



void fatal(const char *) __attribute__((noreturn)); 
void internal(const char *) __attribute__((noreturn)); 
void handleOverflow(const char *) __attribute__((noreturn)); 

void fatal(const char *cp) 
  {fprintf(stderr,"Fatal error: %s\n",cp); 
   exit(EXIT_FAILURE); 
  }



void internal(const char *s) 
  {fprintf(stderr,"gppl/gc internal error: \n\t"); 
   fatal(s);
  }



void handleOverflow(const char *cp) 
  {internal(cp); 
  } 



FILE *safeOpen(const char *name, const char *mode) 
  {FILE *dum; 
   if ((dum = fopen(name,mode)) 
       == 
       ((FILE *)NULL) 
      ) 
      {fprintf(stderr,"fatal error opening %s.\n",name); 
       exit(EXIT_FAILURE); 
      } 
      else 
      return dum; 
  } 



void printStringArray(const char **arr, int sizeofArr, FILE *fp) 
  {for (int i=0;i<(sizeofArr/((int)sizeof(char *)));i++)  
       (void)fputs(arr[i],fp); 
  } 
