/* $Id: bison-directives.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2021-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef BISON_DIRECTIVES_HH_INCLUDED
#define BISON_DIRECTIVES_HH_INCLUDED

#include "location.hh"

#include "semtypes.hh"

// handle Bison '%language "LANGUAGE"' directive
extern void language_directive (const AGC::location&, const std::string&);

// handle Bison '%skeleton "FILE"' directive
extern void skeleton_directive (const AGC::location&, const std::string&);

// handle Bison '%define <API_variable>' directive
extern void API_variable_default (const AGC::location&, const std::string&);

// handle Bison '%define <API_variable> ID' directive
extern void
API_variable_ID (const AGC::location&, const std::string&, const std::string&);

// handle Bison '%define <API_variable> LITERAL' directive
extern void API_variable_LITERAL (const AGC::location&,
                                  const std::string&,
                                  const std::string&);

// handle Bison directives:
// 	'%define <API_variable> { ID }'
// 	'%define <API_variable> { NAMESPACE_ID }'
// 	'%define <API_variable> { <type> }'
extern int
API_variable_EXPR (const AGC::location&, const std::string&, const typeSeq);

#endif // !BISON_DIRECTIVES_HH_INCLUDED
