/* $Id: codegen.hh 2841 2022-12-13 01:08:03Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2020-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef CODEGEN_HH_INCLUDED
#define CODEGEN_HH_INCLUDED

#include "semtypes.hh"

#include "opts.hh"
#include "outMgr.hh"

// start non-terminal for Ox's augmented grammar
#define augNstr "AugNonterm"

extern symDeclSeq global_symDeclSeq;
extern userArgSeq global_userArgSeq;

// if "%destructor { ... } <*>" or "%printer { ... } <*>" occurs ...
extern bool needDestructorForStartSymbol;
extern bool needPrinterForStartSymbol;


extern void finishYfileDefinitionSection (sortCollection&);
extern void finishYfileRulesSection (void);


enum instRefType
{
  imm,
  decor,
  trav
};

class codeGen : public outMgr
{
public:

  codeGen (std::iostream* obj,
           AGC::position* pos = nullptr,
           bool forLexer = false)
      : outMgr (obj, pos, forLexer)
  {}

  codeGen (std::iostream* obj,
           std::string& name,
           int line,
           AGC::position* pos = nullptr,
           bool forLexer = false)
      : outMgr (obj, name, line, pos, forLexer)
  {}

  virtual codeGen& put (char) = 0;

  virtual inline codeGen&
  put (const std::string& str)
  {
    return static_cast<codeGen&> (outMgr::put (str));
  }

  virtual inline codeGen&
  put (int i, int width = 0)
  {
    return static_cast<codeGen&> (outMgr::put (i, width));
  }

  virtual inline codeGen&
  put (long l, int width = 0)
  {
    return static_cast<codeGen&> (outMgr::put (l, width));
  }

  virtual inline codeGen&
  lex (void)
  {
    this->outMgr::lex ();
    return *this;
  }

  virtual inline codeGen&
  gen (void)
  {
    this->outMgr::gen ();
    return *this;
  }

  // implementation found in ctc.cc

  void putStringArray (const char* const* arr, size_t);

  void skeletonText1 (void);  // Ox lexer & parser skeleton
  void skeletonText2 (void);  // Ox lexer & parser skeleton
  void skeletonText3 (void);  // Ox lexer & parser skeleton
  void skeletonText4 (void);  // Ox lexer skeleton
  void skeletonText5 (void);  // Ox parser skeleton
  void skeletonText6 (void);  // Ox parser skeleton
  void skeletonText7 (void);  // Ox parser skeleton
  void skeletonText8 (void);  // Ox parser skeleton
  void skeletonText9 (void);  // Ox parser skeleton
  void skeletonText10 (void); // Ox parser skeleton

  void putRootAttrbRefSection (void);

  void putSolveAndSignalHeader (void);
  void putSolveAndSignalTrailer (void);


  // implementation found in codegen.cc

  void putBisonCXXdefines (void);
  void putOxAttrbsDecl (void);
  void putOxAttrbsUnionDecl (sortCollection&);
  void putYaccAttrbsDecl (void);
  void putYaccUnionDecl (void);

  void putImmediateAttrbRef (attrbPair, bool);

  void
  translateAndOutput (enum instRefType, const attrbRefSeg&, const attrbPair);

  codeGen& putAttrbTypeDecl (const int, const std::string&, const bool);
  codeGen& putGenLeafCall (const std::string&);
  codeGen& putGenIntNodeCall (void);
  codeGen& putDecorateCall (const std::string& firstParam);
  codeGen& putDecorateFwdDecl (void);
  codeGen& putDecoratePrototype (void);
  codeGen& putLHSsemStkElRef (void);
  codeGen& putRHSsemStkElRefList (const std::string& = std::string ());
  codeGen& putReclaimSemStkEls (void);

  codeGen& putPatchlevelComment (void);

  void putParserHeader (void);         // Ox parser header skeleton
  codeGen& putParserHeaderText (void); // Ox parser header skeleton

  void putLexerHeader (enum whichYoxLexer); // Ox lexer header skeleton


  // implementation found in trav.cc

  void putTravCodePerProd (void);
  void putTravCode (void); // Ox traversal skeleton
  void putTravDefines (void);
  codeGen& putDoTraversals (const std::string&);


  // implementation found in treeviz.cc

  void putGenTreeVizLeafCall (const std::string&);
  void putGenTreeVizLeafExtDecl (void);
  void putGenTreeVizNodeCall (const std::string&);
  void putGenTreeVizNodeFwdDecl (void);
  void putTreeVizCode (void); // Ox tree visualization skeleton
  void putTreeVizExtDecl (void);
  void putTreeVizGraphvizIncl (void);
  void putTreeVizId (void);
  void putTreeVizUserDefPrefix (void);

protected:

  // implementation found in codegen.cc

  void putDecorateAttrbRef (attrbPair, attrbPair);
  void putTraversalAttrbRef (attrbPair);

private:

  // implementation found in codegen.cc

  void putParserImmediateAttrbHead (attrbPair ref);
  void putLexerImmediateAttrbHead (void);
};

#endif // !CODEGEN_HH_INCLUDED
