/* $Id: depGs.hh 2841 2022-12-13 01:08:03Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2016-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef DEPGS_HH_INCLUDED
#define DEPGS_HH_INCLUDED

#include "location.hh"

#include "semtypes.hh"

extern long eeCountInitSeq;
extern long attrbOccrSeq;
extern long leafRefCountInitSeq;
extern long entOccrSeq;
extern long gsOccurSeq;

extern bool permitCtimeCircTest;

extern bool hasYaccPvarRefGlob;

extern bool decorationRequested;

// => true if there are any `yyyDecorate()' evaluations
extern bool genRSNodeProdNumSwitchHd;

extern bool memberDepend (attrbPair, bool, bool);

extern void initLDG (void);
extern void processAdepSpec ();

#if defined(G_DEBUG)
// debugging routines
extern void showDepGraph (void);
extern void showGsymInfo (int);
extern void showSymInfoForProd (void);
extern void othershowGsymInfo (void);
#endif


struct refCount
{
  long start;
  long length;

  refCount ()
      : start (-1),
        length (0)
  {}

  refCount (long i, long l)
      : start (i),
        length (l)
  {}
};

extern const refCount installLeafRefCountInitList (const std::string&);


extern void finishYfileAttrbRefSection (const AGC::location&);
extern void finishLfileAttrbRefSection (const AGC::location&);

#endif // !DEPGS_HH_INCLUDED
