/* $Id: fileMgr.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2010-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef FILEMGR_HH_INCLUDED
#define FILEMGR_HH_INCLUDED

#include <fstream>
#include <string>

#include "codegen.hh"



extern bool erase (const std::string&);



class lookaheadBuf;



class fileMgr : public codeGen
{
public:

  fileMgr (bool forLexer = false)
      : codeGen (new std::fstream (),
                 new AGC::position (new std::string ()),
                 forLexer)
  {
    wasOutName = true;
  }

  fileMgr (AGC::position* pos, bool forLexer = false)
      : codeGen (new std::fstream (), pos, forLexer)
  {
    wasOutName = true;
  }

  fileMgr (std::string, bool forLexer = false);

  void open (std::string fileName);

  void close (void);

  void erase (void);

  fileMgr& put (char);

  inline fileMgr&
  put (const std::string& str)
  {
    return static_cast<fileMgr&> (codeGen::put (str));
  }

  inline fileMgr&
  put (int i, int width = 0)
  {
    return static_cast<fileMgr&> (codeGen::put (i, width));
  }

  inline fileMgr&
  put (long l, int width = 0)
  {
    return static_cast<fileMgr&> (codeGen::put (l, width));
  }

  fileMgr& append (fileMgr&);

  fileMgr& append (lookaheadBuf&);

protected:

  inline std::fstream&
  getObj (void) // get the external file
  {
    return static_cast<std::fstream&> (*theObj);
  }

  inline long
  tellp (void)
  {
    return getObj ().tellp ();
  }

private:
};

#endif // !FILEMGR_HH_INCLUDED
