/* $Id: YY_INPUT.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef YY_INPUT_HH_INCLUDED
#define YY_INPUT_HH_INCLUDED

// overload virtual Lexer::LexerInput to treat '\r' or "\r\n" as '\n'
int
Lexer::LexerInput (char* buf, int max_size)
{
  if (yyin.eof () || yyin.fail ())
    return 0;

  (void)yyin.get (buf, max_size);
  int numread = yyin.gcount ();

  if (yyin.eof () && numread == 0)
    return 0;
  if (yyin.fail ())
    { // no characters were extracted ...
      yyin.clear (yyin.rdstate () & ~std::ios::failbit); // clear the fail bit
      if (yyin.peek () == '\n')
        { // ... because the next character is '\n'
          yyin.get (); // skip the following '\n'
	}
      strcpy (buf, "\n");
      return 1;
    }
  if (yyin.bad ())
    return -1;

  if (buf[numread - 1] == '\r')
    {
      buf[numread - 1] = '\n'; // replace the '\r'
      if (yyin.peek () == '\n')
        yyin.get (); // skip the following '\n'
    }
  else if (yyin.peek () == '\n')
    {
      buf[numread] = '\n'; // add '\n' to the buffer
      buf[++numread] = '\0';
      yyin.get (); // skip the following '\n'
    }

  return numread;
}

#endif // !YY_INPUT_HH_INCLUDED
