/* $Id: joinTabs.cc 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2010-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include "ctc.hh"
#include "depGs.hh"
#include "fileMgr.hh"
#include "joinTabs.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "scan.hh"
#include "symTrans.hh"
#include "treeviz.hh"

#define temp1FileID "@temp1.~tab~.c"
#define temp2FileID "@temp2.~tab~.c"
#define temp3FileID "@temp3.~tab~.c"
#define temp4FileID "@temp4.~tab~.c"
#define temp5FileID "@temp5.~tab~.c"
//#define temp6FileID "@temp6.~tab~.c" -- unused
#define temp7FileID "@temp7.~tab~.c"
//#define temp8FileID "@temp8.~tab~.c" -- unused
#define tempSortIndFileID "@temp9.~tab~.c"
#define tempSortNumsFileID "@temp10.~tab~.c"
#define tempProdsIndFileID "@temp11.~tab~.c"
#define tempProdsNumsFileID "@temp12.~tab~.c"
#define tempStringTabFileID "@temp13.~tab~.c"

#define tempTravFileID "@temp.~trav~"


fileMgr temp1Out;
fileMgr temp2Out;
fileMgr temp3Out;
fileMgr temp4Out;
fileMgr temp5Out;
//fileMgr temp6Out; -- unused
fileMgr temp7Out;
//fileMgr temp8Out; -- unused

fileMgr tempSortIndFile;
fileMgr tempSortNumsFile;

fileMgr tempProdsIndFile;
fileMgr tempProdsNumsFile;

fileMgr tempTravOut;

fileMgr tempStringTabFile;



void
openTempFiles (void)
{
  temp1Out.open (temp1FileID);
  temp2Out.open (temp2FileID);
  temp3Out.open (temp3FileID);
  temp4Out.open (temp4FileID);
  temp5Out.open (temp5FileID);
//temp6Out.open (temp6FileID); -- unused
  temp7Out.open (temp7FileID);
//temp8Out.open (temp8FileID); -- unused

  tempSortIndFile.open (tempSortIndFileID);
  tempSortNumsFile.open (tempSortNumsFileID);

  tempProdsIndFile.open (tempProdsIndFileID);
  tempProdsNumsFile.open (tempProdsNumsFileID);

  tempTravOut.open (tempTravFileID);

  tempStringTabFile.open (tempStringTabFileID);
}



void
rmTempFiles (void)
{
  static int called = 0;

  if (called++ != 0)
    return;

  temp1Out.close ();
  temp2Out.close ();
  temp3Out.close ();
  temp4Out.close ();
  temp5Out.close ();
//temp6Out.close (); -- unused
  temp7Out.close ();
//temp8Out.close (); -- unused

  tempSortIndFile.close ();
  tempSortNumsFile.close ();

  tempProdsIndFile.close ();
  tempProdsNumsFile.close ();

  tempTravOut.close ();

  tempStringTabFile.close ();

  temp1Out.erase ();
  temp2Out.erase ();
  temp3Out.erase ();
  temp4Out.erase ();
  temp5Out.erase ();
//temp6Out.erase (); -- unused
  temp7Out.erase ();
//temp8Out.erase (); -- unused

  tempSortIndFile.erase ();
  tempSortNumsFile.erase ();

  tempProdsIndFile.erase ();
  tempProdsNumsFile.erase ();

  tempTravOut.erase ();

  tempStringTabFile.erase ();
}



static void
buildCycleFinder (void)
{
  yoxOutG.gen ()
      .put ("\n\n")
      .put ("#define " + yoxPrefix ("LastProdNum") + " ").put (prodNumSeq - 1)
          .put ("\n")
      .put ("\n\n")
      .put ("static int " + yoxPrefix ("ProdsInd") + "[] = {\n")
      .put ("   0,")
      .append (tempProdsIndFile)
      .gen ()
      .put ("\n").put (gsOccurSeq, 4)
      .put (",\n};\n\n\n")
      .put ("static int " + yoxPrefix ("Prods") + "[][2] = {")
      .append (tempProdsNumsFile)
      .gen ()
      .put ("\n};\n\n\n")
      .put ("static int " + yoxPrefix ("SortsInd") + "[] = {\n")
      .put ("  0,")
      .append (tempSortIndFile)
      .gen ()
      .put ("\n").put (attrbDeclAttrbOccurSeq, 3)
      .put (",\n};\n\n\n")
      .put ("static int " + yoxPrefix ("Sorts") + "[] = {");

  if (tempSortNumsFile.empty ())
    yoxOutG.put ("\n  0,");
  else
    yoxOutG.append (tempSortNumsFile).gen ();

  yoxOutG.put ("\n};\n");

  printStringTab ();

  yoxOutG.gen ()
      .put ("\n\n\n")
      .put ("static const char *const " + yoxPrefix ("StringTab") + "[] = {")
      .append (tempStringTabFile)
      .gen ()
      .put ("\n};\n");

  yoxOutG.skeletonText10 ();
}



void
organizeTables (void)
{
  const yoxControl& ctrl = outMgr::ctrl;

  yoxOutG.skeletonText7 ();

  yoxOutG.gen ()
     .put ("\n\nstatic " + yoxPrefix ("FT") + " " + yoxPrefix ("RCIL") + "[");

  if (eeCountInitSeq != 0)
    {
      yoxOutG.gen ()
          .put ("] = {\n")
          .append (temp2Out)
          .gen ()
          .put ("\n};");
    }
  else
    yoxOutG.gen ().put ("1];");

  yoxOutG.gen ()
      .put ("\n\n"
            "static short " + yoxPrefix ("IIIEL") + "[] = {0,\n");

  for (int i = 1; i < prodNumSeq; i++)
    {
      yoxOutG.put (indToEntsByProdNum[i]).put (",");
      if (i % 10 == 0)
        yoxOutG.put ("\n");
    }

  yoxOutG
      .put ("\n"
            "};"
            "\n\n"
            "static long " + yoxPrefix ("IIEL") + "[] = {\n")
      .append (temp5Out)
      .gen ()
      .put ("\n"
            "};"
            "\n\n")
      .put ("static long " + yoxPrefix ("IEL") + "[");

  if (attrbOccrSeq != 0)
    {
      yoxOutG.gen ()
          .put ("] = {\n")
          .append (temp4Out)
          .gen ()
          .put (entOccrSeq * 2)
          .put (",\n};\n");
    }
  else
    yoxOutG.gen ().put ("1];\n");

  yoxOutG.gen ()
      .put ("\n"
            "static " + yoxPrefix ("FT") + " " + yoxPrefix ("EntL") + "[");

  if (entOccrSeq != 0)
    {
      yoxOutG.gen ()
          .put ("] = {\n")
          .append (temp3Out)
          .gen ()
          .put ("\n};\n");
    }
  else
    yoxOutG.gen ().put ("1];\n");

  yoxOutG.skeletonText8 ();

  if (ctrl.genTreeViz)
    yoxOutG.putTreeVizCode ();

  yoxOutG.append (temp1Out);

  if (permitCtimeCircTest)
    {
      yoxOutG.gen ()
          .put ("\n\n\n")
          .put ("static " + yoxPrefix ("WAT") + " " + yoxPrefix ("LRCIL"))
          .put ("[")
          .put ((leafRefCountInitSeq != 0) ? leafRefCountInitSeq : 1)
          .put ("]");

      if (leafRefCountInitSeq != 0)
        yoxOutG.gen ()
            .put (" = {")
            .append (temp7Out)
            .gen ()
            .put ("\n}");

      yoxOutG.gen ().put (";\n");
    }

  yoxOutG.skeletonText9 ();

  buildCycleFinder ();
}
