/* $Id: lextypes.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2019-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef LEXTYPES_HH_INCLUDED
#define LEXTYPES_HH_INCLUDED

#include <string>

// lexical representation of an attribute occurrence

struct attrbTriple
{
  std::string gsName; // grammar symbol
  int occurrence;     // attribute occurrence in rule
  std::string atName; // attribute name
};


// other lexical semantic types

struct numberRep
{
  std::string spelling;
  long long unsigned value;

  inline numberRep () = default;

  inline numberRep (std::string s, long long unsigned v)
      : spelling (s),
        value (v)
  {}

  inline numberRep (const numberRep& n)
  {
    spelling = n.spelling;
    value = n.value;
  }
};


// distinct 'string' types, so that %printer can be specialized

class char_literal : public std::string
{
public:

  inline char_literal ()
      : std::string ()
  {}

  inline char_literal (const char* c)
      : std::string (c)
  {}

  inline char_literal (const std::string& c)
      : std::string (c)
  {}
};

class name_literal : public std::string
{
public:

  inline name_literal ()
      : std::string ()
  {}

  inline name_literal (const char* n)
      : std::string (n)
  {}

  inline name_literal (const std::string& n)
      : std::string (n)
  {}

  // hack: convert a literal to a name
  inline name_literal (const char_literal& n)
      : std::string (n)
  {}
};

class keyword : public std::string
{
public:

  inline keyword ()
      : std::string ()
  {}

  inline keyword (const char* k)
      : std::string (k)
  {}

  inline keyword (const std::string& k)
      : std::string (k)
  {}

  // hack: convert a name to a keyword
  inline keyword (const name_literal& k)
      : std::string (k)
  {}
};

#endif // !LEXTYPES_HH_INCLUDED
