/* $Id: memMgr.cc 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2010-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <cassert>

#include <string>

#include "memMgr.hh"
#include "outMgr.hh"

#include "scan.hh"



// =============================== class memMgr ================================

// public:

void
memMgr::clear (void)
{
  // throw away any contents
  outMgr::clear ();
  wasOutName = false;

  origin.line = 0;

  // and, reset the internal memory file
  getObj ().str (std::string ());
  getObj ().clear ();
}



memMgr&
memMgr::put (char c)
{
  if (origin.line == 0)
    origin = *curPos;

  putCharInWriteBuffer (c);

  return *this;
}



memMgr&
memMgr::append (lookaheadBuf& whichBuf)
{
  assert (whichBuf != *this);

  if (whichBuf.empty ())
    return *this;

  put (whichBuf.getObj ().str ());

  if (origin.line == 0)
    origin = whichBuf.origin;

  // flush any pending whitespace
  if (!whichBuf.whitespace.empty ())
    put (whichBuf.whitespace);

  wasOutName = false;

  return *this;
}

// protected:

// private:

// ================================ end memMgr =================================

// ============================= class refSegBuf ===============================

// public:

// called at the start of the evaluation expression in each attribute
// definition, at the start of an @auto(inh|syn) attribute definition,
// at the start of each dynamic traversal modifier, and at the start of
// of each traversal action specification

void
refSegBuf::init (const AGC::position startPos)
{
  clear (); // throws away any comments & white space preceding the annunciator
  origin = startPos;
  curPos = &origin;
}

void
refSegBuf::putAttrbPair (attrbPair ref)
{
  put (attrbNumPairDelimChar);
  put (ref.pos);
  put (attrbNumPairDelimChar);
  put (ref.num);
  put (attrbNumPairDelimChar);;
}

// protected:

void
refSegBuf::putOutLineDirective (void)
{
  assert (!heOptions->noLineDirectives);

  wasOutName = true;
  lastInlineno = -1; // fake it, for now
}

// private:

// ============================== end refSegBuf ================================
