/* $Id: messages.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2019-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef MESSAGES_HH_INCLUDED
#define MESSAGES_HH_INCLUDED

#include "location.hh"

extern void yox_error (const AGC::location&, const std::string&);
extern void yox_error (const AGC::location&, std::ostringstream&);

extern void yox_warning (const AGC::location&, const std::string&);
extern void yox_warning (const AGC::location&, std::ostringstream&);

extern void yox_fatal (const AGC::location&, const std::string&)
    __attribute__ ((__noreturn__));
extern void yox_fatal (const AGC::location&, std::ostringstream&)
    __attribute__ ((__noreturn__));

extern void yox_internal_fail (const std::string&, int, const std::string&)
    __attribute__ ((__noreturn__));
extern void yox_internal_fail (const std::string&, int, std::ostringstream&)
    __attribute__ ((__noreturn__));

extern void yox_alloc_fail (const std::string&, int)
    __attribute__ ((__noreturn__));

extern void agc_syntax_error (const AGC::location&, const std::string&);
extern void agc_syntax_error (const AGC::location&, std::ostringstream&);

// error/warning messages used in multiple places
extern void ambiguousReturn (const AGC::location&);
extern void conflictingYYSTYPEdefinitions (const AGC::location&,
                                           const std::string&,
                                           const std::string&);
extern void escapeCharMisuse (const AGC::location&);
extern void multipleYYSTYPEdefinitions (const AGC::location&,
                                        const std::string&);
extern void multipleReturnTokens (const AGC::location&);

extern void notSupported (const AGC::location&, const std::string&);

extern void returnNonterminalAsToken (const AGC::location&, const std::string&);

extern void stringContainsNewline (const AGC::location&, const AGC::position&);
extern void unbalancedParensInMacro (const AGC::location&);
extern void unexpectedATLBRACE (const AGC::location&);
extern void unexpectedChar (const AGC::location&, char, const std::string&);
extern void unexpectedEOF (const AGC::location&) __attribute__ ((__noreturn__));
extern void unknownNodeType (const AGC::location&);
extern void unmatchedRPAREN (const AGC::location&);

extern void unterminatedAction (const AGC::location&, const AGC::position&)
    __attribute__ ((__noreturn__));
extern void unterminatedAttrDeclaration (const AGC::location&,
                                         const AGC::position&)
    __attribute__ ((__noreturn__));
extern void unterminatedAttrRefSection (const AGC::location&,
                                        const AGC::position&)
    __attribute__ ((__noreturn__));
extern void unterminatedCode (const AGC::location&, const AGC::position&)
    __attribute__ ((__noreturn__));
extern void unterminatedComment (const AGC::location&, const AGC::position&)
    __attribute__ ((__noreturn__));
extern void unterminatedCPPdirective (const AGC::location&,
                                      const AGC::position&)
    __attribute__ ((__noreturn__));
extern void unterminatedString (const AGC::location&, const AGC::position&)
    __attribute__ ((__noreturn__));
extern void unterminatedYaccUnion (const AGC::location&, const AGC::position&)
    __attribute__ ((__noreturn__));

#endif // !MESSAGES_HH_INCLUDED
