/* $Id: YY_LOCATIONS.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2019-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef YY_LOCATIONS_HH_INCLUDED
#define YY_LOCATIONS_HH_INCLUDED

// customize location processing

void
yylloc_update (AGC::location& yylloc)
{
  REFLEX_OPTION_lexer::yylloc_update (yylloc);

  if (!EOLstate.atEND && *(matcher ().end () - 1) == '\n')
    {
      // the current match ends with an EOL char

      EOLstate.atEOL = true;
      yylloc.begin.column++;
      yylloc.end.line = yylloc.begin.line + 1;
      yylloc.end.column = 0;
      EOLstate.EOLpos = yylloc.begin;
      if (matcher ().columns () > 1)
        EOLstate.EOLpos.column += matcher ().columns () - 1;
    }
  else if (EOLstate.atEND)
    {
      // the current match is at @ EOF

      if (EOLstate.atEOL)
        yylloc.begin = EOLstate.EOLpos;
      else
        yylloc.end.column++;
    }
  else
    {
      EOLstate.atEOL = false;
      yylloc.begin.column++;
      yylloc.end.column += 2;
    }

  if (matcher ().at_end ())
    EOLstate.atEND = true;
}

#endif // !YY_LOCATIONS_HH_INCLUDED
