/* $Id: symTrans.hh 2836 2022-12-10 22:02:43Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2010-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef SYMTRANS_HH_INCLUDED
#define SYMTRANS_HH_INCLUDED

#include <unordered_map>
#include <vector>

#include "location.hh"

#include "lextypes.hh"
#include "opts.hh"
#include "semtypes.hh"

// There is a vector of `attrbDirection' defining whether each attribute in the
// sort associated with a grammar symbol instance in the decorated parse tree is
// inherited or synthesized, or undefind if not yet known
typedef std::vector<attrbDirection> attrbDirectionSeq;


// Each 'gSym' (representing a grammar symbol) has a copy
// of the 'sort' for that symbol.
struct gSym
{
  long symNum;
  std::string alias;
  AGC::position posOfDefn;
  sort symSort;
  attrbDirectionSeq attrbDirections;

  gSym ()
      : symNum (0),
        alias (std::string ()),
	posOfDefn (AGC::position (nullptr, -1, -1)),
        symSort (sort ())
  {}
};

// table of grammar symbols & associated attribute definitions
typedef std::unordered_map<std::string, gSym> gSymTableType;

extern gSymTableType gSymTable;


// represents a symbol in a grammar production
struct prodSymbol
{
  std::string lexeme; // name
  int symOccr;        // 0-based occurrence # of the symbol; 1st occurrence is 0
  int yaccSemIndex;   // 0-based position of the symbol; actions have a position
};

typedef std::vector<prodSymbol> production;

extern int maxProdLength;

extern int prodNumSeq;

extern int yaccSemIndex;
extern int prodBufferPointer2;

extern production prodBuffer2;


// index to dependents by production number
extern std::vector<long> indToEntsByProdNum;

// ************************************************************************

// increasing sequence of numbers representing each declared attribute;
// corresponds to the index in yyySorts[] in the generated evaluator
extern int attrbDeclAttrbOccurSeq;

extern bool atLeastOneSymHasAttrbs;

extern const std::string& stringLookup (const std::string&, bool);

extern int stringSerialNoLookup (const std::string&);
extern void showStringTab (void);
extern void printStringTab (void);

extern const std::string& gSymLookup (const std::string&);

extern std::string omittedRetSym; // gSymTable omitted Lexer return token

// install & memoize the `omitted Lexer return token'
extern void setupOmittedRetSym (void);

extern std::string errorSym;      // gSymTable `error' token
extern void setupErrorSym (void); // install & memoize the `error' token

extern std::string startSymLexeme;
extern void installStartSym (const AGC::location&, const std::string&);

extern void putAttrbSort (const sort&);
extern void addAttrbsToGsym (const AGC::location&,
                             const std::string&,
                             const sort&);

extern void installGsymIfNew (const AGC::location&, std::string&);
extern void installGsymAsTokenIfNew (const AGC::location&, const std::string&);
extern void installAliasToToken (const AGC::location&,
                                 const std::string&,
                                 const AGC::location&,
                                 const std::string&);
extern void installGsymAsNontermIfNew (const AGC::location&,
                                       const std::string&);

// create a new attribute w/ copy rule annotations (if any)
extern sortAttrb newAttrbID (const AGC::location&,
                             const std::string&,
                             enum attrbDirection,
                             bool);

extern void incProdNumSeq (void);
extern void putSymbolInProdBuffer2 (const std::string&);
extern attrbPair transAttrbTripleToAttrbPair (const AGC::location&,
                                              const attrbTriple&);
extern void attrbAutoDef (const AGC::location&,
                          const std::string&,
                          enum attrbDirection,
                          bool);

#endif // !SYMTRANS_HH_INCLUDED
