%option bison-bridge bison-locations
%{
#include <string.h>

#include "types.h"
#include "parser.h"
#include "util.h"

static int yycolno=1;

#define set_yylloc(width) {yylloc->first_line    = yylineno; \
                           yylloc->first_column  = yycolno; \
                           yylloc->last_line     = yylineno; \
                           yylloc->last_column   = ((yycolno+=width)-1);}

int i;
static char leaf_label[100];
%}

IDENT ([a-zA-Z_][a-zA-Z0-9_]*) 

%% 

\n         yycolno=1;  
[ \t\f]    yycolno++;  
const      {set_yylloc(yyleng); return(CONST);}
type       {set_yylloc(yyleng); return(TYPE);}
[0-9]+     {set_yylloc(yyleng);
            return(NUM);
            @{sscanf(yytext,"%d",&i); @NUM.iVal@ = i;
              snprintf(leaf_label,sizeof(leaf_label),
                       "NUM|%d@[%d:%d,%d:%d]",
                       i,
                       yylloc->first_line,yylloc->first_column,
                       yylloc-> last_line,yylloc-> last_column);
             agset(@NUM.NIO@, "label", leaf_label);
            @}
           }
\(         {set_yylloc(1); return('(');}
\)         {set_yylloc(1); return(')');}
\,         {set_yylloc(1); return(',');}
\;         {set_yylloc(1); return(';');}
\+         {set_yylloc(1); return('+');}
\-         {set_yylloc(1); return('-');}
\=         {set_yylloc(1); return('=');}
\.\.       {set_yylloc(2); return(DOTDOT);}
{IDENT}    {set_yylloc(yyleng);
            return(ID);
            @{@ID.loc@ = *yylloc;
              @ID.sym@ = addSymbolIfNew(g_SymbolTable(),yylloc,yytext);
              snprintf(leaf_label,sizeof(leaf_label),
                       "IDENT|%s@[%d:%d,%d,%d]",
                       yytext,
                       yylloc->first_line,yylloc->first_column,
                       yylloc-> last_line,yylloc-> last_column);
              agset(@ID.NIO@, "label", leaf_label);
            @}
           }
\{         eat_comment("}",&yycolno,&yylineno);
\(\*       eat_comment("*)",&yycolno,&yylineno);
.          {fprintf(stderr,"line %d:%d illegal character\n",yylineno,yycolno); yycolno++;}

%%
#if !defined(FLEX_SCANNER) || !defined(__cplusplus)
int yyyinput(void){return input();} /* `lex', or `flex' with C compiler */
#else
int yyyinput(void){return yyinput();} /* `flex' with C++ compiler */
#endif

int yywrap(void){return 1;}

#include <getopt.h>
#include <libgen.h>

/* options descriptor */
static struct option opts[] = {
   { "input",   required_argument, NULL, 'i' },
   { "output",  required_argument, NULL, 'o' },
   { NULL,      0,                 NULL, 0   }
};

int
main (int argc, char *argv[]) {
  char *inputFilename = (char *) NULL;
  char *DOTfilename = (char *) NULL;
  int ch;

#if YYDEBUG
  if (getenv("YYDEBUG") != 0) yydebug = 1;
#endif

  opterr = 0; /* disable getopt() generated error messages */
  while ((ch = getopt_long(argc, argv, ":i:o:", opts, NULL)) != -1) {
    switch (ch) {
      case 'i' : /* --input filename */
		 inputFilename = optarg;
		 break;
      case 'o' : /* --output filename */
		 DOTfilename = optarg;
		 break;
      case ':' : /* missing option argument */
		 switch (optopt) {
		   case 'i' : /* missing --input filename  */
		   case 'o' : /* missing --output filename */
		   default  : /* something else is wrong   */
			      ;
		 }
      case '?' : /* unknown option */
      default  : fprintf(stderr, "error in command line option: %s\n", argv[optind-1]);
		 exit(EXIT_FAILURE);
    }
  }
  argc -= optind; argv += optind;
  /* process positional arguments */
  if (argc == 0) {
    yyin = stdin;
  } else if (argc == 1) {
    inputFilename = argv[argc-1];
    yyin = fopen (inputFilename, "r");
    if (!yyin) {
       perror (inputFilename);
       exit(-1);
    }
  }

  if (yyparse() == 0) dumpSymbolTable();

  if (inputFilename != (char *) NULL) {

    if (DOTfilename == NULL) {
      char *dir = strdup(dirname(inputFilename));
      char *base = strdup(basename(inputFilename));
      char *pt;

      strcat(dir, "/");
      pt = strrchr(base, '.');
      if (pt == (char *) NULL) /* append ".gv" to inputFilename */;
      else /* replace trailing extension with ".gv" */
	*pt = '\0'; /* truncate inputFilename at the last '.' */
      DOTfilename = (char *)malloc(strlen(dir)+strlen(base)+1+2+1);
      strcat(strcat(strcpy(DOTfilename, dir), base), ".gv");
    }

    {FILE *out = fopen(DOTfilename, "w");
     agwrite(yyyTreeVizGraph, out);
     agclose(yyyTreeVizGraph); fclose(out);
    }
  }
}
