%{

#include <string>

#include "driver.hh"
#include "parser.hh"

extern void yyerror (const char *, ...);

%}

%token ASSIGN

%union {
   int   val;
}

%token IDENTIFIER
%token <val> NUMBER
%type  <val> exp

%start unit

%left '+' '-'
%left '*' '/'

@attributes { int Val; } exp NUMBER
@attributes { std::string Name; } IDENTIFIER
@attributes { @void stg; } assignment unit

%%

unit: assignments exp  @{ @m @unit.stg@; ::drv.result = @exp.Val@; @};

assignments:
  /* empty */
| assignments assignment;

assignment:
  IDENTIFIER ASSIGN exp @{ @m @assignment.stg@;
			      ::drv.variables[@IDENTIFIER.Name@] = @exp.Val@; @};

exp:
  NUMBER      @{ @i @exp.Val@ = @NUMBER.Val@; @}
| IDENTIFIER  @{ @i @exp.Val@ = ::drv.variables[@IDENTIFIER.Name@]; @}
| exp '+' exp @{ @i @exp.0.Val@ = (@exp.1.Val@ + @exp.2.Val@); @}
| exp '-' exp @{ @i @exp.0.Val@ = (@exp.1.Val@ - @exp.2.Val@); @}
| exp '*' exp @{ @i @exp.0.Val@ = (@exp.1.Val@ * @exp.2.Val@); @}
| exp '/' exp { 
		if ($3 == 0)
		  yyerror("invalid division by zero");
	      }
	      @{ @i @exp.0.Val@ = (@exp.1.Val@ / @exp.2.Val@); @}
| '(' exp ')' @{ @i @exp.0.Val@ = @exp.1.Val@; @};
%%

#if defined (USE_FLEXCXX)

#include <FlexLexer.h>

int yylex (void)
{
  return ::drv.theScanner->yylex ();
}

#else

extern int yylex (void);

#endif

int
driver::parse (const std::string& f)
{
  file = f;
  scan_begin ();
  yydebug = trace_parsing;
  int res = yyparse ();
  scan_end ();
  return res;
}
