%{
#include <cstdio>

#include "types.hh"
#include "parser.hh"
#include "scan.hh"
#include "util.hh"

#define YYPOSN YYLTYPE
#define YYREDUCEPOSNFUNC ReducePosn

%}

%union {
    int       iVal;
    symbolP   sVal;
    typeSpecP tVal;
    enumListP eVal;
}

%token CONST TYPE DOTDOT
%token <sVal> ID
%token <iVal> NUM

%type  <iVal> constant expr
%type  <tVal> type simple_type
%type  <eVal> id_list

%left '+' '-'

%start decls

@attributes { YYLTYPE loc; } '+' '-' '(' ')'
@autoinh table
@attributes { symTable table; } decl_list
@attributes { symTable table; } decl const_decl const_defs type_decl type_defs
@attributes { symTable table; int iVal; YYLTYPE loc; } constant expr NUM
@attributes { char *sym; YYLTYPE loc; } ID
@attributes { symTable table; symbolP symTabEl; YYLTYPE loc; } const_def type_def
@attributes { symTable table; typeSpecP tVal; YYLTYPE loc; } type simple_type
@attributes { symTable table; enumListP eVal; YYLTYPE loc; } id_list

%%

decls	:	decl_list	@{ @i @decl_list.table@ = g_SymbolTable(); @}
      	;

decl_list
	:	decl
	|	decl_list decl
	;

decl	:	const_decl
	|	type_decl
	;

const_decl
	:	CONST const_defs ';'
	;

const_defs
	:	const_def
	|	const_defs ';' const_def
	;

const_def
	:	ID '=' constant      { newConst(locSpan(YYPOSNARG(1),YYPOSNARG(3)),$1,$3); }
				    @{ @i @const_def.loc@ = locSpan(@ID.loc@,@constant.loc@);
				       @i @const_def.symTabEl@ =
					  newOxConst(@const_def.table@,@const_def.loc@,@ID.sym@,@constant.iVal@);
				    @}
	;

constant:	NUM                  { $$ = $1; }
				    @{ @i @constant.iVal@ = @NUM.iVal@;
				       @i @constant.loc@ = @NUM.loc@;
				    @}
	|	'+' NUM              { $$ = $2; }
				    @{ @i @constant.iVal@ = @NUM.iVal@;
				       @i @constant.loc@ = locSpan(@'+'.loc@,@NUM.loc@);
				    @}
	|	'-' NUM              { $$ = -$2; }
				    @{ @i @constant.iVal@ = -@NUM.iVal@;
				       @i @constant.loc@ = locSpan(@'-'.loc@,@NUM.loc@);
				    @}
	|	ID                   { $$ = valOf(YYPOSNARG(1),$1); }
				    @{ @i @constant.iVal@ = oxValOf(@constant.table@,@ID.loc@,@ID.sym@);
				       @i @constant.loc@ = @ID.loc@;
				    @}
	|	'+' ID               { $$ = valOf(YYPOSNARG(2),$2); }
				    @{ @i @constant.iVal@ = oxValOf(@constant.table@,@ID.loc@,@ID.sym@);
				       @i @constant.loc@ = locSpan(@'+'.loc@,@ID.loc@);
				    @}
	|	'-' ID               { $$ = -valOf(YYPOSNARG(2),$2); }
				    @{ @i @constant.iVal@ = -oxValOf(@constant.table@,@ID.loc@,@ID.sym@);
				       @i @constant.loc@ = locSpan(@'-'.loc@,@ID.loc@);
				    @}
	;

type_decl
	:	TYPE type_defs ';'
	;

type_defs
	:	type_def
	|	type_defs ';' type_def
	;

type_def:	ID '=' type          { newType(locSpan(YYPOSNARG(1),YYPOSNARG(3)),$1,$3); }
				    @{ @i @type_def.loc@ = locSpan(@ID.loc@,@type.loc@);
				       @i @type_def.symTabEl@ =
					  newOxType(@type_def.table@,@type_def.loc@,@ID.sym@,@type.tVal@);
				    @}
	;

type	:	simple_type
				    @{ @i @type.loc@ = @simple_type.loc@;
				       @i @type.tVal@ = @simple_type.tVal@;
				    @}
	;

	/* the rules for `simple_type' may require backtracking to disambiguate */
simple_type
	:	'(' id_list ')'      { $$ = makeEnumType(YYPOSNARG(2),$2); }
				    @{ @i @simple_type.loc@ = locSpan(@'('.loc@,@')'.loc@);
				       @i @simple_type.tVal@ =
					   makeEnumType(@id_list.loc@,@id_list.eVal@);
				    @}
	|	expr DOTDOT expr     { $$ = makeSubrangeType(locSpan(YYPOSNARG(1),YYPOSNARG(3)),$1,$3); }
				    @{ @i @simple_type.loc@ = locSpan(@expr.0.loc@,@expr.1.loc@);
				       @i @simple_type.tVal@ =
					  makeSubrangeType(@simple_type.loc@,
							   @expr.0.iVal@,@expr.1.iVal@);
				    @}
	|	ID                   { $$ = typeOf(YYPOSNARG(1),$1); }
				    @{ @i @simple_type.loc@ = @ID.loc@;
				       @i @simple_type.tVal@ = oxTypeOf(@simple_type.table@,@ID.loc@,@ID.sym@);
				    @}
	;

id_list	:	ID                   { $$ = newEnum(YYPOSNARG(1),$1); }
				    @{ @i @id_list.loc@ = @ID.loc@;
				       @i @id_list.eVal@ = newOxEnum(@id_list.table@,@ID.loc@,@ID.sym@);
				    @}
	|	id_list ',' ID       { $$ = addToEnum(YYPOSNARG(3),$1,$3); }
				    @{ @i @id_list.0.loc@ = locSpan(@id_list.1.loc@,@ID.loc@);
				       @i @id_list.0.eVal@ = addToOxEnum(@id_list.table@,
									 @ID.loc@,
									 @id_list.1.eVal@,
								         @ID.sym@);
				    @}
	;

expr	:	'(' expr ')'         { $$ = $2; }
				    @{ @i @expr.0.iVal@ = @expr.1.iVal@;
				       @i @expr.0.loc@ = locSpan(@'('.loc@,@')'.loc@);
				    @}
	|	'+' expr             { $$ = $2; }
				    @{ @i @expr.0.iVal@ = @expr.1.iVal@;
				       @i @expr.0.loc@ = locSpan(@'+'.loc@,@expr.1.loc@);
				    @}
	|	'-' expr             { $$ = -$2; }
				    @{ @i @expr.0.iVal@ = -@expr.1.iVal@;
				       @i @expr.0.loc@ = locSpan(@'-'.loc@,@expr.1.loc@);
				    @}
	|	expr '+' expr        { $$ = $1 + $3; }
				    @{ @i @expr.0.iVal@ = @expr.1.iVal@ + @expr.2.iVal@;
				       @i @expr.0.loc@ = locSpan(@expr.1.loc@,@expr.2.loc@);
				    @}
	|	expr '-' expr        { $$ = $1 - $3; }
				    @{ @i @expr.0.iVal@ = @expr.1.iVal@ - @expr.2.iVal@;
				       @i @expr.0.loc@ = locSpan(@expr.1.loc@,@expr.2.loc@);
				    @}
	|	ID                   { $$ = valOf(YYPOSNARG(1),$1); }
				    @{ @i @expr.iVal@ = oxValOf(@expr.table@,@ID.loc@,@ID.sym@);
				       @i @expr.loc@ = @ID.loc@;
				    @}
	|	NUM                  { $$ = $1; }
				    @{ @i @expr.iVal@ = @NUM.iVal@;
				       @i @expr.loc@ = @NUM.loc@;
				    @}
	;

%%

#define YYLLOC_DEFAULT(loc, rhs, n) \
do \
{ \
    if (n == 0) \
    { \
        (loc).first_line   = ((rhs)[-1]).last_line; \
        (loc).first_column = ((rhs)[-1]).last_column; \
        (loc).last_line    = ((rhs)[-1]).last_line; \
        (loc).last_column  = ((rhs)[-1]).last_column; \
    } \
    else \
    { \
        (loc).first_line   = ((rhs)[ 0 ]).first_line; \
        (loc).first_column = ((rhs)[ 0 ]).first_column; \
        (loc).last_line    = ((rhs)[n-1]).last_line; \
        (loc).last_column  = ((rhs)[n-1]).last_column; \
    } \
} while (0)

#define UserType void *

void ReducePosn(YYPOSN  &ret,
		YYPOSN  *terms,
		YYSTYPE *term_vals,
		int      term_no,
		int      stk_pos,
		int      yychar,
		YYPOSN  &yyposn)
{YYLLOC_DEFAULT(ret,terms,term_no);}

/* redefine `YYCALLREDUCEPOSN(e)' to ignore 'e' */
#undef YYCALLREDUCEPOSN
#define YYCALLREDUCEPOSN(e)   \
	if(reduce_posn) {     \
	  YYREDUCEPOSNFUNC(yyps->pos, &(yyps->psp)[1-yym], &(yyps->vsp)[1-yym], \
			   yym, yyps->psp - yyps->ps, yychar, yyposn);          \
	  reduce_posn = 0;    \
	}

/* redefine `YYREDUCEPOSNFUNCARG' as empty */
#undef YYREDUCEPOSNFUNCARG
#define YYREDUCEPOSNFUNCARG
