#include <cstdlib>
#include <cstring>
#include <cstdio>

#include "types.hh"
#include "parser.hh"
#include "scan.hh"
#include "util.hh"

int
yylex(void)
{
   extern int yy_lex(YYSTYPE *,YYLTYPE *);
   return yy_lex(&yylval,&yylloc);
}

void
yyerror_detailed(const char *s, int errt, YYSTYPE errt_value, YYLTYPE errt_posn)
{
  printf("%s @ [%d:%d,%d:%d]\n", s, errt_posn.first_line, errt_posn.first_column,
				    errt_posn.last_line , errt_posn.last_column);
}


/* from the lexer */
extern int
g_input(void);

void
eat_comment(const char *end,int *yycolno,int *yylineno) {
    switch (end[0]) {
    case '}':
        { char c;
          while((c = g_input()) != '}') ;
	  return;
        }
    case '*':
        { char c, lastc=' ';
          while((c = g_input())) {
	      if(lastc=='*' && c==')')
	          break;
              else
	          lastc = c;
          }
        }
    }
}

#define MAXSYMBOLS 1000

/* Yacc-attribute maintained symbol table */
struct symbolTableItem yaccSymTab[MAXSYMBOLS];
int nextYaccSym = 0;

symbolP
addYaccSymbolIfNew(YYLTYPE *yyloc_sym, char *sym) {
    char *s, *n;

    for(int i=0; i < nextYaccSym; i++)
        if (strcmp(sym, yaccSymTab[i].sym) == 0) return &yaccSymTab[i];

    if (nextYaccSym == MAXSYMBOLS) {
        yyerror_detailed("addYaccSymbolIfNew: MAXSYMBOLS exceeded",
			 0,       /* not used, but required by btyaccpa.ske */
			 yylval, /* not used, but required by btyaccpa.ske */
			 *yyloc_sym
			);
        exit(-1);
    }

    if ((n = (char *)malloc(strlen(sym)+1)) != NULL)
        s = strcpy(n,sym);
    else {
        yyerror_detailed("addYaccSymbolIfNew: malloc() failure",
			 0,      /* not used, but required by btyaccpa.ske */
			 yylval, /* not used, but required by btyaccpa.ske */
			 *yyloc_sym
			);
        exit(-1);
    }

    yaccSymTab[nextYaccSym].sym = s;
    yaccSymTab[nextYaccSym].isConstVal = false;
    yaccSymTab[nextYaccSym].constVal   = 0;
    yaccSymTab[nextYaccSym].isEnumLit  = false;
    yaccSymTab[nextYaccSym].whichEnum  = NULL;
    yaccSymTab[nextYaccSym].isTypeName = false;
    yaccSymTab[nextYaccSym].whichType  = NULL;

    return &yaccSymTab[nextYaccSym++];
}

void
dumpYaccSymbolTable(void) {
    for(int i=0; i < nextYaccSym; i++) {
        fprintf(stdout, "symbol %s:", yaccSymTab[i].sym);
        if (yaccSymTab[i].isConstVal)
            fprintf(stdout, " constVal = %d", yaccSymTab[i].constVal);
        if (yaccSymTab[i].isEnumLit)
            fprintf(stdout, " isEnumLiteral = %d", yaccSymTab[i].whichEnum->constVal);
        if (yaccSymTab[i].isTypeName) {
            fputs(" isType", stdout);
            if (yaccSymTab[i].whichType != NULL)
                switch (yaccSymTab[i].whichType->tag) {
                case enumT:
                    fputs(" enum", stdout);
                    break;
                case rangeT:
                    fprintf(stdout, " range[%d .. %d]",
                                    yaccSymTab[i].whichType->type.rangeType.start,
                                    yaccSymTab[i].whichType->type.rangeType.end);
                    break;
                }
        }
        fputc('\n', stdout);
    }
}

/* Ox-attribute maintained symbol table */
static struct symbolTableItem oxGlobalSymbolTable[MAXSYMBOLS];
int nextOxSym = 0;

symTable g_SymbolTable() { return (symTable)&oxGlobalSymbolTable; };

char *
addOxSymbolIfNew(symTable oxSymTab, YYLTYPE *yyloc_sym, char *sym) {
    char *s, *n;

    for(int i=0; i < nextOxSym; i++)
        if (strcmp(sym, oxSymTab[i].sym) == 0) return oxSymTab[i].sym;

    if (nextOxSym == MAXSYMBOLS) {
        yyerror_detailed("addOxSymbolIfNew: MAXSYMBOLS exceeded",
			 0,      /* not used, but required by btyaccpa.ske */
			 yylval, /* not used, but required by btyaccpa.ske */
			 *yyloc_sym
			 );
        exit(-1);
    }

    if ((n = (char *)malloc(strlen(sym)+1)) != NULL)
        s = strcpy(n,sym);
    else {
        yyerror_detailed("addOxSymbolIfNew: malloc() failure",
			 0,      /* not used, but required by btyaccpa.ske */
			 yylval, /* not used, but required by btyaccpa.ske */
			 *yyloc_sym
			 );
        exit(-1);
    }

    oxSymTab[nextOxSym].sym = s;
    oxSymTab[nextOxSym].isConstVal = false;
    oxSymTab[nextOxSym].constVal   = 0;
    oxSymTab[nextOxSym].isEnumLit  = false;
    oxSymTab[nextOxSym].whichEnum  = NULL;
    oxSymTab[nextOxSym].isTypeName = false;
    oxSymTab[nextOxSym].whichType  = NULL;
    nextOxSym++;

    return s;
}

void
dumpOxSymbolTable(void) {
    static struct symbolTableItem *oxSymTab = (symTable)&oxGlobalSymbolTable;
    for(int i=0; i < nextOxSym; i++) {
        fprintf(stdout, "symbol %s:", oxSymTab[i].sym);
        if (oxSymTab[i].isConstVal)
            fprintf(stdout, " constVal = %d", oxSymTab[i].constVal);
        if (oxSymTab[i].isEnumLit)
            fprintf(stdout, " isEnumLiteral = %d", oxSymTab[i].whichEnum->constVal);
        if (oxSymTab[i].isTypeName) {
            fputs(" isType", stdout);
            if (oxSymTab[i].whichType != NULL)
                switch (oxSymTab[i].whichType->tag) {
                case enumT:
                    fputs(" enum", stdout);
                    break;
                case rangeT:
                    fprintf(stdout, " range[%d .. %d]",
                                    oxSymTab[i].whichType->type.rangeType.start,
                                    oxSymTab[i].whichType->type.rangeType.end);
                    break;
                }
        }
        fputc('\n', stdout);
    }
}

bool
cmpYaccToOxSymbolTable(void) {
    static struct symbolTableItem *oxSymTab = (symTable)&oxGlobalSymbolTable;
    bool result = true;

    if (nextYaccSym != nextOxSym) {
        fprintf(stdout, "sizeof(yaccSymTab) != sizeof(oxSymTab)\n");
        return false;
    }

    for(int i=0; i < nextOxSym; i++) {
        if (strcmp(yaccSymTab[i].sym,oxSymTab[i].sym) != 0) result = false;
        if (yaccSymTab[i].isConstVal != oxSymTab[i].isConstVal) result = false;
        if (yaccSymTab[i].isEnumLit  != oxSymTab[i].isEnumLit ) result = false;
        if (yaccSymTab[i].isTypeName != oxSymTab[i].isTypeName) result = false;
        else if (yaccSymTab[i].isTypeName) {
            if (yaccSymTab[i].whichType == NULL && oxSymTab[i].whichType == NULL) result = false;
            else {
                if (yaccSymTab[i].whichType->tag != oxSymTab[i].whichType->tag) result = false;
                else {
                    switch (yaccSymTab[i].whichType->tag) {
                    case enumT:
                        /* missing: compare the list of enum literals */
                        break;
                    case rangeT:
                        if ((yaccSymTab[i].whichType->type.rangeType.start !=
                             oxSymTab[i].  whichType->type.rangeType.start) ||
                            (yaccSymTab[i].whichType->type.rangeType.end !=
                             oxSymTab[i].  whichType->type.rangeType.end))
                            result = false;
                        break;
                    }
                }
            }
        }
    }

    return result;
}

void
newConst(YYLTYPE yyloc, symbolP sym, int iVal) {
    if (sym->isConstVal) {
        const char msg1[] = "newConst: [";
        const char msg2[] = "] is already a constant";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym->sym)+strlen(msg2)+1);
        if (msg == NULL) {
	    yyerror_detailed("newConst: malloc() failure",
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc
			     );
            exit(-1);
        } else
	    yyerror_detailed(strcat(strcat(strcpy(msg,msg1),sym->sym),msg2),
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc
			     );
    } else {
        sym->isConstVal = true;
        sym->constVal   = iVal;
    }
}

symbolP
newOxConst(symTable oxSymTab, YYLTYPE yyloc, char *sym, int iVal) {
    int which;

    for(which=0; which  < nextOxSym; which++)
        if (strcmp(sym, oxSymTab[which].sym) == 0) break;

    if (oxSymTab[which].isConstVal) {
        const char msg1[] = "newOxConst: [";
        const char msg2[] = "] is already a constant";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym)+strlen(msg2)+1);
        if (msg == NULL) {
	    yyerror_detailed("newOxConst: malloc() failure",
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc
			     );
            exit(-1);
        } else
	    yyerror_detailed(strcat(strcat(strcpy(msg,msg1),sym),msg2),
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc
			     );
    } else {
        oxSymTab[which].isConstVal = true;
        oxSymTab[which].constVal   = iVal;
    }

    return &oxSymTab[which];
}


YYLTYPE
locSpan(YYLTYPE left, YYLTYPE right) {
    YYLTYPE loc;
    loc.first_line   = left.first_line;
    loc.first_column = left.first_column;
    loc.last_line    = right.last_line;
    loc.last_column  = right.last_column;
    return loc;
}

int
valOf(YYLTYPE yyloc_sym, symbolP sym) {
    if (!sym->isConstVal) {
        const char msg1[] = "valOf: [";
        const char msg2[] = "] is not a constant";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym->sym)+strlen(msg2)+1);
        if (msg == NULL) {
	    yyerror_detailed("valOf: malloc() failure",
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc_sym
			     );
            exit(-1);
        } else {
	    yyerror_detailed(strcat(strcat(strcpy(msg,msg1),sym->sym),msg2),
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc_sym
			     );
            return 0;
        }
    } else
        return sym->constVal;
}

int
oxValOf(symTable oxSymTab, YYLTYPE yyloc_sym, char *sym) {
    int which;

    for(which=0; which  < nextOxSym; which++)
        if (strcmp(sym, oxSymTab[which].sym) == 0) break;

    if (!oxSymTab[which].isConstVal) {
        const char msg1[] = "oxValOf: [";
        const char msg2[] = "] is not a constant";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym)+strlen(msg2)+1);
        if (msg == NULL) {
	    yyerror_detailed("oxValOf: malloc() failure",
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc_sym
			     );
            exit(-1);
        } else {
	    yyerror_detailed(strcat(strcat(strcpy(msg,msg1),sym),msg2),
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc_sym
			     );
            return 0;
        }
    } else
        return oxSymTab[which].constVal;
}

void
newType(YYLTYPE yyloc, symbolP sym, typeSpecP spec) {
    if (sym->isTypeName) {
        const char msg1[] = "newType: [";
        const char msg2[] = "] is already a typename";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym->sym)+strlen(msg2)+1);
        if (msg == NULL) {
	    yyerror_detailed("newType: malloc() failure",
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc
			     );
            exit(-1);
        } else
	    yyerror_detailed(strcat(strcat(strcpy(msg,msg1),sym->sym),msg2),
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc
			     );
    }

    sym->isTypeName = true;
    sym->whichType  = spec;
}

symbolP
newOxType(symTable oxSymTab, YYLTYPE yyloc, char *sym, typeSpecP spec) {
    int which;

    for(which=0; which  < nextOxSym; which++)
        if (strcmp(sym, oxSymTab[which].sym) == 0) break;

    if (oxSymTab[which].isTypeName) {
        const char msg1[] = "newOxType: [";
        const char msg2[] = "] is already a typename";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym)+strlen(msg2)+1);
        if (msg == NULL) {
	    yyerror_detailed("newOxType: malloc() failure",
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc
			     );
            exit(-1);
        } else
	    yyerror_detailed(strcat(strcat(strcpy(msg,msg1),sym),msg2),
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc
			     );
    }

    oxSymTab[which].isTypeName = true;
    oxSymTab[which].whichType  = spec;

    return &oxSymTab[which];
}

typeSpecP
makeEnumType(YYLTYPE yyloc_list, enumListP list) {
    typeSpecP newTypeSpec = (typeSpecP)malloc(sizeof(struct typeSpec));
    if (newTypeSpec == NULL) {
	yyerror_detailed("makeEnumType: malloc() failure",
			 0,      /* not used, but required by btyaccpa.ske */
			 yylval, /* not used, but required by btyaccpa.ske */
			 yyloc_list
			 );
        exit(-1);
    }

    newTypeSpec->tag = enumT;
    newTypeSpec->type.enumType = *list;

    return newTypeSpec;
}

typeSpecP
makeSubrangeType(YYLTYPE yyloc, int start, int end) {
    typeSpecP newTypeSpec = (typeSpecP)malloc(sizeof(struct typeSpec));
    if (newTypeSpec == NULL) {
	yyerror_detailed("makeSubrangeType: malloc() failure",
			 0,      /* not used, but required by btyaccpa.ske */
			 yylval, /* not used, but required by btyaccpa.ske */
			 yyloc
			 );
        exit(-1);
    }

    newTypeSpec->tag = rangeT;
    newTypeSpec->type.rangeType.start = start;
    newTypeSpec->type.rangeType.end   = end;

    return newTypeSpec;
}

typeSpecP
typeOf(YYLTYPE yyloc_sym, symbolP sym) {
    if (sym->isTypeName)
        return sym->whichType;
    else {
        const char msg1[] = "typeOf: [";
        const char msg2[] = "] is not a typename";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym->sym)+strlen(msg2)+1);
        if (msg == NULL) {
	    yyerror_detailed("typeOf: malloc() failure",
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc_sym
			     );
            exit(-1);
        } else
	    yyerror_detailed(strcat(strcat(strcpy(msg,msg1),sym->sym),msg2),
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc_sym
			     );
        return NULL;
    }
}

typeSpecP
oxTypeOf(symTable oxSymTab, YYLTYPE yyloc_sym, char *sym) {
    int which;

    for(which=0; which  < nextOxSym; which++)
        if (strcmp(sym, oxSymTab[which].sym) == 0) break;

    if (oxSymTab[which].isTypeName)
        return oxSymTab[which].whichType;
    else {
        const char msg1[] = "oxTypeOf: [";
        const char msg2[] = "] is not a typename";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym)+strlen(msg2)+1);
        if (msg == NULL) {
	    yyerror_detailed("oxTypeOf: malloc() failure",
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc_sym
			     );
            exit(-1);
        } else
	    yyerror_detailed(strcat(strcat(strcpy(msg,msg1),sym),msg2),
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc_sym
			     );
        return (typeSpecP) 0;
    }
}

enumListP
newEnum(YYLTYPE yyloc_sym, symbolP sym) {
    if (sym->isEnumLit) {
        const char msg1[] = "newEnum: [";
        const char msg2[] = "] is already an enum literal";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym->sym)+strlen(msg2)+1);
        if (msg == NULL) {
	    yyerror_detailed("newEnum: malloc() failure",
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc_sym
			     );
            exit(-1);
        } else
	    yyerror_detailed(strcat(strcat(strcpy(msg,msg1),sym->sym),msg2),
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc_sym
			     );
    }

    enumListP newL = (enumListP)malloc(sizeof(struct enumList));
    if (newL == NULL) {
	yyerror_detailed("newEnum: malloc() failure",
			 0,      /* not used, but required by btyaccpa.ske */
			 yylval, /* not used, but required by btyaccpa.ske */
			 yyloc_sym
			 );
        exit(-1);
    }

    newL->enumLit  = sym;
    newL->constVal = 1;
    newL->next     = NULL;

    sym->isEnumLit = true;
    sym->whichEnum = newL;

    return newL;
}

enumListP
newOxEnum(symTable oxSymTab, YYLTYPE yyloc_sym, char *sym) {
    int which;

    for(which=0; which  < nextOxSym; which++)
        if (strcmp(sym, oxSymTab[which].sym) == 0) break;

    if (oxSymTab[which].isEnumLit) {
        const char msg1[] = "newOxEnum: [";
        const char msg2[] = "] is already an enum literal";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym)+strlen(msg2)+1);
        if (msg == NULL) {
	    yyerror_detailed("newOxEnum: malloc() failure",
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc_sym
			     );
            exit(-1);
        } else
	    yyerror_detailed(strcat(strcat(strcpy(msg,msg1),sym),msg2),
			     0,      /* not used, but required by btyaccpa.ske */
			     yylval, /* not used, but required by btyaccpa.ske */
			     yyloc_sym
			     );
    }

    enumListP newL = (enumListP)malloc(sizeof(struct enumList));
    if (newL == NULL) {
	yyerror_detailed("newOxEnum: malloc() failure",
			 0,      /* not used, but required by btyaccpa.ske */
			 yylval, /* not used, but required by btyaccpa.ske */
			 yyloc_sym
			 );
        exit(-1);
    }

    newL->enumLit  = &oxSymTab[which]; 
    newL->constVal = 1;
    newL->next     = NULL;

    oxSymTab[which].isEnumLit = true;
    oxSymTab[which].whichEnum = newL;

    return newL;
}

enumListP
addToEnum(YYLTYPE yyloc_sym, enumListP eL, symbolP sym) {
    enumListP tL = eL;
    enumListP newL = newEnum(yyloc_sym, sym);

    while (tL->next != NULL) tL = tL->next;

    newL->constVal = tL->constVal+1;

    tL->next = newL;

    return eL;
}

enumListP
addToOxEnum(symTable oxSymTab, YYLTYPE yyloc_sym, enumListP eL, char *sym) {
    enumListP tL = eL;
    enumListP newL = newOxEnum(oxSymTab,yyloc_sym, sym);

    while (tL->next != NULL) tL = tL->next;

    newL->constVal = tL->constVal+1;

    tL->next = newL;

    return eL;
}
