%{
#include <stdlib.h>

#include "types.h"
#include "util.h"

nameList *btYaccDeclList = NULL;
nameList *oxDeclList = NULL;

extern int yylex(void);
static void yyerror(const char *);
%}

%token <cVal> GLOBAL LOCAL
%token <tVal> REAL INTEGER
%token <id>   NAME

%type <nList> declaration_list declaration namelist(<cVal>, <tVal>) locnamelist(<tVal>)
%type <cVal>  class
%type <tVal>  type

%destructor	{ nameList *p = $$;
		  while (p != NULL)
		  { nameList *pp = p;
		    p = p->next;
		    free(pp->s); free(pp);
		  }
		} <nList>

%union
{
    classT	cVal;
    typeT	tVal;
    nameList *	nList;
    name	id;
}

%start input

@attributes { @autosyn classT cVal; } GLOBAL LOCAL class
@attributes { @autosyn typeT  tVal; } REAL INTEGER type
@attributes { @autosyn nameList * nList; } declaration_list declaration input
@attributes { nameList * nList; @autoinh classT cVal; @autoinh typeT tVal; } namelist
@attributes { @autosyn nameList * nList; typeT tVal; } locnamelist
@attributes { name id; } NAME

%%
input	:	declaration_list { btYaccDeclList = $1; }
	       @{ @i @input.nList@ = oxDeclList = @declaration_list.nList@; @}
	;

declaration_list
	:	declaration ';' { $$ = $1; }
	|	declaration_list declaration ';' { $$ = append($1, $2); }
	       @{ @i @declaration_list.0.nList@ = append(@declaration_list.1.nList@,
							 @declaration.nList@);
	       @}
	;

declaration
	:	class type namelist($1, $2) { $$ = $3; }
	       @{ @i @namelist.cVal@ = @class.cVal@;
		  @i @namelist.tVal@ = @type.tVal@;
		  @i @declaration.nList@ = @namelist.nList@;
	       @}
	|	type locnamelist($1) { $$ = $2; }
	       @{ @i @locnamelist.tVal@ = @type.tVal@;
		  @i @declaration.nList@ = @locnamelist.nList@;
	       @}
	;

class	:	GLOBAL
	|	LOCAL
	;

type	:	REAL
	|	INTEGER
	;

namelist($c, $t)
	:	namelist NAME 
		{ $$ = append($1, mkNameList(mkSymbol($<tVal>t, $<cVal>c, $2)));
		}
	       @{ @i @namelist.0.nList@ =
			append(@namelist.1.nList@,
			       mkNameList(mkSymbol(@namelist.0.tVal@,
						   @namelist.0.cVal@,
						   @NAME.id@
						  )
					 )
			      );
	       @}
	|	NAME
		{ $$ = mkNameList(mkSymbol($t, $c, $1)); }
	       @{ @i @namelist.nList@ =
			mkNameList(mkSymbol(@namelist.tVal@, @namelist.cVal@, @NAME.id@));
	       @}
	;

locnamelist($t)
	:	namelist(cLOCAL, $t) { $$ = $1; }
		@{ @i @namelist.cVal@ = cLOCAL; @}
	;
%%
static
void yyerror(const char *msg)
{
  fprintf (stderr, "%s\n", msg);
}
