%{
/* $Id: gppl.L 2539 2022-05-23 00:29:05Z teshields $ */ 

#include "gppl_glob.h" 
#include "scan.h"
#include "parser.h" 

%}

IDENT ([a-zA-Z_][a-zA-Z0-9_]*) 

%% 
int            return(INT);        @{ @INT.line@     = yylineno; @}  
float          return(TKFLOAT);    @{ @TKFLOAT.line@ = yylineno; @} 
bool           return(BOOL);       @{ @BOOL.line@    = yylineno; @} 
void           return(VOID);       @{ @VOID.line@    = yylineno; @} 
var            return(VAR); 
if             return(IF);         @{ @IF.line@      = yylineno; @} 
then           return(THEN); 
else           return(ELSE); 
while          return(WHILE);      @{ @WHILE.line@   = yylineno; @} 
return         return(RETURN);     @{ @RETURN.line@  = yylineno; @} 
exit           return(EXIT);       @{ @EXIT.line@    = yylineno; @} 
input          return(INPUT); 
output         return(OUTPUT); 
\"             {char *strDum;
                strDum = copyLexemeOfSTRING(); 
                return (STRING);  
                                   @{ @STRING.str@   = strDum;   @}
               }
\/\*           eatComment(); 
[ \n\t\f]*     ;  
[0-9]+\.[0-9]+ return(FCONST);     @{ sscanf((char *)yytext,"%f",
                                             &(@FCONST.val@.f)
                                            ); 
                                   @}
[0-9]+         return(ICONST);     @{ sscanf((char *)yytext,"%d",
                                             &(@ICONST.val@.i)
                                            ); 
                                   @}
EOF            return(EOFCONST);  
TRUE           return(BCONST);     @{ @BCONST.val@.b = 1;        @} 
FALSE          return(BCONST);     @{ @BCONST.val@.b = 0;        @} 
\(             return('(');        @{ @'('.line@     = yylineno; @}
\[             return('[');        @{ @'['.line@     = yylineno; @}
[,]            return(',');        @{ @','.line@     = yylineno; @}
[{]            return('{');
[}]            return('}');
[)]            return(')');
[;]            return(';');
[\]]           return(']');
&&             return(AND);        @{ @AND.line@     = yylineno; @} 
\|\|           return(OR);         @{ @OR.line@      = yylineno; @} 
!              return(NOT);        @{ @NOT.line@     = yylineno; @} 
[*]            return('*');        @{ @'*'.line@     = yylineno; @} 
[+]            return('+');        @{ @'+'.line@     = yylineno; @} 
[/]            return('/');        @{ @'/'.line@     = yylineno; @} 
[\-]           return('-');        @{ @'-'.line@     = yylineno; @} 
=              return(GETS);       @{ @GETS.line@    = yylineno; @} 
==             return(EQ);         @{ @EQ.line@      = yylineno; @} 
!=             return(NEQ);        @{ @NEQ.line@     = yylineno; @} 
"<"            return(LT);         @{ @LT.line@      = yylineno; @} 
">"            return(GT);         @{ @GT.line@      = yylineno; @} 
"<="           return(LEQ);        @{ @LEQ.line@     = yylineno; @} 
">="           return(GEQ);        @{ @GEQ.line@     = yylineno; @} 
{IDENT}        {
                return(ID);   
                @{ @ID.str@ = stringLookup((char *)yytext,yyleng,1); 
                   @ID.line@ = yylineno; 
                @} 
               } 
.              {fprintf(stderr,"line %d: illegal character\n",yylineno); 
                handleSemError();
               }
%%
#if (!defined(FLEX_SCANNER) && !defined(REFLEX_VERSION)) || !defined(__cplusplus)
int yyyinput(void){return input();} /* `lex', or `flex' with C compiler */
#else
int yyyinput(void){return yyinput();} /* `flex' with C++ compiler, or `reflex' */
#endif

int yywrap(void){return 1;}
