/* $Id: gppl.Y 1960 2020-12-26 20:06:21Z teshields $ */ 

/* a compiler for a general purpose programming language */ 

%{
#include "gppl_glob.h"
#include "scan.h"
#include "sem.h"
#include "code.h"
#include "util.h"

#if YYBTYACC && !YYBYACC
/* the `btyacc' 3.0 skeleton omits the following */
#include "parser.h"
#endif

extern int yylex(void);
%}


%start prog    

%token ID ICONST FCONST BCONST EOFCONST
%token INT TKFLOAT BOOL VOID 
%token VAR  
%token IF THEN ELSE WHILE   
%token EQ LT GT GEQ LEQ NEQ  
%token AND OR NOT  
%token INPUT OUTPUT STRING 
%token RETURN EXIT  

%right GETS
%left OR 
%left AND 
%nonassoc EQ NEQ  
%nonassoc LT LEQ GT GEQ   
%left '+' '-'
%left '*' '/' 
%right NOT

@traversal t 

@attributes {struct env *env; 
             regNumType maxRegNum;
            } 
            execElem statement 
            returnStatement exitStatement
            ifStatement whileStatement  

@attributes {struct env *env; 
             regNumType regNum;
             regNumType maxRegNum;
             struct sym *formParamList;
             struct sym *func;
             lineNumType line;
            } 
            actParamList    

@attributes {struct env *env; 
             struct sym *varLoc,*funcLoc; 
             regNumType maxRegNum;
            } 
            block blockElemList 

@attributes {struct sym *loc; 
             enum stgClass Class;
            } 
            stgDecl   

@attributes {struct sym *loc; 
             struct sym *funcSym; 
            } 
            paramDeclList   

@attributes {struct sym *loc; 
            } 
            paramDecl  

@attributes {struct env *env; 
             struct sym *loc; 
            } 
            funcDecl 

@attributes {enum symType type; 
             lineNumType line; 
            } 
            typeSpec  

@attributes {enum symType type; 
             struct sym *loc;
             enum stgClass Class; 
            } 
            stgDecl2 stgDecl2List 

@attributes {struct dim *dimList;
            }
            dimSeq

@attributes {struct env *env; 
             enum symType type; 
             regNumType regNum;
             regNumType maxRegNum;
            } 
            funcInvoc assignExpr 

@attributes {struct env *env; 
             enum symType type; 
             regNumType regNum;
             regNumType maxRegNum;
             struct sym *formalMatch; 
             enum pred isStgRef;
            } 
            expr simpleExpr  

@attributes {struct env *env; 
             regNumType regNum;
             regNumType maxRegNum;
             struct dim *dimList;  
             struct sym *sym;
             lineNumType line; 
            } 
            indexSeq

@attributes {struct env *env; 
             enum symType type; 
             regNumType regNum;
             regNumType maxRegNum;
             struct sym *formalMatch;
            } 
            stgRef 

@attributes {char *str; 
             lineNumType line; 
             struct sym *sym; 
            } 
            ID  

@attributes {char *str;
            }
            STRING

@attributes {lineNumType line; 
            } 
            AND OR NOT 
            GETS 
            GT LT EQ NEQ GEQ LEQ 
            '*' '+' '/' '-' '(' ',' '['
            FOR IF WHILE DO 
            INT TKFLOAT BOOL VOID
            RETURN EXIT

@attributes {labelType label;
             regNumType decisionRegNum;
            }
            then wDecision 

@attributes {labelType label;
            }
            else while

@attributes {union num val; 
            } 
            ICONST FCONST BCONST 

@attributes {long val; 
            } 
            iConstExpr

@macro constSimpleExpr(SYM,TYPE)
   @i @simpleExpr.type@ = TYPE;   
   @i @simpleExpr.maxRegNum@ = @simpleExpr.regNum@;
   @i @simpleExpr.isStgRef@ = False;
   @t cgConstSimpleExpr(@simpleExpr.regNum@,TYPE,@SYM.val@); 
@end


@macro unrOpExprdefs(op)
  @i @expr.type@ = unrOpTypeCheck(@expr.1.type@,op,@op.line@);
  @i @expr.1.env@ = @expr.env@;
  @i @expr.1.regNum@ = @expr.regNum@;
  @i @expr.maxRegNum@ = @expr.1.maxRegNum@;
  @i @expr.1.formalMatch@ = lambdaSym;
  @i @expr.isStgRef@ = False;
@end


@macro binOpExprdefs(op) 
  @i @expr.type@ = binOpTypeCheck(@expr.1.type@,@expr.2.type@,op,@op.line@);
  @i @expr.1.env@ = @expr.env@;
  @i @expr.2.env@ = @expr.env@;
  @i @expr.1.regNum@ = @expr.regNum@;
  @i @expr.2.regNum@ = @expr.regNum@ + 1;
  @i @expr.maxRegNum@ = max(@expr.1.maxRegNum@,@expr.2.maxRegNum@);
  @i @expr.1.formalMatch@ = lambdaSym;
  @i @expr.2.formalMatch@ = lambdaSym;
  @i @expr.isStgRef@ = False;
@end

@macro binArithExpr(op) 
  @t cgBinArithExpr(@expr.1.regNum@,op,@expr.type@); 
@end

@macro relExpr(op)
  @t cgRelExpr(@expr.1.regNum@,op,@expr.1.type@);
@end  

@macro binLogExpr(op)
  @t cgBinLogExpr(@expr.1.regNum@,op);
@end  


%%
prog	
	:	block	
		@{ @i @block.env@ = newEnv(lambdaEnv,
		                           lambdaSym, 
		                           @block.varLoc@,
		                           @block.funcLoc@,
		                           @block.maxRegNum@
		                          ); 
		@}  
	; 

block	
	:	'{'	blockElemList	'}' 
		@{ @i @blockElemList.env@ = @block.env@; 
		   @i @block.maxRegNum@ = @blockElemList.maxRegNum@;
		   @i @block.varLoc@ = @blockElemList.varLoc@; 
		   @i @block.funcLoc@ = @blockElemList.funcLoc@; 
		   @t cgBlockExit(@block.env@); 
		@} 
	; 

blockElemList 
	:
		@{ @i @blockElemList.varLoc@ = lambdaSym;
		   @i @blockElemList.funcLoc@ = lambdaSym; 
		   @i @blockElemList.maxRegNum@ = FIRSTREGNUM - 1;
		   @t cgBlockEnter(@blockElemList.env@); 
		@}
	|	blockElemList	execElem	
		@{ @i @blockElemList.1.env@ = @blockElemList.env@; 
		   @i @blockElemList.maxRegNum@ = 
		         max(@blockElemList.1.maxRegNum@,
		             @execElem.maxRegNum@ 
		            ); 
		   @i @blockElemList.varLoc@ = @blockElemList.1.varLoc@; 
		   @i @blockElemList.funcLoc@ = @blockElemList.1.funcLoc@; 
		   @i @execElem.env@ = @blockElemList.env@; 
		@} 
	|	blockElemList	stgDecl	
		@{ @i @blockElemList.1.env@ = @blockElemList.env@; 
		   @i @blockElemList.maxRegNum@ = @blockElemList.1.maxRegNum@;
		   @i @blockElemList.varLoc@ =  
		        concatLocs(@blockElemList.1.varLoc@,@stgDecl.loc@); 
		   @i @blockElemList.funcLoc@ = @blockElemList.1.funcLoc@; 
		   @i @stgDecl.Class@ = VARIABLEC;
		@}
	|	blockElemList	funcDecl	
		@{ @i @blockElemList.1.env@ = @blockElemList.env@; 
		   @i @blockElemList.maxRegNum@ = @blockElemList.1.maxRegNum@;
		   @i @blockElemList.varLoc@ = @blockElemList.1.varLoc@; 
		   @i @blockElemList.funcLoc@ = 
		        concatLocs(@blockElemList.1.funcLoc@,@funcDecl.loc@); 
		   @i @funcDecl.env@ = @blockElemList.env@; 
		@}
	; 
 
execElem	
	:	statement
		@{ @i @statement.env@ = @execElem.env@; 
		   @i @execElem.maxRegNum@ = @statement.maxRegNum@;
		@}
	|	block	
		@{ @i @block.env@ = newEnv(@execElem.env@,
		                           lambdaSym,
		                           @block.varLoc@,
		                           @block.funcLoc@,
		                           @block.maxRegNum@
		                          ); 
		   @i @execElem.maxRegNum@ = FIRSTREGNUM - 1;
		@} 
	;

statement 
	:	';'	
		@{ @i @statement.maxRegNum@ = FIRSTREGNUM - 1;
		@}
	|	expr	';'	
		@{ @i @expr.env@ = @statement.env@; 
		   @i @expr.regNum@ = FIRSTREGNUM; 
		   @i @statement.maxRegNum@ = @expr.maxRegNum@;
		   @i @expr.formalMatch@ = lambdaSym; 
		@}
	|	ifStatement
		@{ @i @ifStatement.env@ = @statement.env@; 
		   @i @statement.maxRegNum@ = @ifStatement.maxRegNum@; 
		@}
	|	whileStatement
		@{ @i @whileStatement.env@ = @statement.env@; 
		   @i @statement.maxRegNum@ = @whileStatement.maxRegNum@; 
		@}
	|	returnStatement
		@{ @i @returnStatement.env@ = @statement.env@; 
		   @i @statement.maxRegNum@ = @returnStatement.maxRegNum@; 
		@}
	|	exitStatement
		@{ @i @exitStatement.env@ = @statement.env@; 
		   @i @statement.maxRegNum@ = @exitStatement.maxRegNum@; 
		@}
	; 

expr	
	:	assignExpr
		@{ @i @expr.type@ = @assignExpr.type@;
		   @i @assignExpr.env@ = @expr.env@;
		   @i @assignExpr.regNum@ = @expr.regNum@;
		   @i @expr.maxRegNum@ = @assignExpr.maxRegNum@;
		   @i @expr.isStgRef@ = False;
		@}
	|	simpleExpr
		@{ @i @expr.type@ = @simpleExpr.type@;
		   @i @simpleExpr.env@ = @expr.env@;
		   @i @simpleExpr.regNum@ = @expr.regNum@; 
		   @i @expr.maxRegNum@ = @simpleExpr.maxRegNum@;
		   @i @simpleExpr.formalMatch@ = @expr.formalMatch@;
		   @i @expr.isStgRef@ = @simpleExpr.isStgRef@;
		@}
	|	'('	expr	')'
		@{ @i @expr.type@ = @expr.1.type@;
		   @i @expr.1.env@ = @expr.env@;
		   @i @expr.1.regNum@ = @expr.regNum@; 
		   @i @expr.maxRegNum@ = @expr.1.maxRegNum@;
		   @i @expr.1.formalMatch@ = @expr.formalMatch@;  
		   @i @expr.isStgRef@ = @expr.1.isStgRef@;
		@}
	|	'-'	expr		
		@{ unrOpExprdefs('-') 
		   @t cgUnrArithExpr(@expr.regNum@,'-',@expr.type@); 
		@}
	|	'+'	expr	
		@{ unrOpExprdefs('+') 
		   @t cgUnrArithExpr(@expr.regNum@,'+',@expr.type@); 
		@}
	|	NOT	expr		
		@{ unrOpExprdefs(NOT) 
		   @t cgLogNegExpr(@expr.regNum@);
		@}
	|	expr	'*'	expr
		@{ binOpExprdefs('*') 
		   binArithExpr('*') 
		@}
	|	expr	'/'	expr
		@{ binOpExprdefs('/') 
		   binArithExpr('/') 
		@}
	|	expr	'+'	expr
		@{ binOpExprdefs('+') 
		   binArithExpr('+') 
		@}
	|	expr	'-'	expr
		@{ binOpExprdefs('-') 
		   binArithExpr('-') 
		@}
	|	expr	EQ	expr
		@{ binOpExprdefs(EQ) 
		   relExpr(EQ)
		@}
	|	expr	NEQ	expr
		@{ binOpExprdefs(NEQ) 
		   relExpr(NEQ)
		@}
	|	expr	LT	expr
		@{ binOpExprdefs(LT) 
		   relExpr(LT)
		@}
	|	expr	GT	expr
		@{ binOpExprdefs(GT) 
		   relExpr(GT)
		@}
	|	expr	LEQ	expr
		@{ binOpExprdefs(LEQ) 
		   relExpr(LEQ)
		@}
	|	expr	GEQ	expr
		@{ binOpExprdefs(GEQ) 
		   relExpr(GEQ)
		@}
	|	expr	AND	expr
		@{ binOpExprdefs(AND) 
		   binLogExpr(AND)
		@}
	|	expr	OR	expr
		@{ binOpExprdefs(OR) 
		   binLogExpr(OR)
		@}
	;

assignExpr	
	:	stgRef 	GETS	expr 
		@{ @i @assignExpr.type@ = assignmentTypeCheck(@stgRef.type@,
		                                              @expr.type@,
		                                              @GETS.line@
		                                             );
		   @i @stgRef.env@ = @assignExpr.env@;
		   @i @expr.env@ = @assignExpr.env@;
		   @i @expr.regNum@ = @assignExpr.regNum@ + 1; 
		   @i @stgRef.regNum@ = @assignExpr.regNum@; 
		   @i @assignExpr.maxRegNum@ = 
		         max(@stgRef.maxRegNum@,@expr.maxRegNum@);
		   @i @stgRef.formalMatch@ = lambdaSym; 
		   @i @expr.formalMatch@ = lambdaSym; 
		   @t cgAssign(@stgRef.regNum@,@assignExpr.type@); 
		@}
	;

simpleExpr	
	:	stgRef 
		@{ @i @simpleExpr.type@ = @stgRef.type@; 
		   @i @stgRef.env@ = @simpleExpr.env@;
		   @i @stgRef.regNum@ = @simpleExpr.regNum@;
		   @i @simpleExpr.maxRegNum@ = @stgRef.maxRegNum@;
		   @i @stgRef.formalMatch@ = @simpleExpr.formalMatch@; 
		   @i @simpleExpr.isStgRef@ = True;
		   @t cgCondDeref(@simpleExpr.formalMatch@,@simpleExpr.regNum@,
		                  @simpleExpr.type@
		                 ); 
		@}
	|	funcInvoc
		@{ @i @simpleExpr.type@ = @funcInvoc.type@; 
		   @i @funcInvoc.env@ = @simpleExpr.env@;
		   @i @funcInvoc.regNum@ = @simpleExpr.regNum@;
		   @i @simpleExpr.maxRegNum@ = @funcInvoc.maxRegNum@;
		   @i @simpleExpr.isStgRef@ = False;
		@}
 	|	ICONST
		@{ constSimpleExpr(ICONST,(INTT))   @}
 	|	FCONST
		@{ constSimpleExpr(FCONST,\(FLOATT\)) @}
 	|	BCONST
		@{ constSimpleExpr(BCONST,(BOOLT))  @}
 	|	EOFCONST
		@{ @i @simpleExpr.type@ = INTT;   
   		   @i @simpleExpr.maxRegNum@ = @simpleExpr.regNum@;
   		   @i @simpleExpr.isStgRef@ = False;
   		   @t cgEOF(@simpleExpr.regNum@); 
		@}
	;

iConstExpr	
 	:	ICONST
		@{ @i @iConstExpr.val@ = @ICONST.val@.i;  
		@}
	|	'-'	iConstExpr
		@{ @i @iConstExpr.val@ = - @iConstExpr.1.val@ ; 
		@}
	|	'+'	iConstExpr
		@{ @i @iConstExpr.val@ =   @iConstExpr.1.val@ ; 
		@}
	|	'('	iConstExpr	')'
		@{ @i @iConstExpr.val@ =   @iConstExpr.1.val@ ; 
		@}
	|	iConstExpr	'*'	iConstExpr
		@{ @i @iConstExpr.val@ = @iConstExpr.1.val@ * 
                                           @iConstExpr.2.val@ ;  
		@}
	|	iConstExpr	'+'	iConstExpr
		@{ @i @iConstExpr.val@ = @iConstExpr.1.val@ + 
                                           @iConstExpr.2.val@ ;  
		@}
	|	iConstExpr	'/'	iConstExpr
		@{ @i @iConstExpr.val@ = @iConstExpr.1.val@ / 
                                           @iConstExpr.2.val@ ;  
		@}
	|	iConstExpr	'-'	iConstExpr
		@{ @i @iConstExpr.val@ = @iConstExpr.1.val@ - 
                                           @iConstExpr.2.val@ ;  
		@}
	;

stgRef	
	:	ID	indexSeq
		@{ @i @ID.sym@ = 
                        stgRefLookup(@stgRef.env@,@ID.str@,@ID.line@);
		   @i @indexSeq.env@ = @stgRef.env@;
		   @i @indexSeq.regNum@ = @stgRef.regNum@;
		   @i @stgRef.type@ = @ID.sym@->type;
		   @i @indexSeq.dimList@ = 
		         adjustDimList(@stgRef.formalMatch@,@ID.sym@,@ID.line@);
		   @i @stgRef.maxRegNum@ = @indexSeq.maxRegNum@;
		   @i @indexSeq.sym@ = @ID.sym@;
		   @i @indexSeq.line@ = @ID.line@;
		@}
	;

indexSeq
	:	
		@{ @i @indexSeq.maxRegNum@ = @indexSeq.regNum@; 
		   @t cgEmptyIndexSeq(@indexSeq.env@,@indexSeq.sym@,
		         @indexSeq.line@,@indexSeq.regNum@,@indexSeq.dimList@
		                     );
		@}
	|	indexSeq	'['	expr	']'
		@{ @i @indexSeq.1.env@ = @indexSeq.env@;
		   @i @expr.env@ = @indexSeq.env@;
		   @i @indexSeq.1.regNum@ = @indexSeq.regNum@; 
		   @i @expr.regNum@ = @indexSeq.regNum@ + 1; 
		   @i @indexSeq.maxRegNum@ = max(@indexSeq.1.maxRegNum@,
		                               @expr.maxRegNum@
		                              );
		   @i @indexSeq.1.dimList@ = cdrDimList(@indexSeq.dimList@,
		                                        @indexSeq.sym@,
		                                        @indexSeq.line@
		                                       ); 
		   @i @indexSeq.1.sym@ = @indexSeq.sym@;
		   @i @indexSeq.1.line@ = @indexSeq.line@;
		   @i @expr.formalMatch@ = lambdaSym;
		   @t cgIndexSeq(@indexSeq.sym@,@indexSeq.regNum@,
		                 @indexSeq.dimList@
		                );
		@}
	;

funcInvoc
	:	ID	'('	')'
		@{ @i @funcInvoc.type@ = @ID.sym@->type; 
		   @i @ID.sym@ = 
		        funcInvocLookup(@funcInvoc.env@,@ID.str@,@ID.line@);
		   @i @funcInvoc.maxRegNum@ = @funcInvoc.regNum@;
		   @t cgFuncCall(@funcInvoc.env@,@ID.sym@,
				 @funcInvoc.regNum@,True,@ID.line@
		                ); 
		@}
	|	ID	'('	actParamList	')'
		@{ @i @actParamList.env@ = @funcInvoc.env@;
		   @i @funcInvoc.type@ = @ID.sym@->type; 
		   @i @ID.sym@ = 
		        funcInvocLookup(@funcInvoc.env@,@ID.str@,@ID.line@);
		   @i @actParamList.regNum@ = @funcInvoc.regNum@;
		   @i @funcInvoc.maxRegNum@ = @actParamList.maxRegNum@;
		   @i @actParamList.formParamList@ = @ID.sym@->paramLoc;
		   @i @actParamList.func@ = @ID.sym@;
		   @i @actParamList.line@ = @ID.line@;
		   @t cgFuncCall(@funcInvoc.env@,@ID.sym@,
		                 @funcInvoc.regNum@,False,@ID.line@
                                ); 
		@}
	|	typeSpec	'('	expr	')'
		@{ @i @expr.env@ = @funcInvoc.env@;
		   @i @funcInvoc.type@ = @typeSpec.type@; 
		   @i @expr.regNum@ = @funcInvoc.regNum@;
		   @i @funcInvoc.maxRegNum@ = @expr.maxRegNum@;
		   @i @expr.formalMatch@ = lambdaSym;
		   @t cgCast(@expr.regNum@,@typeSpec.type@,@expr.type@);
		@}
	|	INPUT	'('	STRING	','	stgRef	')' 
		@{ @i @stgRef.env@ = @funcInvoc.env@;
		   @i @funcInvoc.type@ = INTT; 
		   @i @stgRef.regNum@ = @funcInvoc.regNum@;
		   @i @funcInvoc.maxRegNum@ = @stgRef.maxRegNum@;
		   @i @stgRef.formalMatch@ = lambdaSym;
		   @t cgInput(@STRING.str@,@stgRef.regNum@,
		              @stgRef.type@,@'('.line@
		             ); 
		@}
	|	OUTPUT	'('	STRING	')' 
		@{ @i @funcInvoc.type@ = INTT; 
		   @i @funcInvoc.maxRegNum@ = @funcInvoc.regNum@;
		   @t cgOutputNoArg(@STRING.str@,@funcInvoc.regNum@); 
		@}
	|	OUTPUT	'('	STRING	','	expr	')' 
		@{ @i @expr.env@ = @funcInvoc.env@;
		   @i @funcInvoc.type@ = INTT; 
		   @i @expr.regNum@ = @funcInvoc.regNum@;
		   @i @funcInvoc.maxRegNum@ = @expr.maxRegNum@;
		   @i @expr.formalMatch@ = lambdaSym;
		   @t cgOutputOneArg(@STRING.str@,@expr.regNum@,@expr.type@); 
		@}
	;

actParamList
	:	expr
		@{ @i @expr.regNum@ = @actParamList.regNum@;
		   @i @actParamList.maxRegNum@ = @expr.maxRegNum@;
		   @i @expr.env@ = @actParamList.env@;
		   @i @expr.formalMatch@ = @actParamList.formParamList@; 
		   @t {paramMatchCheck(@actParamList.formParamList@,
				       @expr.type@,
				       @expr.isStgRef@,True,
				       @actParamList.func@,@actParamList.line@
				      ); 
		       cgPush(@actParamList.regNum@);}
		@}
	|	actParamList	','	expr	
		@{ @i @actParamList.1.env@ = @actParamList.env@;
		   @i @expr.env@ = @actParamList.env@;
		   @i @actParamList.1.regNum@ = @actParamList.regNum@;
		   @i @expr.regNum@ = @actParamList.regNum@;
		   @i @actParamList.maxRegNum@ = max(@actParamList.1.maxRegNum@,
		                                     @expr.maxRegNum@
		                                    );
		   @i @actParamList.1.formParamList@ = 
		        cdrParamList(@actParamList.formParamList@,
		                     @actParamList.func@,
		                     @actParamList.line@
		                    ); 
		   @i @actParamList.1.func@ = @actParamList.func@;
		   @i @actParamList.1.line@ = @actParamList.line@;
		   @i @expr.formalMatch@ = @actParamList.formParamList@; 
		   @t {paramMatchCheck(@actParamList.formParamList@,
				       @expr.type@,
				       @expr.isStgRef@,False,
				       @actParamList.func@,@actParamList.line@
				      ); 
		       cgPush(@actParamList.regNum@);}
		@}
	;

funcDecl
	:	typeSpec	ID	'('	paramDeclList	')'	block
		@{ @i @block.env@ = newEnv(@funcDecl.env@,
		                           @ID.sym@,
		                           @block.varLoc@,
		                           @block.funcLoc@,
		                           @block.maxRegNum@
		                          ); 
		   @i @funcDecl.loc@ = @ID.sym@; 
		   @i @ID.sym@ = newFuncSym(@ID.str@,@ID.line@,
		                            @typeSpec.type@,
		                            newLabelNum(),
		                            @paramDeclList.loc@ 
		                           );
		   @i @paramDeclList.funcSym@ = @ID.sym@; 
		   @t cgFuncTrailer(@ID.sym@); 
		@} 
	;

paramDeclList
	:
		@{ @i @paramDeclList.loc@ = lambdaSym; 
		   @t cgFuncAddress(@paramDeclList.funcSym@); 
		@}
	|	paramDeclList	paramDecl
		@{ @i @paramDeclList.loc@ = 
                        concatLocs(@paramDecl.loc@,@paramDeclList.1.loc@); 
		   @i @paramDeclList.1.funcSym@ = @paramDeclList.funcSym@; 
		@}
	;	

paramDecl	
	:		stgDecl		
		@{ @i @paramDecl.loc@ = @stgDecl.loc@; 
		   @i @stgDecl.Class@ = VALPARAMC;
		@}
	|	VAR	stgDecl	
		@{ @i @paramDecl.loc@ = @stgDecl.loc@; 
		   @i @stgDecl.Class@ = REFPARAMC;
		@}
	; 

stgDecl	
	:		typeSpec	stgDecl2List	';'
		@{ @i @stgDecl.loc@ = @stgDecl2List.loc@; 
		   @i @stgDecl2List.type@ = @typeSpec.type@;
		   @i @stgDecl2List.Class@ = @stgDecl.Class@;
		@}
	;

typeSpec
	:	INT
		@{ @i @typeSpec.type@ = INTT;
		   @i @typeSpec.line@ = @INT.line@;
		@}
	|	TKFLOAT
		@{ @i @typeSpec.type@ = FLOATT;
		   @i @typeSpec.line@ = @TKFLOAT.line@;
		@}
	|	BOOL
		@{ @i @typeSpec.type@ = BOOLT;
		   @i @typeSpec.line@ = @BOOL.line@;
		@}
	|	VOID
		@{ @i @typeSpec.type@ = VOIDT;
		   @i @typeSpec.line@ = @VOID.line@;
		@}
	;

stgDecl2List
	:	stgDecl2	
		@{ @i @stgDecl2List.loc@ = @stgDecl2.loc@;
		   @i @stgDecl2.type@ = @stgDecl2List.type@;
		   @i @stgDecl2.Class@ = @stgDecl2List.Class@;
		@}
	|	stgDecl2List	','	stgDecl2
		@{ @i @stgDecl2List.loc@ = concatLocs(@stgDecl2.loc@,
		                                      @stgDecl2List.1.loc@
		                                     ); 
		   @i @stgDecl2List.1.type@ = @stgDecl2List.type@;
		   @i @stgDecl2List.1.Class@ = @stgDecl2List.Class@;
		   @i @stgDecl2.type@ = @stgDecl2List.type@;
		   @i @stgDecl2.Class@ = @stgDecl2List.Class@;
		@}
	;

stgDecl2	
	:	ID	dimSeq
		@{ @i @stgDecl2.loc@ = @ID.sym@; 
		   @i @ID.sym@ = newStgSym(@ID.str@,@ID.line@,
		                           @stgDecl2.type@,
		                           @stgDecl2.Class@,
		                           @dimSeq.dimList@
		                          );
		@}
	;

dimSeq	:	
		@{ @i @dimSeq.dimList@ = lambdaDim; 
		@}
	|	dimSeq	'['	iConstExpr	']'
		@{ @i @dimSeq.dimList@ = 
		         consDimList(@iConstExpr.val@,@dimSeq.1.dimList@,
		                     @'['.line@  
		                    ); 
		@}
	;

ifStatement
	:	IF	expr	then	execElem	
		@{ @i @execElem.env@ = @ifStatement.env@; 
		   @i @expr.env@ = @ifStatement.env@;
		   @i @expr.regNum@ = FIRSTREGNUM;
		   @i @ifStatement.maxRegNum@ = max(@expr.maxRegNum@,
		                                    @execElem.maxRegNum@
		                                   );
		   @i @expr.formalMatch@ = lambdaSym;
		   @i @then.label@ = newLabelNum(); 
		   @i @then.decisionRegNum@ = @expr.regNum@;
		   @t cgIfThen(@expr.type@,@IF.line@,@then.label@); 
		@} 

	|	IF	expr	then	execElem	else	execElem
		@{ @i @expr.env@ = @ifStatement.env@; 
                   @i @execElem.env@ = @ifStatement.env@;  
		   @i @execElem.1.env@ = @ifStatement.env@; 
		   @i @expr.regNum@ = FIRSTREGNUM;
		   @i @ifStatement.maxRegNum@ = 
                         max(@expr.maxRegNum@,
		             max(@execElem.maxRegNum@,
		                 @execElem.1.maxRegNum@
		                )
		            );
		   @i @expr.formalMatch@ = lambdaSym;
		   @i @then.label@ = newLabelNum(); 
		   @i @then.decisionRegNum@ = @expr.regNum@;
		   @i @else.label@ = @then.label@; 
		   @t cgIfThenElse(@expr.type@,@IF.line@,@then.label@); 
		@} 
	;

then 	
	:	THEN
		@{ @t cgThen(@then.decisionRegNum@,@then.label@); 
		@}
	;

else	
	:	ELSE
		@{ @t cgElse(@else.label@); 
		@}
	;

whileStatement
	:	while	'(' 	expr	')'	wDecision	execElem
		@{ @i @expr.env@ = @whileStatement.env@; 
		   @i @execElem.env@ = @whileStatement.env@; 
		   @i @expr.regNum@ = FIRSTREGNUM;
		   @i @whileStatement.maxRegNum@ = max(@expr.maxRegNum@,
		                                       @execElem.maxRegNum@
		                                      );
		   @i @expr.formalMatch@ = lambdaSym;
		   @i @while.label@ = newLabelNum(); 
		   @i @wDecision.label@ = @while.label@;
		   @i @wDecision.decisionRegNum@ = @expr.regNum@;
		   @t cgWhileStatement(@expr.type@,@'('.line@,@while.label@); 
		@} 
	;

while	
	:	WHILE
		@{ @t cgWhile(@while.label@); 
		@}
	;

wDecision
	:
		@{ @t cgWdecision(@wDecision.decisionRegNum@,@wDecision.label@);
		@}
	;

returnStatement
	:	RETURN	expr	';'
		@{ @i @expr.env@ = @returnStatement.env@; 
		   @i @expr.regNum@ = FIRSTREGNUM;
		   @i @returnStatement.maxRegNum@ = @expr.maxRegNum@;
		   @i @expr.formalMatch@ = lambdaSym;
		   @t cgReturnStmt(@returnStatement.env@,@expr.regNum@,
		                   @expr.type@,@RETURN.line@
		                  ); 
		@} 
	|	RETURN	';'
		@{ @i @returnStatement.maxRegNum@ = FIRSTREGNUM - 1;
		   @t cgReturnStmt(@returnStatement.env@,DONTCARE,
		                   VOIDT,@RETURN.line@
		                  ); 
		@} 
	;

exitStatement
	:	EXIT	expr	';'
		@{ @i @expr.env@ = @exitStatement.env@; 
		   @i @expr.regNum@ = FIRSTREGNUM;
		   @i @exitStatement.maxRegNum@ = @expr.maxRegNum@;
		   @i @expr.formalMatch@ = lambdaSym;
		   @t cgExitStmt(@expr.regNum@,@expr.type@,@EXIT.line@); 
		@} 
	;

%% 
#include <stdio.h>

int yyparse(void);

int main(int argc, char *argv[]) 
  {int parseStatus;

   extern FILE *yyin;
   if (argc > 1)
     yyin = fopen (argv[argc-1], "r");
   else
    yyin = stdin;

   scanSetup(); 
   semSetup(); 
   tempOut = safeOpen(tempOutID,"w"); 
   cgHeader();  
   parseStatus = yyparse();
   cgTrailer();  
   fclose(tempOut); 
   return(parseStatus || (doCodeGen == False)); 
  } 



