%{
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "types.h"
#include "lit.h"
#include "sem.h"
#include "util.h"

bool dump_literals = false,
     dump_names = false;

extern int yylineno;

extern int yylex(void);
void yyerror(const char *s);
%}

%token _BECOMES			/* ':=' */
%token _BITS
%token _BITS_DENOTATION
%token _BOLD_TAG
%token _BOOL
%token _BYTES
%token _CHAR
%token _COMPL
%token _FALSE
%token _HEAP
%token _INT
%token _INTEGRAL_DENOTATION
%token _LOC
%token _LONG
%token _MODE
%token _PLAIN_TAG
%token _PROC
%token _REAL
%token _REAL_DENOTATION
%token _REF
%token _SHORT
%token _STRING
%token _STRING_DENOTATION
%token _STRUCT
%token _TRUE
%token _UNION
%token _VOID

%left '^'
%left '+' '-'
%left '>' '<'
%left '/' '*'
%nonassoc UMINUS

%start input

@attributes { int ct; @void done; @autoinh int size; @autoinh declSet *set; /* declSet set[] */ }
	      declaration_list

@attributes { declSet declSet; } declaration
	      mode_declaration variable_declaration identity_declaration

@attributes { int ct; @void done; @autoinh int size; @autoinh decl *set; /* decl set[] */ }
	      mode_joined_definition proc_identity_joined_definition

@attributes { mode_defn inh_mode; int ct; @void done; @autoinh int size; @autoinh decl *set; /* decl set[] */ }
	      variable_joined_definition identity_joined_definition

@attributes { decl dcl; } mode_definition proc_identity_definition

@attributes { decl dcl; mode_defn inh_mode; } variable_definition identity_definition 

@attributes { name_defn name; } _BOLD_TAG _PLAIN_TAG

@attributes { @autosyn mode_defn syn_mode; } declarer mode_indication mode_standard
              rows_declarator procedure_declarator formal_procedure_plan reference_declarator
              INT_sithety REAL_sithety BITS_sithety BYTES_sithety COMPL_sithety

@attributes { int ct; @void done; @autoinh int size; @autoinh mode_defn *set; /* mode set[] */ }
              joined_declarer

@attributes { @autosyn mode_defn syn_mode; }
              union_declarator structure_declarator

@attributes { int ct; @void done; @autoinh int size; @autoinh field **set; /* field set[][] */ }
              portrayer

@attributes { mode_defn inh_mode; int ct; @void done; @autoinh int size; @autoinh field *set; /* field set[] */ }
              joined_fields

@attributes { mode_defn syn_mode; } routine_text generator assignation

@attributes { @autosyn mode_defn syn_mode; @autoinh mode_defn inh_mode; @autoinh bool coerce; } unit formula

@attributes { lit_defn lit; mode_defn syn_mode; } denotation boolean_denotation
              BITS_DENOTATION_sithety INTEGRAL_DENOTATION_sithety
              REAL_DENOTATION_sithety STRING_DENOTATION_sithety

@attributes { lit_defn lit; } _BITS_DENOTATION _INTEGRAL_DENOTATION
              _REAL_DENOTATION _STRING_DENOTATION

@attributes { int num_short; } SHORT_sequence

@attributes { int num_long; } LONG_sequence

%%

input	:	declaration_list
		@{ @i @declaration_list.set@ =
			(declSet *)checked_calloc(@declaration_list.ct@, sizeof(declSet));
		   @i @declaration_list.size@ = @declaration_list.ct@;
		@}
	;

declaration_list
	:	declaration ';'
		@{ @i @declaration_list.0.ct@ = 1;
		   @m declaration_list.0.done ;
		      @declaration_list.0.set@[@declaration_list.0.size@ - 1] = @declaration.declSet@;
		@}
	|	declaration ';' declaration_list
		@{ @i @declaration_list.0.ct@ = @declaration_list.1.ct@ + 1;
		   @m declaration_list.0.done ;
		      @declaration_list.0.set@[@declaration_list.0.size@ -
						    @declaration_list.1.ct@ - 1] = @declaration.declSet@;
		@}
	;

declaration
	:	mode_declaration
		@{ @i @declaration.declSet@ = @mode_declaration.declSet@; @}
	|	variable_declaration
		@{ @i @declaration.declSet@ = @variable_declaration.declSet@; @}
	|	identity_declaration
		@{ @i @declaration.declSet@ = @identity_declaration.declSet@; @}
	;

mode_declaration
	:	_MODE mode_joined_definition
		@{ @m mode_declaration.declSet ;
			{@mode_declaration.declSet@.set  = @mode_joined_definition.set@;
			 @mode_declaration.declSet@.size = @mode_joined_definition.ct@;}
		   @i @mode_joined_definition.set@ =
			(decl *)checked_calloc(@mode_joined_definition.ct@, sizeof(decl));
		   @i @mode_joined_definition.size@ = @mode_joined_definition.ct@;
		@}
	;

mode_joined_definition
	:	mode_definition
		@{ @i @mode_joined_definition.0.ct@ = 1;
		   @m mode_joined_definition.0.done ;
			@mode_joined_definition.0.set@[@mode_joined_definition.0.size@ - 1] =
				@mode_definition.dcl@;
		@}
	|	mode_definition ',' mode_joined_definition
		@{ @i @mode_joined_definition.0.ct@ = @mode_joined_definition.1.ct@ + 1;
		   @m mode_joined_definition.0.done ;
		      @mode_joined_definition.0.set@[@mode_joined_definition.0.size@ -
						    @mode_joined_definition.1.ct@ - 1] =
				@mode_definition.dcl@;
		@}
	;

mode_definition
	:	_BOLD_TAG '=' declarer
		@{ @m mode_definition.dcl ;
			{s_mode_name(@_BOLD_TAG.name@);
			 s_mode(@_BOLD_TAG.name@,@declarer.syn_mode@);
			 @mode_definition.dcl@ = newModeDecl(@_BOLD_TAG.name@, @declarer.syn_mode@);}
		@}
	;

variable_declaration
	:	declarer variable_joined_definition
		@{ @m variable_declaration.declSet ;
			{@variable_declaration.declSet@.set  = @variable_joined_definition.set@;
			 @variable_declaration.declSet@.size = @variable_joined_definition.ct@;}
		   @i @variable_joined_definition.set@ =
			(decl *)checked_calloc(@variable_joined_definition.ct@, sizeof(decl));
		   @i @variable_joined_definition.size@ = @variable_joined_definition.ct@;
		   @i @variable_joined_definition.inh_mode@ = @declarer.syn_mode@;
		@}
	|	generator variable_joined_definition
		@{ @m variable_declaration.declSet ;
			{@variable_declaration.declSet@.set  = @variable_joined_definition.set@;
			 @variable_declaration.declSet@.size = @variable_joined_definition.ct@;}
		   @i @variable_joined_definition.set@ =
			(decl *)checked_calloc(@variable_joined_definition.ct@, sizeof(decl));
		   @i @variable_joined_definition.size@ = @variable_joined_definition.ct@;
		   @i @variable_joined_definition.inh_mode@ = @generator.syn_mode@;
		@}
	;

variable_joined_definition
	:	variable_definition
		@{ @i @variable_joined_definition.0.ct@ = 1;
		   @m variable_joined_definition.0.done ;
			@variable_joined_definition.0.set@[@variable_joined_definition.0.size@ - 1] =
				@variable_definition.dcl@;
		   @i @variable_definition.inh_mode@          = @variable_joined_definition.0.inh_mode@;
		@}
	|	variable_definition ',' variable_joined_definition
		@{ @i @variable_joined_definition.0.ct@ = @variable_joined_definition.1.ct@ + 1;
		   @m variable_joined_definition.0.done ;
		      @variable_joined_definition.0.set@[@variable_joined_definition.0.size@ -
						    @variable_joined_definition.1.ct@ - 1] =
				@variable_definition.dcl@;
		   @i @variable_definition.inh_mode@          = @variable_joined_definition.0.inh_mode@;
		   @i @variable_joined_definition.1.inh_mode@ = @variable_joined_definition.0.inh_mode@;
		@}
	;

variable_definition
	:	_PLAIN_TAG
		@{ @m variable_definition.dcl ;
			{s_ident_name(@_PLAIN_TAG.name@);
			 s_mode(@_PLAIN_TAG.name@,@variable_definition.inh_mode@);
			 @variable_definition.dcl@ = newDecl(@variable_definition.inh_mode@, @_PLAIN_TAG.name@);}
		@}
	|	_PLAIN_TAG _BECOMES unit
		@{ @m variable_definition.dcl ;
			{s_ident_name(@_PLAIN_TAG.name@);
			 s_mode(@_PLAIN_TAG.name@,@variable_definition.inh_mode@);
			 @variable_definition.dcl@ = newDecl(@variable_definition.inh_mode@, @_PLAIN_TAG.name@);}
		   @i @unit.inh_mode@           = @variable_definition.inh_mode@;
		   @i @unit.coerce@             = (@unit.inh_mode@.tag != @unit.syn_mode@.tag);
		@}
	;

identity_declaration
	:	declarer identity_joined_definition
		@{ @m identity_declaration.declSet ;
			{@identity_declaration.declSet@.set  = @identity_joined_definition.set@;
			 @identity_declaration.declSet@.size = @identity_joined_definition.ct@;}
		   @i @identity_joined_definition.set@ =
			(decl *)checked_calloc(@identity_joined_definition.ct@, sizeof(decl));
		   @i @identity_joined_definition.size@ = @identity_joined_definition.ct@;
		   @i @identity_joined_definition.inh_mode@ = @declarer.syn_mode@;
		@}
	|	_PROC proc_identity_joined_definition
		@{ @m identity_declaration.declSet ;
			{@identity_declaration.declSet@.set  = @proc_identity_joined_definition.set@;
			 @identity_declaration.declSet@.size = @proc_identity_joined_definition.ct@;}
		   @i @proc_identity_joined_definition.set@ =
			(decl *)checked_calloc(@proc_identity_joined_definition.ct@, sizeof(decl));
		   @i @proc_identity_joined_definition.size@ = @proc_identity_joined_definition.ct@;
		@}
	;

identity_joined_definition
	:	identity_definition
		@{ @i @identity_joined_definition.0.ct@ = 1;
		   @m identity_joined_definition.0.done  ;
			@identity_joined_definition.0.set@[@identity_joined_definition.0.size@ - 1] =
				@identity_definition.dcl@;
		   @i @identity_definition.inh_mode@ = @identity_joined_definition.0.inh_mode@;
		@}
	|	identity_definition ',' identity_joined_definition
		@{ @i @identity_joined_definition.0.ct@ = @identity_joined_definition.1.ct@ + 1;
		   @m identity_joined_definition.0.done  ;
		      @identity_joined_definition.0.set@[@identity_joined_definition.0.size@ -
						    @identity_joined_definition.1.ct@ - 1] =
				@identity_definition.dcl@;
		   @i @identity_definition.inh_mode@          = @identity_joined_definition.0.inh_mode@;
		   @i @identity_joined_definition.1.inh_mode@ = @identity_joined_definition.0.inh_mode@;
		@}
	;

identity_definition
	:	_PLAIN_TAG '=' unit
		@{ @m identity_definition.dcl ;
			{s_ident_name(@_PLAIN_TAG.name@);
			 s_mode(@_PLAIN_TAG.name@,@identity_definition.inh_mode@);
			 @identity_definition.dcl@ = newDecl(@identity_definition.inh_mode@, @_PLAIN_TAG.name@);}
		   @i @unit.inh_mode@           = @identity_definition.inh_mode@;
		   @i @unit.coerce@             = (@unit.inh_mode@.tag != @unit.syn_mode@.tag);
		@}
	;

proc_identity_joined_definition
	:	proc_identity_definition
		@{ @i @proc_identity_joined_definition.0.ct@ = 1;
		   @m proc_identity_joined_definition.0.done ;
			@proc_identity_joined_definition.0.set@[@proc_identity_joined_definition.0.size@ - 1] =
				@proc_identity_definition.dcl@;
		@}
	|	proc_identity_definition ',' proc_identity_joined_definition
		@{ @i @proc_identity_joined_definition.0.ct@ = @proc_identity_joined_definition.1.ct@ + 1;
		   @m proc_identity_joined_definition.0.done ;
		      @proc_identity_joined_definition.0.set@[@proc_identity_joined_definition.0.size@ -
						    @proc_identity_joined_definition.1.ct@ - 1] =
				@proc_identity_definition.dcl@;
		@}
	;

proc_identity_definition
	:	_PLAIN_TAG '=' routine_text
		@{ @m proc_identity_definition.dcl ;
			{s_ident_name(@_PLAIN_TAG.name@);
			 s_mode(@_PLAIN_TAG.name@,@routine_text.syn_mode@);
			 @proc_identity_definition.dcl@ = newDecl(@routine_text.syn_mode@, @_PLAIN_TAG.name@);}
		@}
	;

mode_indication
	:	mode_standard
	|	_BOLD_TAG
		@{ @i @mode_indication.syn_mode@ = gMode(@_BOLD_TAG.name@); @}
	;

declarer:	mode_indication
	|	structure_declarator
	|	rows_declarator
	|	procedure_declarator
	|	reference_declarator
	|	union_declarator
	;

mode_standard
	:	BITS_sithety
	|	_BOOL
		@{ @i @mode_standard.syn_mode@ = mkPrimitiveMode(bool_mode); @}
	|	BYTES_sithety
	|	_CHAR
		@{ @i @mode_standard.syn_mode@ = mkPrimitiveMode(char_mode); @}
	|	COMPL_sithety
	|	INT_sithety
	|	REAL_sithety
	|	_STRING
		@{ @i @mode_standard.syn_mode@ = mkPrimitiveMode(string_mode); @}
	|	_VOID
		@{ @i @mode_standard.syn_mode@ = mkPrimitiveMode(void_mode); @}
	;

INT_sithety
	:	_INT
		@{ @i @INT_sithety.syn_mode@ = mkSithetyMode(int_mode, 0, 0); @}
	|	SHORT_sequence _INT
		@{ @i @INT_sithety.syn_mode@ = mkSithetyMode(int_mode, @SHORT_sequence.num_short@, 0); @}
	|	LONG_sequence _INT
		@{ @i @INT_sithety.syn_mode@ = mkSithetyMode(int_mode, 0, @LONG_sequence.num_long@); @}
	;

SHORT_sequence
	:	_SHORT
		@{ @i @SHORT_sequence.0.num_short@ = 1; @}
	|	_SHORT SHORT_sequence
		@{ @i @SHORT_sequence.0.num_short@ = @SHORT_sequence.1.num_short@ +1; @}
	;

LONG_sequence
	:	_LONG
		@{ @i @LONG_sequence.0.num_long@ = 1; @}
	|	_LONG LONG_sequence
		@{ @i @LONG_sequence.0.num_long@ = @LONG_sequence.1.num_long@ +1; @}
	;

REAL_sithety
	:	_REAL
		@{ @i @REAL_sithety.syn_mode@ = mkSithetyMode(real_mode, 0, 0); @}
	|	SHORT_sequence _REAL
		@{ @i @REAL_sithety.syn_mode@ = mkSithetyMode(real_mode, @SHORT_sequence.num_short@, 0); @}
	|	LONG_sequence _REAL
		@{ @i @REAL_sithety.syn_mode@ = mkSithetyMode(real_mode, 0, @LONG_sequence.num_long@); @}
	;

COMPL_sithety
	:	_COMPL
		@{ @i @COMPL_sithety.syn_mode@ = mkSithetyMode(compl_mode, 0, 0); @}
	|	SHORT_sequence _COMPL
		@{ @i @COMPL_sithety.syn_mode@ = mkSithetyMode(compl_mode, @SHORT_sequence.num_short@, 0); @}
	|	LONG_sequence _COMPL
		@{ @i @COMPL_sithety.syn_mode@ = mkSithetyMode(compl_mode, 0, @LONG_sequence.num_long@); @}
	;

BITS_sithety
	:	_BITS
		@{ @i @BITS_sithety.syn_mode@ = mkSithetyMode(bits_mode, 0, 0); @}
	|	SHORT_sequence _BITS
		@{ @i @BITS_sithety.syn_mode@ = mkSithetyMode(bits_mode, @SHORT_sequence.num_short@, 0); @}
	|	LONG_sequence _BITS
		@{ @i @BITS_sithety.syn_mode@ = mkSithetyMode(bits_mode, 0, @LONG_sequence.num_long@); @}
	;

BYTES_sithety
	:	_BYTES
		@{ @i @BYTES_sithety.syn_mode@ = mkSithetyMode(bytes_mode, 0, 0); @}
	|	SHORT_sequence _BYTES
		@{ @i @BYTES_sithety.syn_mode@ = mkSithetyMode(bytes_mode, @SHORT_sequence.num_short@, 0); @}
	|	LONG_sequence _BYTES
		@{ @i @BYTES_sithety.syn_mode@ = mkSithetyMode(bytes_mode, 0, @LONG_sequence.num_long@); @}
	;

structure_declarator
	:	_STRUCT '(' portrayer ')'
		@{ @i @structure_declarator.syn_mode@ = mkStructureMode(@portrayer.set@);
		   @i @portrayer.set@ =
			(field **)checked_calloc(@portrayer.ct@, sizeof(field *));
		   @i @portrayer.size@ = @portrayer.ct@;
		@}
	;

portrayer
	:	declarer joined_fields
		@{ @i @portrayer.0.ct@ = 1;
		   @m  portrayer.0.done ;
			@portrayer.0.set@[@portrayer.0.size@ - 1] = @joined_fields.set@;
		   @i @joined_fields.inh_mode@ = @declarer.syn_mode@;
		   @i @joined_fields.set@ =
			(field *)checked_calloc(@joined_fields.ct@, sizeof(field));
		   @i @joined_fields.size@ = @joined_fields.ct@;
		@}
	|	declarer joined_fields ',' portrayer /* this rule requires LR(2) */
		@{ @i @portrayer.0.ct@ = @portrayer.1.ct@ + 1;
		   @m  portrayer.0.done ;
			@portrayer.0.set@[@portrayer.0.size@ - @portrayer.1.ct@ - 1] =
				@joined_fields.set@;
		   @i @joined_fields.inh_mode@ = @declarer.syn_mode@;
		   @i @joined_fields.set@ =
			(field *)checked_calloc(@joined_fields.ct@, sizeof(field));
		   @i @joined_fields.size@ = @joined_fields.ct@;
		@}
	;

joined_fields
	:	_PLAIN_TAG
		@{ @i @joined_fields.0.ct@ = 1;
		   @m  joined_fields.0.done ;
		       { field new_field;
			 new_field.name = @_PLAIN_TAG.name@;
			 new_field.mode = @joined_fields.inh_mode@;
		         s_ident_name(@_PLAIN_TAG.name@);
			 s_mode (new_field.name, new_field.mode);
			 @joined_fields.0.set@[@joined_fields.0.size@ - 1] = new_field;
		       }
		@}
	|	_PLAIN_TAG ',' joined_fields
		@{ @i @joined_fields.0.ct@ = @joined_fields.1.ct@ + 1;
		   @m  joined_fields.0.done ;
		       { field new_field;
			 new_field.name = @_PLAIN_TAG.name@;
			 new_field.mode = @joined_fields.0.inh_mode@;
		         s_ident_name(@_PLAIN_TAG.name@);
			 s_mode (new_field.name, new_field.mode);
			 @joined_fields.0.set@[@joined_fields.0.size@ - @joined_fields.1.ct@ - 1] = new_field;
		       }
		   @i @joined_fields.1.inh_mode@  = @joined_fields.0.inh_mode@;
		@}
	;

rows_declarator
	:	'[' rower_list ']' declarer
		@{ @i @rows_declarator.syn_mode@ =
			mkArrayMode(NULL /* dummy rower for now */,@declarer.syn_mode@);
		@}
	;

rower_list
	:	rower
	|	rower ',' rower_list
	;

rower	:	unit ':' unit
		@{ @i @unit.0.inh_mode@ = mkSithetyMode(int_mode, 0, 0);
		   @i @unit.0.coerce@   = false;
		   @i @unit.1.inh_mode@ = mkSithetyMode(int_mode, 0, 0);
		   @i @unit.1.coerce@   = false;
		@}
	|	unit
		@{ @i @unit.0.inh_mode@ = mkSithetyMode(int_mode, 0, 0);
		   @i @unit.0.coerce@   = false;
		@}
	|	':'
	|	/* empty */
	;

procedure_declarator
	:	_PROC formal_procedure_plan
		@{ @i @procedure_declarator.syn_mode@ = @formal_procedure_plan.syn_mode@; @}
	;

formal_procedure_plan
	:	declarer
		@{ @i @formal_procedure_plan.syn_mode@ = mkProcMode(NULL, @declarer.syn_mode@); @}
	|	'(' joined_declarer ')' declarer
		@{ @i @formal_procedure_plan.syn_mode@ = mkProcMode(@joined_declarer.set@,@declarer.syn_mode@);
		   @i @joined_declarer.set@ =
			(mode_defn *)checked_calloc(@joined_declarer.ct@, sizeof(mode_defn));
		   @i @joined_declarer.size@ = @joined_declarer.ct@;
		@}
	;

joined_declarer
	:	declarer
		@{ @i @joined_declarer.0.ct@ = 1;
		   @m joined_declarer.0.done ;
		      @joined_declarer.0.set@[@joined_declarer.0.size@ - 1] =
				@declarer.syn_mode@;
		@}
	|	declarer ',' joined_declarer
		@{ @i @joined_declarer.0.ct@ = @joined_declarer.1.ct@ + 1;
		   @m joined_declarer.0.done ;
		      @joined_declarer.0.set@[@joined_declarer.0.size@ - @joined_declarer.1.ct@ -1] =
				@declarer.syn_mode@;
		@}
	;

reference_declarator
	:	_REF declarer
		@{ @i @reference_declarator.syn_mode@ = mkRefMode(@declarer.syn_mode@); @}
	;

union_declarator
	:	_UNION '(' joined_declarer ')'
		@{ @i @union_declarator.syn_mode@ = mkUnitedMode(@joined_declarer.set@);
		   @i @joined_declarer.set@ =
			(mode_defn *)checked_calloc(@joined_declarer.ct@, sizeof(mode_defn));
		   @i @joined_declarer.size@ = @joined_declarer.ct@;
		@}
	;

unit	:	assignation
	|	formula
	;

assignation
	:	_PLAIN_TAG _BECOMES unit
		@{ @i @assignation.syn_mode@ = gMode(@_PLAIN_TAG.name@);
		   @i @unit.inh_mode@        = gMode(@_PLAIN_TAG.name@);
		   @i @unit.coerce@          = (@unit.inh_mode@.tag != @unit.syn_mode@.tag);
		@}
	|	generator _BECOMES unit
		@{ @i @assignation.syn_mode@ = @generator.syn_mode@;
		   @i @unit.inh_mode@        = @generator.syn_mode@;
		   @i @unit.coerce@          = (@unit.inh_mode@.tag != @unit.syn_mode@.tag);
		@}
	;

formula	:	formula '+' formula
		@{ @i @formula.0.syn_mode@ = balanceMode(@formula.1.syn_mode@, @formula.2.syn_mode@);
		   @i @formula.1.coerce@   = (@formula.0.coerce@ | (@formula.1.inh_mode@.tag != @formula.1.syn_mode@.tag));
		   @i @formula.2.coerce@   = (@formula.0.coerce@ | (@formula.2.inh_mode@.tag != @formula.2.syn_mode@.tag));
		@}
	|	formula '-' formula
		@{ @i @formula.0.syn_mode@ = balanceMode(@formula.1.syn_mode@, @formula.2.syn_mode@);
		   @i @formula.1.coerce@   = (@formula.0.coerce@ | (@formula.1.inh_mode@.tag != @formula.1.syn_mode@.tag));
		   @i @formula.2.coerce@   = (@formula.0.coerce@ | (@formula.2.inh_mode@.tag != @formula.2.syn_mode@.tag));
		@}
	|	formula '>' formula
		@{ @i @formula.0.syn_mode@ = mkPrimitiveMode(bool_mode);
		   @i @formula.1.coerce@   = (@formula.0.coerce@ | (@formula.1.inh_mode@.tag != @formula.1.syn_mode@.tag));
		   @i @formula.2.coerce@   = (@formula.0.coerce@ | (@formula.2.inh_mode@.tag != @formula.2.syn_mode@.tag));
		@}
	|	formula '<' formula
		@{ @i @formula.0.syn_mode@ = mkPrimitiveMode(bool_mode);
		   @i @formula.1.coerce@   = (@formula.0.coerce@ | (@formula.1.inh_mode@.tag != @formula.1.syn_mode@.tag));
		   @i @formula.2.coerce@   = (@formula.0.coerce@ | (@formula.2.inh_mode@.tag != @formula.2.syn_mode@.tag));
		@}
	|	formula '/' formula
		@{ @i @formula.0.syn_mode@ = balanceMode(@formula.1.syn_mode@, @formula.2.syn_mode@);
		   @i @formula.1.coerce@   = (@formula.0.coerce@ | (@formula.1.inh_mode@.tag != @formula.1.syn_mode@.tag));
		   @i @formula.2.coerce@   = (@formula.0.coerce@ | (@formula.2.inh_mode@.tag != @formula.2.syn_mode@.tag));
		@}
	|	formula '*' formula
		@{ @i @formula.0.syn_mode@ = balanceMode(@formula.1.syn_mode@, @formula.2.syn_mode@);
		   @i @formula.1.coerce@   = (@formula.0.coerce@ | (@formula.1.inh_mode@.tag != @formula.1.syn_mode@.tag));
		   @i @formula.2.coerce@   = (@formula.0.coerce@ | (@formula.2.inh_mode@.tag != @formula.2.syn_mode@.tag));
		@}
	|	'+' formula %prec UMINUS
		@{ @i @formula.0.syn_mode@ = @formula.1.syn_mode@;
		@}
	|	'-' formula %prec UMINUS
		@{ @i @formula.0.syn_mode@ = @formula.1.syn_mode@;
		@}
	|	'(' formula ')'
		@{ @i @formula.0.syn_mode@ = @formula.1.syn_mode@;
		@}
	|	denotation
		@{ @i @formula.0.syn_mode@ = @denotation.syn_mode@; @}
	|	_PLAIN_TAG
		@{ @i @formula.0.syn_mode@ = gMode(@_PLAIN_TAG.name@); @}
	;

routine_text
	:	'(' declaration ')' declarer ':' unit
		@{ @i @routine_text.syn_mode@ = @declarer.syn_mode@;
		   @i @unit.inh_mode@         = @declarer.syn_mode@;
		   @i @unit.coerce@           = (@unit.inh_mode@.tag != @unit.syn_mode@.tag);
		@}
	|	declarer ':' unit
		@{ @i @routine_text.syn_mode@ = @declarer.syn_mode@;
		   @i @unit.inh_mode@         = @declarer.syn_mode@;
		   @i @unit.coerce@           = (@unit.inh_mode@.tag != @unit.syn_mode@.tag);
		@}
	;

generator
	:	_LOC declarer
		@{ @i @generator.syn_mode@ = @declarer.syn_mode@; @}
	|	_HEAP declarer
		@{ @i @generator.syn_mode@ = @declarer.syn_mode@; @}
	;

denotation
	:	INTEGRAL_DENOTATION_sithety
		@{ @i @denotation.lit@      = @INTEGRAL_DENOTATION_sithety.lit@;
		   @i @denotation.syn_mode@ = @INTEGRAL_DENOTATION_sithety.syn_mode@;
		@}
	|	REAL_DENOTATION_sithety
		@{ @i @denotation.lit@      = @REAL_DENOTATION_sithety.lit@;
		   @i @denotation.syn_mode@ = @REAL_DENOTATION_sithety.syn_mode@;
		@}
	|	boolean_denotation
		@{ @i @denotation.lit@      = @boolean_denotation.lit@;
		   @i @denotation.syn_mode@ = @boolean_denotation.syn_mode@;
		@}
	|	BITS_DENOTATION_sithety
		@{ @i @denotation.lit@      = @BITS_DENOTATION_sithety.lit@;
		   @i @denotation.syn_mode@ = @BITS_DENOTATION_sithety.syn_mode@;
		@}
	|	_STRING_DENOTATION
		@{ @i @denotation.lit@      = @_STRING_DENOTATION.lit@;
		   @i @denotation.syn_mode@ = mkPrimitiveMode(string_mode);
		@}
	;

INTEGRAL_DENOTATION_sithety
	:	_INTEGRAL_DENOTATION
		@{ @i @INTEGRAL_DENOTATION_sithety.lit@      = @_INTEGRAL_DENOTATION.lit@;
		   @i @INTEGRAL_DENOTATION_sithety.syn_mode@ = mkSithetyMode(int_mode, 0, 0);
		@}
	|	SHORT_sequence _INTEGRAL_DENOTATION
		@{ @i @INTEGRAL_DENOTATION_sithety.lit@      = @_INTEGRAL_DENOTATION.lit@;
		   @i @INTEGRAL_DENOTATION_sithety.syn_mode@ = mkSithetyMode(int_mode, @SHORT_sequence.num_short@, 0);
		@}
	|	LONG_sequence _INTEGRAL_DENOTATION
		@{ @i @INTEGRAL_DENOTATION_sithety.lit@      = @_INTEGRAL_DENOTATION.lit@;
		   @i @INTEGRAL_DENOTATION_sithety.syn_mode@ = mkSithetyMode(int_mode, 0, @LONG_sequence.num_long@);
		@}
	;

REAL_DENOTATION_sithety
	:	_REAL_DENOTATION
		@{ @i @REAL_DENOTATION_sithety.lit@      = @_REAL_DENOTATION.lit@;
		   @i @REAL_DENOTATION_sithety.syn_mode@ = mkSithetyMode(real_mode, 0, 0);
		@}
	|	SHORT_sequence _REAL_DENOTATION
		@{ @i @REAL_DENOTATION_sithety.lit@      = @_REAL_DENOTATION.lit@;
		   @i @REAL_DENOTATION_sithety.syn_mode@ = mkSithetyMode(real_mode, @SHORT_sequence.num_short@, 0);
		@}
	|	LONG_sequence _REAL_DENOTATION
		@{ @i @REAL_DENOTATION_sithety.lit@      = @_REAL_DENOTATION.lit@;
		   @i @REAL_DENOTATION_sithety.syn_mode@ = mkSithetyMode(real_mode, 0, @LONG_sequence.num_long@);
		@}
	;

boolean_denotation
	:	_TRUE
		@{ @i s_bool_lit((@boolean_denotation.lit@ = add_lit("TRUE")), "TRUE");
		   @i @boolean_denotation.syn_mode@        = mkPrimitiveMode(bool_mode);
		@}
	|	_FALSE
		@{ @i s_bool_lit((@boolean_denotation.lit@ = add_lit("FALSE")), "FALSE");
		   @i @boolean_denotation.syn_mode@        = mkPrimitiveMode(bool_mode);
		@}
	;

BITS_DENOTATION_sithety
	:	_BITS_DENOTATION
		@{ @i @BITS_DENOTATION_sithety.lit@      = @_BITS_DENOTATION.lit@;
		   @i @BITS_DENOTATION_sithety.syn_mode@ = mkSithetyMode(bits_mode, 0, 0);
		@}
	|	SHORT_sequence _BITS_DENOTATION
		@{ @i @BITS_DENOTATION_sithety.lit@      = @_BITS_DENOTATION.lit@;
		   @i @BITS_DENOTATION_sithety.syn_mode@ = mkSithetyMode(bits_mode, @SHORT_sequence.num_short@, 0);
		@}
	|	LONG_sequence _BITS_DENOTATION
		@{ @i @BITS_DENOTATION_sithety.lit@  = @_BITS_DENOTATION.lit@;
		   @i @BITS_DENOTATION_sithety.syn_mode@ = mkSithetyMode(bits_mode, 0, @LONG_sequence.num_long@);
		@}
	;

%%

#include "name.h"

void yyerror(const char *s)
{
  fprintf( stderr, "%s near %d\n", s, yylineno );
}

int
main (int argc, char **argv)
{
  extern FILE *yyin;

  for (int i=1; argv[i]; i++) {
    if      ( strcmp ( argv[i], "--dump-lit"     ) == 0 ) dump_literals = true;
    else if ( strcmp ( argv[i], "--dump-name"    ) == 0 ) dump_names = true;
    else if ( strcmp ( argv[i], "--dump-tables"  ) == 0 ) dump_literals = dump_names = true;
    else yyin = fopen (argv[i], "r");
  }

  yyparse();

  /* dump out the literal table */
  if ( dump_literals ) dump_lit ();
  /* dump out the name table */
  if ( dump_names ) dump_name ();
}
