#if defined (__cplusplus)
#include <cstring>
#include <cstdio>
#include <cstdlib>
#else
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#endif

#include "types.h"
#include "name.h"
#include "util.h"
#include "symbol.h"

static S_table name_set;
static bool name_set_initialized = false;
static void name_set_init ( void )
{ name_set = S_empty (); /* Make a new table */
  name_set_initialized = true; }

static name_defn in_name_set ( char *str )
{ if ( !name_set_initialized ) name_set_init();

  /* HACK: this doesn't deal with nested ranges */

  /* return binding of "str", or NULL if "str" is unbound. */
  return (name_defn) S_look ( name_set, S_Symbol ( toString ( str ) ) );
}

name_defn add_name ( char *str )
{ name_defn name;

  /* get binding of "str", or NULL if "str" is unbound. */
  if ( ( name = in_name_set ( str ) ) ) return name;

  /* HACK: this doesn't deal with nested ranges */
  /* really need to maintain nested range structure in indicator table */

  /* Enter a binding "sym->value" into "t", shadowing but not deleting
   *    any previous binding of "sym". */
  name = (name_defn) checked_malloc ( sizeof (struct _name_defn) );
  name->symb = S_Symbol ( toString ( str ) );
  name->kind = undef_name;
  S_enter ( name_set, name->symb, (void *) name );

  return name;
}

string g_name ( name_defn name )
{ return S_name ( name->symb ); }

void s_oper_name ( name_defn op )
{ ((name_defn) S_look ( name_set, op->symb ))->kind = oper_name; }

void s_mode_name ( name_defn md )
{ ((name_defn) S_look ( name_set, md->symb ))->kind = mode_name; }

void s_ident_name ( name_defn id )
{ ((name_defn) S_look ( name_set, id->symb ))->kind = ident_name; }

bool is_oper_name ( name_defn op )
{
  if ( op->kind == oper_name ) return true;
  return false;
}

bool is_mode_name ( name_defn md )
{
  if ( md->kind == mode_name ) return true;
  return false;
}

void s_mode ( name_defn name, mode_defn mode )
{
  ((name_defn) S_look ( name_set, name->symb ))->mode = mode;
}

mode_defn g_mode ( name_defn name )
{
  return ((name_defn) S_look ( name_set, name->symb ))->mode;
}

static void show_name ( S_symbol sym, name_defn binding )
{
  fprintf ( stderr, "Key: [%s] \n", S_name ( sym ) );
}

void dump_name ( void )
{
  if ( !name_set_initialized ) name_set_init();
  fprintf ( stderr, "\n\nIndicator Table:\n" );
  S_dump ( name_set, (void (*)(S_symbol, void *)) show_name );
}
