#ifndef __types_h
#define __types_h

#if defined (__cplusplus)
#include <cstdbool>
#else
#include <stdbool.h>
#endif

typedef char * string;

/* indicate the end of the line - required if backing up across a \n */
#define EOL -666

/* specific attribute grammar attribute types */

/* modes */
enum mode {
  bool_mode, char_mode, string_mode, void_mode,           /* simple modes */
  int_mode, real_mode, compl_mode, bits_mode, bytes_mode, /* sithety modes */
  struct_mode, union_mode, array_mode, ref_mode,          /* aggregate modes */
  proc_mode
};

typedef struct simple_mode_defn {
  enum mode tag;
} simple_mode_defn;

typedef struct sithety_mode_defn {
  enum mode tag;
  struct {int num_short; int num_long;} size;
} sithety_mode_defn;

typedef int declarer_list; /* temporary dummy type */
typedef int rower_list; /* temporary dummy type */

union mode_defn;
typedef struct united_mode_defn {
  enum mode tag;
  union mode_defn *decl_list;
} united_mode_defn;

struct field;
typedef struct struct_mode_defn {
  enum mode tag;
  struct field **field_list;
} struct_mode_defn;

typedef struct array_mode_defn {
  enum mode tag;
  rower_list *rower_list;
} array_mode_defn;

typedef struct ref_mode_defn {
  enum mode tag;
} ref_mode_defn;

typedef struct proc_mode_defn {
  enum mode tag;
  union mode_defn *params;
} proc_mode_defn;

typedef union mode_defn {
  enum mode            tag;
  simple_mode_defn    _simple;
  sithety_mode_defn   _sithety;
  struct_mode_defn    _struct;
  united_mode_defn    _united;
  array_mode_defn     _array;
  ref_mode_defn       _ref;
  proc_mode_defn      _proc;
} mode_defn;

#include "name.h"

/* declared symbols */
typedef struct
{ name_defn name;
  mode_defn mode;
} decl;

typedef struct
{ decl *set;
  int   size;
} declSet;

/* struct fields */
typedef struct field
{ name_defn name;
  mode_defn mode;
} field;

typedef struct
{ field **set;
  int     size;
} fieldSet;

#endif
