/* $Id: agc.Y 2625 2022-08-02 02:07:14Z teshields $ */ 

/* The information in this file is the intellectual property of 
 * Kurt M. Bischoff.  It may not be used, copied, or transferred 
 * without permission from the author. 
 *
 * Copyright (c) 1992, Kurt M. Bischoff.  All rights reserved.
 *
 * Beta Test Release--not for general distribution.
 */ 

%token ID  
%token IDCOLON  /* identifier (but not literal) followed by colon */   
%token NUM  /* [0-9]+ */ 

	/* reserved words:  %type => TYPE,  %left => LEFT, etc.  */  
%token LEFT RIGHT NONASSOC TOKEN PREC TYPE START UNION 
%token MARK MARK2  /* the %% mark */  
%token LCURL  /* the %{ mark */  

	/* tokens of (yox - yacc) */ 
%token STRUCT UNIONC ENUM ATTRIBUTES LBRACE RBRACE ATMACRO ATEND
%token ATTRAV PREORDER POSTORDER LEFTTORIGHT RIGHTTOLEFT DISABLE
%token TRAVID
%token ATI ATCAPI ATM ATE ATID 
%token PRUNE DECORATE 
%token CHAR SHORT INT LONG SIGNED UNSIGNED FLOAT DOUBLE VOID 
%token EOSF RETURN  

%start spec   

%{ 
#include <stdio.h> 
#include "glob.h" 
#include "opts.h" 
#include "trav.h"
#if YYBTYACC && !YYBYACC
/* the `btyacc' 3.0 skeleton omits the following */
#include "parser.h"
#endif

extern void initMacroDef(char *);
extern void terminateMacroDef(void);
extern void installMacroFrmPrm(char *);
extern void bufferAnAttrbID(char *); 
extern void afterAttrbDeclList(void);
extern void setupDumNodeSym(void);
extern void prepareDataStructures1(void); 
extern void prepareDataStructures2(void); 
extern void prepareDataStructuresMD(void); 
extern void prepareDataStructuresMU(void); 
extern void openIOfilesAndSetOptions(int,char *[]); 
extern void closeIOfiles(void);
extern void openTempFiles(void);
extern void rmTempFiles(void);
extern void erasf(const char *);
extern void flushLAbuffer(int,int);
extern void genImmCrefForAttrbRef(int,int,int,FILE *);
extern void checkAttrbsSynOrInh(void); 
extern void doPruneCondChecks(void); 
extern void printIntNodeGenInvocation(void); 
extern void installLeafRefCountInitList(void);
extern void conveyLocalDepGraphInfo(void); 
extern void ldgTopSort(void);
extern int memberDepend(int,int,int,int); 
extern void processAdepSpec(void); 
extern void organizeTables(void); 
extern void scanInit4(void);
extern void initLDG(void); 
extern void printSASheader(void); 
extern void printSAStrailer(void); 
extern void printTravCode(void); 
extern void genTravCrefForAttrbRef(int,int,FILE *); 
extern int someTravEnabled(void);
extern void putSymbolInProdBuffer2(int);
extern int installStartSym(char *);
extern void initForAttrbDeclList(void);
extern int installGsymAsTokenIfNew(char *);
extern int installGsymAsNontermIfNew(char *);
extern int installGsymAfterAttrbDeclList(char *);
extern void incProdNumSeq(void);


extern int suppressPruning;
extern char *startSymLexeme; 
extern int bufPosInProd,bufAttrbNum; 
extern int prodNumSeq; 
extern FILE *yyin; 
extern int *plineno; 
extern int line,nLbraces;
extern char *IDbuffer; 
extern char *gsBuffer; 
extern char *dumNodeSym; 

extern int pruneRequested; 
extern int decorationRequested; 
extern int errorProd; 
extern int retObjectSeen; 

extern int yylex(void);

#if YYBTYACC && !YYBYACC
extern void yyerror(const char *, ...);
#else
extern void yyerror(const char *);
#endif
%} 

%% 
spec 	: 	specG 	
		EOSF 
		specLseq 
		{organizeTables(); 
		} 
	; 

specG	:	defs 	MARK 	
		{setupDumNodeSym(); 
		 printSASheader();
		 suppressPruning |= someTravEnabled();
		} 
		rules  	
		{printSAStrailer(); 
		} 
		root_attrb_ref_section 
		{fprintf(temp1Out,"} /* yyyExecuteRRsection */ \n"); 
	         printTravCode(); 
		}  
		tail     
		{flushLAbuffer(LA1,outG);} 
	; 

tail 	:       {fprintf(yoxOutG,"\n%%%%\n");} 
	|  	MARK  
	;  
 
root_attrb_ref_section  
	: 	
	; 

defs	:	
	|	defs def 
	; 

def 	:	START ID {installStartSym(startSymLexeme);}  
	|	UNION {line = *plineno;}      
 	|	LCURL {line = *plineno;}    
	|	rword tag nlistT   
	|	TYPE tag nlist   
	|	ATTRIBUTES 
		{initForAttrbDeclList();} 
		'{' attribute_decl_list '}' 
                {afterAttrbDeclList();} 
              	grammar_symbol_list 
	|	ATMACRO	ID	
		{initMacroDef(IDbuffer);}
		'('	macroFrmPrmList	')'	
		ATEND	
		{terminateMacroDef();}
	|	ATTRAV 	
		{clearTravBuf();}
		travSpecSeq	
		travIDseq	
	; 

rword	: 	TOKEN
	|	LEFT 
	| 	RIGHT 
	|	NONASSOC
	; 
	
tag 	:	 
	|	'<'  ID  '>'  
	; 

nlist 	: 	ID 
	| 	nlist ID 
	|	nlist  ','  ID 
	; 

nlistT 	: 	nmnoT 
	| 	nlistT nmnoT 
	|	nlistT  ','  nmnoT 
	; 

nmnoT 	:	ID {installGsymAsTokenIfNew(IDbuffer);} 
                /* yacc will catch the error later */     
	|	NUM 
	; 

attribute_decl_list  
	: 
	| 	attribute_decl_list attribute_decl   
	; 

attribute_decl  
	:	type_specifier id_seq stars ID 
		{bufferAnAttrbID(IDbuffer);} 
		field_length_spec 
		';' 
	|	VOID id_seq2 ID  
		{bufferAnAttrbID(IDbuffer);} ';' 
	; 

type_specifier
	: STRUCT ID   
	| UNIONC ID 
	| ENUM ID
	| ID  	
	| type_keyword_list type_keyword  
	; 

type_keyword_list 
	: 
	|	type_keyword_list type_keyword 
	; 

type_keyword 
	: CHAR | SHORT | INT | LONG | FLOAT | DOUBLE | SIGNED | UNSIGNED 
	;

id_seq 	:	
	|	id_seq stars ID {bufferAnAttrbID(IDbuffer);} ','  
	; 

id_seq2	:	
	|	id_seq2 ID {bufferAnAttrbID(IDbuffer);} ','  
	; 

stars 	: 
	| 	stars '*'  
	;

field_length_spec 
	:	
	|	':' NUM 
	;  
	
grammar_symbol_list 
	:	
	|	grammar_symbol_list ID 
		{installGsymAfterAttrbDeclList(IDbuffer);}  
	;

macroFrmPrmList
	:	
	|	macroFrmPrmList	ID	','
		{installMacroFrmPrm(IDbuffer);}
	;

travSpecSeq
	:	
	|	travSpecSeq	travSpec
	;

travIDseq
	:	
	|	travIDseq	ID	
		{newTrav(IDbuffer);}
	;

travSpec
	:	PREORDER	{travBuf.PRE = 1;}
	|	POSTORDER	{travBuf.POST = 1;}
	|	LEFTTORIGHT	{travBuf.LR = 1;}
	|	RIGHTTOLEFT	{travBuf.RL = 1;}
	|	DISABLE		{travBuf.DA = 1;}
	;
	
rules 	:	IDCOLON 
		{prodBufferPointer2 = 0; errorProd = 0; incProdNumSeq(); 
                 putSymbolInProdBuffer2(installGsymAsNontermIfNew(IDbuffer)); 
                } 
		rbody attrb_def_section  
	|	rules rule  
	; 

rule 	:	IDCOLON  
		{prodBufferPointer2 = 0; errorProd = 0; incProdNumSeq();  
                 putSymbolInProdBuffer2(installGsymAsNontermIfNew(IDbuffer)); 
                } 
 		rbody attrb_def_section  
	| 	'|'  
		{prodBufferPointer2 = 1; errorProd = 0; incProdNumSeq();  
		} 
		rbody attrb_def_section  
	; 

rbody 	: 
	|	rbody ID 
		{if (IDbuffer[0] != '\'')  
                    putSymbolInProdBuffer2(installGsymAsNontermIfNew(IDbuffer));
                    else 
                    putSymbolInProdBuffer2(installGsymAsTokenIfNew(IDbuffer));  
                }  
	|	rbody PREC ID   
	|	rbody act  
	; 

act  	: 	{line = *plineno; nLbraces = 1;} 
		'{' 
		'}'         
	; 

attrb_def_section 
	: 	{initLDG();} 
		a_d_s_body 
		{/*showDepGraph(stderr);*/ 
                 checkAttrbsSynOrInh();
                 doPruneCondChecks(); 
		 if (errorProd) fprintf(yoxOutG,"{yyyYok=0;if(0){");  
		    else        fprintf(yoxOutG,"{if(yyyYok){");  
		 printIntNodeGenInvocation();    
		 ldgTopSort();  
                 outputTravCodePerProd();
		 conveyLocalDepGraphInfo();  
		 fprintf(yoxOutG,"}}");   
   		 flushLAbuffer(LA1,outG);  
		}   
	|	attrb_def_section ';' 
	; 

a_d_s_body 
	: 	
	|	LBRACE 
		{pruneRequested = 0; 
		 decorationRequested = 0;
                } 
		nodeCmdList 
		attrb_def_list 
		RBRACE   
	;

nodeCmdList 
	: 
	|	nodeCmd nodeCmdList 
	; 

nodeCmd 
	:	PRUNE 	{pruneRequested = 1;}  
	|	DECORATE  {decorationRequested = 1;}  
	; 

attrb_def_list 
	:	
	|	attrb_def_list trav_segm 
	|	attrb_def_list attrb_def {processAdepSpec();}  
	; 

attrb_def:	ATI ID 
		{memberDepend(bufPosInProd,bufAttrbNum,1,1); 
		} 
		im_dependee_ref_seq 

	|	ATCAPI ID 
		{memberDepend(bufPosInProd,bufAttrbNum,1,1); 
		} 
		mm_dependee_ref_seq 

	|	ATM dependent_name_seq ';' mm_dependee_ref_seq 

	|	ATE dependent_name_seq ':' dependee_name_seq ';' ref_seq   

	; 

im_dependee_ref_seq  	/* im = implicit mode */
	:	
	|	im_dependee_ref_seq ID 
		{if (!memberDepend(bufPosInProd,bufAttrbNum,1,0))
		    memberDepend(bufPosInProd,bufAttrbNum,0,1);
		    else 
		    ywarning("implicit dependent depends on itself\n"); 
		} 
	;
 
mm_dependee_ref_seq 	/* mm = mixed mode */ 
	:	
	|	mm_dependee_ref_seq ID 
		{if (!memberDepend(bufPosInProd,bufAttrbNum,1,0))
		    memberDepend(bufPosInProd,bufAttrbNum,0,1);
		} 
	;
 
dependent_name_seq  
	:	ID 	
		{memberDepend(bufPosInProd,bufAttrbNum,1,1);} 
	|	dependent_name_seq ID 
		{if (memberDepend(bufPosInProd,bufAttrbNum,1,1)) 
                    yerror("duplicate dependent\n"); 
		} 
	; 
 
dependee_name_seq  
	:	
	|	dependee_name_seq ID 
		{if (memberDepend(bufPosInProd,bufAttrbNum,1,0)) 
                    yerror("attribute depends immediately on itself\n"); 
		    else 
		    {if (memberDepend(bufPosInProd,bufAttrbNum,0,1)) 
                        yerror("duplicate dependee\n"); 
                    } 
		} 
	; 
 
ref_seq : 
	|	ref_seq ID  
		{if (!memberDepend(bufPosInProd,bufAttrbNum,0,0) &&  
		     !memberDepend(bufPosInProd,bufAttrbNum,1,0) 
		    ) 
    yerror("in explicit mode, referred to attribute not named explicitly\n"); 
		}
	; 

trav_segm
	:	ATID
	|	ATID	'('	')'
	|	ATID	'('	')' 	'('	')'
	;

specLseq:	specL
	|	specL	EOSF	specLseq
	; 

specL 	:	MARK Lrules
	|	MARK Lrules MARK2
	|	     Lrules
	; 

Lrules 	: 	
	|	Lrules 
		RETURN 
		{prodBufferPointer2 = 0;  errorProd = 0;} 
		returnObject 
		';'    
		attrb_def_sectionL 
		{checkAttrbsSynOrInh();
		}
	;

returnObject 
	: 	
	|	ID 
		{putSymbolInProdBuffer2(installGsymAsTokenIfNew(IDbuffer));  
		 installLeafRefCountInitList();  
	 	 initLDG();
    		} 
	;

attrb_def_sectionL 
	:	 
	| 	LBRACE 
		ref_seqL 
		RBRACE 
	; 
	
ref_seqL  
	:	
	|	ref_seqL ID
		{
		 if (memberDepend(bufPosInProd,bufAttrbNum,1,1)) 
                    yerror("attribute multiply defined\n"); 
		 genImmCrefForAttrbRef(bufPosInProd,bufAttrbNum,0,
				       Lfile[currentLfile].outPtr 
				      );  
                 processAdepSpec(); 
		} 
	; 
 
%% 

int yyparse(void);

int main(int argc, char *argv[]) 
{openIOfilesAndSetOptions(argc,argv);  
 if (stripOxConstructs)
    {fprintf(yoxOutG,"%%{\nvoid yyyerror(){}\n");
     fprintf(yoxOutG,"void yyyCheckForResizes(){}\n%%}\n");
     scanInit4(); 
     yylex();
     closeIOfiles(); 
     return(0);
    }
 openTempFiles();  
 prepareDataStructures1(); 
 prepareDataStructures2(); 
 prepareDataStructuresMD(); 
 prepareDataStructuresMU(); 
 yyparse(); 
 closeIOfiles(); 
 rmTempFiles(); 
 if (removeOutput) rmOutputFiles(); 
 if (!buildHeaderFile) erasf(yoxOutFileStrH);   
 if (removeOutput)
    return(-1); 
    else 
    return(0); 
} 
 
/* First choice for determining the type of node to put on the tree at shift 
 * time is to find it in a well-formed object of a return statement.  Second
 * choice is to look at the grammar symbol in the first attribute reference 
 * of the attribute definition section if such section is nonempty.  If all
 * else fails, assume the node is attribute-less.  
 */ 

