/* $Id: depGs.c 2098 2021-02-16 02:47:30Z teshields $ */ 

/* The information in this file is the intellectual property of 
 * Kurt M. Bischoff.  It may not be used, copied, or transferred 
 * without permission from the author. 
 *
 * Copyright (c) 1992, Kurt M. Bischoff.  All rights reserved.
 *
 * Beta Test Release--not for general distribution.
 */ 


#if defined (__cplusplus)
#include <cstdio>
#include <cstdlib>
#include <climits>
#else
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#endif

#include "glob.h" 
#include "opts.h" 
#include "scan.h" 
#include "trav.h"

#define G_DEBUG 1 

extern struct gSym *gSymTable; 
extern int maxProdLength; 
extern int maxNattrbsPerSym; 
extern void clearErrorTrav(void);

extern void translateAndOutputRefSeg(enum instRefType,long *,long,int,int,FILE *); 

extern FILE *temp1Out,*temp2Out,*temp3Out,
            *temp4Out,*temp5Out,*temp6Out, 
            *temp7Out; 

extern int maxNprods; 
extern int prodNumSeq; 

long *indToEntsByProdNum; 
long maxNterms = 1024; 
long *leafRefCountInitIndex;  
long *leafRefCountListLength;  
long eeCountInitSeq = 0;  
long attrbOccrSeq = 0;  
long leafRefCountInitSeq = 0; 
long entOccrSeq = 0; 

int permitCtimeCircTest = 1; 
int maxNentsPerDepSpec = 256; 
int maxNeesPerDepSpec = 256;

/* for keeping track of the dependents in a dependency specification */ 
int dependentSeq = 0; 
/* for keeping track of the dependees in a dependency specification */ 
int dependeeSeq = 0; 


/* At the time I designed this, it seemed that it would be handy to store 
   both the local dependency graph and its inverse.  As I write this 
   comment, I don't know if they are both used.  --KB 
   The comments refer to both graphs as the LDG.  
 */
/* the space for storing lists of dependents in the LDG */ 
struct attrbOccr *dependent; 
/* the space for storing lists of dependees in the LDG */ 
struct attrbOccr *dependee; 



/* Each struct ldgAdjListHeads corresponds to an attribute occurrence in 
   the rule being parsed.  
   The following are predicates on the attribute occurrences of the rule: 
 
   defdThisProd:    is defined in this rule?  i.e., is an output occurrence? 
   hasYaccPvarRef:  definition has a reference to a Yacc pseudovariable? 
   inhFellowEnt:    has a fellow dependent that is inherited? 
   synFellowEnt:    has a fellow dependent that is synthesized? 
   (fellow dependents are possible using the explicit mode annunciator @e)
 */
struct ldgAdjListHeads {unsigned int defdThisProd     : 1;  
                        unsigned int hasYaccPvarRef   : 1;  
                        unsigned int inhFellowEnt     : 1;  
                        unsigned int synFellowEnt     : 1;  
                        long refTextStart,refTextEnd; 
                        long entList;  /* the occurrence's dependents */ 
                        long eeList;   /* the occurrence's dependees  */ 
                        short eeCount; /* the occurrence's dependee count */
                       }; 

int hasYaccPvarRefGlob; 
extern char *refSegBuffer; 
extern long refSegBufferPointer,refSegBufferPointerHold,refSegBufferSize; 

struct ldgAdjListHeads *ldgALhead;  
int maxNattrbOccrsPerProd = 2048;  

/* index of pointers to the first attribute occurrence in each symbol */ 
long *ldgALheadInd; 


 
struct ldgNodeStgType {struct attrbOccr node; 
                       long next; 
                      }; 
 
struct ldgNodeStgType *ldgNodeStg; 
long ldgNodeStgFree;  
int ldgNodeStgSize = 8192;  


/* Used in the consistency and completeness tests. 
   if (change is nonzero) 
       mark the occurrence as defined or undefined (depending on newVal) ;
   issue error message if setting it when already set; 
   return the truth value of "was it defined before calling this?".
 */
int markOrCheckAttrbOccrDefd(int symPos, int attrbPos, int change, int newVal) 
  {unsigned int prev; 
   prev = ldgALhead[ldgALheadInd[symPos]+attrbPos].defdThisProd; 
   if (change) ldgALhead[ldgALheadInd[symPos]+attrbPos].defdThisProd = newVal; 
   if (change && newVal && prev) 
      {yerror("attribute occurrence multiply defined \n   "); 
       fprintf(stderr,"at position %d in production.       %s  %s\n",
               symPos,
               gSymTable[prodBuffer2[symPos].gSymTabInd].lexeme,  
               gSymTable[prodBuffer2[symPos].gSymTabInd].attrbID[attrbPos]   
              );   
      }  
   return (prev);  
  } 



int pruneRequested; 
int decorationRequested; 

struct attrbOccr *ldgTopSortQ;  
long ldgTopSortQfront,ldgTopSortQrear; 
 
void prepareDataStructures2(void) 
  {int i; 
   if ((dependent = (struct attrbOccr *) malloc(maxNentsPerDepSpec * 
                                                sizeof(struct attrbOccr)
                                               )
       )  
       == 
       ((struct attrbOccr *) NULL) 
      )   
      fatal("malloc error in dependent list allocation\n"); 

 
   if ((dependee = (struct attrbOccr *) malloc(maxNeesPerDepSpec * 
                                               sizeof(struct attrbOccr)
                                              )
       )  
       == 
       ((struct attrbOccr *) NULL) 
      )   
      fatal("malloc error in dependee list allocation\n");  


   if ((ldgTopSortQ = (struct attrbOccr *) malloc(maxNattrbOccrsPerProd * 
                                                  sizeof(struct attrbOccr)
                                                 )
       )  
       == 
       ((struct attrbOccr *) NULL) 
      )   
      fatal("malloc error in ldgTopSortQ allocation\n");  


   if ((ldgALhead = (struct ldgAdjListHeads *) malloc(maxNattrbOccrsPerProd * 
                                               sizeof(struct ldgAdjListHeads)
                                                     )
       )  
       == 
       ((struct ldgAdjListHeads *) NULL) 
      )   
      fatal("malloc error in local dep graph adj list head allocation\n");  

   for (i=0;i<maxNattrbOccrsPerProd;i++) 
       {ldgALhead[i].defdThisProd = 0; 
        ldgALhead[i].hasYaccPvarRef = 0; 
        ldgALhead[i].inhFellowEnt = 0; 
        ldgALhead[i].synFellowEnt = 0; 
        ldgALhead[i].refTextStart = 0; 
        ldgALhead[i].refTextEnd = -1; 
        ldgALhead[i].entList = 0;  
        ldgALhead[i].eeList = 0; 
        ldgALhead[i].eeCount = -1; 
       } 

 
   if ((ldgNodeStg = 
       (struct ldgNodeStgType *) malloc(ldgNodeStgSize * 
                                        sizeof(struct ldgNodeStgType)
                                       )
       )  
       == 
       ((struct ldgNodeStgType *) NULL) 
      )   
      fatal("malloc error in local dep graph adj list space allocation\n");  


   if ((ldgALheadInd = (long *) malloc((maxProdLength+1) * sizeof(long)))  
       == 
       ((long *) NULL) 
      )   
      fatal("malloc error in local dep graph head index allocation\n");  


   if ((indToEntsByProdNum = (long *) malloc((maxNprods+1) * sizeof(long)))  
       == 
       ((long *) NULL) 
      )   
      fatal("malloc error in index to dependents by prod num allocation\n");  


   if ((refSegBuffer = (char *) malloc((refSegBufferSize) * sizeof(char)))  
       == 
       ((char *) NULL) 
      )   
      fatal("malloc error in reference segment buffer allocation\n");  


   if ((leafRefCountInitIndex = (long *) malloc((maxNterms+1) * sizeof(long)))  
       == 
       ((long *) NULL) 
      )   
      fatal("malloc error in allocation of leaf ref count init list stg\n");  
   for (i=0;i<maxNterms+1;i++) leafRefCountInitIndex[i] = -1; 


   if ((leafRefCountListLength= (long *) malloc((maxNterms+1) * sizeof(long)))  
       == 
       ((long *) NULL) 
      )   
      fatal("malloc error in allocation of leaf ref count list length stg\n");  
   for (i=0;i<maxNterms+1;i++) leafRefCountListLength[i] = 0; 

   if ((travArr = (struct trav *) malloc((1+maxNtravs) * sizeof(struct trav)))  
       == 
       ((struct trav *) NULL) 
      )   
      fatal("malloc error in allocation of traversal space\n");  
   travArr++;
   clearTravBuf();
   newTrav("undeclared");
   clearErrorTrav();

   if ((travSegBufArr = 
          (struct travSegBuf *) malloc((maxNtravSegBufs) * 
                                        sizeof(struct travSegBuf))
                                      )  
       == 
       ((struct travSegBuf *) NULL) 
      )   
      fatal("malloc error in allocation of traversal segment buffers\n");  
   lambdaTSB = travSegBufArr + maxNtravSegBufs;
   for (i=0; i<maxNtravSegBufs; i++)
     {travSegBufArr[i].orderRevCondStartPtr = 0; 
      travSegBufArr[i].orderRevCondStopPtr = 0; 
      travSegBufArr[i].actionStartPtr = 0; 
      travSegBufArr[i].actionStopPtr = 0; 
     } 
   travSegBufAvail = 0; 
  } 



/* Inserts an occurrence into the image of the dependency expression.
   Searches the dependee or dependent list (as specified by entNOTee)
   and remembers whether the occurrence is already in that list.
   The sorting and binary search are probably overkill, since Ox 
   presently is I/O-bound.
   Inserts the occurrence into the selected list if (insert != 0).
   Returns true if the occurrence was already in the given list.
 */ 
int memberDepend(int symPos, int attrbNum, int entNOTee, int insert) 
  {struct attrbOccr *occrList; 
   int *seq,limit; 
   int i,lower,upper;
   int member=0; 

   if ((symPos == -1) || (attrbNum == -1)) return(0);  
   if (entNOTee) 
      {occrList=dependent; 
       seq = &dependentSeq; 
       limit=maxNentsPerDepSpec;    
       if (insert) markOrCheckAttrbOccrDefd(symPos,attrbNum,1,1); 
      }  
      else    
      {occrList=dependee; 
       seq = &dependeeSeq; 
       limit=maxNeesPerDepSpec;    
      }  
   lower = 0; 
   upper = *seq - 1; 
   while (lower <= upper) 
     {i = (lower+upper)/2; 
      if (symPos == occrList[i].symPos && 
          attrbNum == occrList[i].attrbNum 
         ) 
         {member = 1; break;}   
         else 
         {if (symPos < occrList[i].symPos ||    
              (symPos == occrList[i].symPos && 
               attrbNum < occrList[i].attrbNum 
              ) 
             )   
             lower = i+1; 
             else 
             upper = i-1; 
         }  
     } 
   if (insert && !member) 
      {if (limit == *seq) 
          {if (entNOTee) 
              handleOverflow(NPDS);  
              else  
              handleOverflow(EPDS); 
          }  
       for (i= *seq;i>lower;i--) occrList[i] = occrList[i-1];   
       occrList[lower].symPos = symPos; 
       occrList[lower].attrbNum = attrbNum; 
       (*seq)++; 
      }  
   return(member);  
  } 



/* get ready to start a new local dependency graph */
void initLDG(void) 
  {int i; 
   int ldgALheadIndSeq = 0;  
   for (i=0; i<prodBufferPointer2; i++) 
       {ldgALheadInd[i] = ldgALheadIndSeq; 
        ldgALheadIndSeq += gSymTable[prodBuffer2[i].gSymTabInd].nAttrbs; 
        if (ldgALheadIndSeq >= maxNattrbOccrsPerProd) 
           handleOverflow(AOPP);  
       }  
   ldgALheadInd[prodBufferPointer2] = ldgALheadIndSeq; 
   ldgNodeStgFree = 1; 
   refSegBufferPointerHold = 0;  
   refSegBufferPointer = 0;  
   ldgTopSortQfront = 0; 
   ldgTopSortQrear = -1; 
  } 



/* handle one dependency specification (done once for each @i, @e, or @m).
 */ 
void processAdepSpec(void) 
  {int i,j; 
   struct ldgAdjListHeads *pHead;
   long cdr,*lp; 
   pHead= &(ldgALhead[ldgALheadInd[dependent[0].symPos]+dependent[0].attrbNum]);
   pHead->refTextStart = refSegBufferPointerHold;  
   pHead->refTextEnd = refSegBufferPointer - 1; 
   pHead->hasYaccPvarRef = hasYaccPvarRefGlob; 
   for (i=0;i<dependentSeq;i++) 
       {if (dependent[i].symPos) 
           pHead->inhFellowEnt = 1; 
           else 
           pHead->synFellowEnt = 1; 
       }
   for (i=0;i<dependentSeq;i++) 
       {pHead = 
        &(ldgALhead[ldgALheadInd[dependent[i].symPos]+dependent[i].attrbNum]); 
        if ((pHead->eeCount = dependeeSeq)  == 0) 
           {ldgTopSortQ[++ldgTopSortQrear].symPos = dependent[i].symPos;  
            ldgTopSortQ[ldgTopSortQrear].attrbNum = dependent[i].attrbNum; 
           }  
        for (j=0;j<dependeeSeq;j++) 
            {if (ldgNodeStgFree == ldgNodeStgSize) 
                handleOverflow(LGNS);  
             cdr = pHead->eeList;   
             pHead->eeList = ldgNodeStgFree;   
             ldgNodeStg[pHead->eeList].node = dependee[j]; 
             ldgNodeStg[pHead->eeList].next = cdr; 
             ldgNodeStgFree++;  
            }  
       } 
   for (i=0;i<dependeeSeq;i++) 
       {lp = &(ldgALhead[ldgALheadInd[dependee[i].symPos] + 
                         dependee[i].attrbNum
                        ].entList);   
        while ((*lp != 0) && (ldgNodeStg[*lp].node.symPos == 0)) 
          lp = &(ldgNodeStg[*lp].next); 
        for (j=0;j<dependentSeq;j++) 
             {if (ldgNodeStgFree == ldgNodeStgSize) 
                 handleOverflow(LGNS);  
              cdr = *lp;  
              *lp = ldgNodeStgFree;   
              ldgNodeStg[*lp].node = dependent[j]; 
              ldgNodeStg[*lp].next = cdr; 
              ldgNodeStgFree++; 
             }  
       } 
   dependentSeq = 0;  
   dependeeSeq = 0; 
   hasYaccPvarRefGlob = 0;  
   refSegBufferPointerHold = refSegBufferPointer;  
  } 


 
int startSymMarkedSyn = 0;  

/* consistency and completeness check for one rule */
void checkAttrbsSynOrInh(void) 
  {int i,j,nA,defdInThisProd,inh,RHS;
   struct gSym *pSym; 
   for (i=0; i<prodBufferPointer2; i++) 
       {RHS = i; 
        pSym = &(gSymTable[prodBuffer2[i].gSymTabInd]);   
        nA = pSym->nAttrbs;
        if (!RHS && !startSymMarkedSyn && (pSym->lexeme == startSymLexeme))
           {pSym->fileOfFirstDef = inFileStr;
            pSym->lineOfFirstDef = *plineno;
            for (j=0;j<nA;j++) 
                {pSym->aaList[j].IorSknown = 1; 
                 pSym->aaList[j].InotS = 0; 
                } 
            startSymMarkedSyn = 1; 
           }
        if (pSym->lineOfFirstDef == -1) 
           {pSym->fileOfFirstDef = inFileStr;
            pSym->lineOfFirstDef = *plineno;
            for (j=0;j<nA;j++) 
                {pSym->aaList[j].IorSknown = 1; 
                 defdInThisProd = markOrCheckAttrbOccrDefd(i,j,1,0); 
                 if ((defdInThisProd && RHS) || (!defdInThisProd && !RHS)) 
                    pSym->aaList[j].InotS = 1; 
                    else  
                    pSym->aaList[j].InotS = 0; 
                } 
           }
           else 
           {for (j=0;j<nA;j++) 
                {defdInThisProd = markOrCheckAttrbOccrDefd(i,j,1,0); 
                 inh = pSym->aaList[j].InotS; 
                 if (defdInThisProd) 
                    {if (!RHS && inh)  
                        {yerror("attribute inherited in the rule\n"); 
                         fprintf(stderr,"%s%d of %s.\n%s     %s   %s\n",
                         "  ending near line ",
                         pSym->lineOfFirstDef,  
                         pSym->fileOfFirstDef,  
                         "  synthesized on LHS of this rule. ",
                         pSym->lexeme,
                         pSym->attrbID[j]  
                                ); 
                        } 
                     if (RHS && !inh)  
                        {yerror("attribute synthesized in the rule\n"); 
                         fprintf(stderr,"%s%d of %s.\n%s%d%s     %s   %s\n\n",
                         "  ending near line ",
                         pSym->lineOfFirstDef,  
                         pSym->fileOfFirstDef,  
                         "  inherited at grammar symbol ",
                         i, 
                         " of this rule.", 
                         pSym->lexeme,
                         pSym->attrbID[j]  
                                ); 
                        } 
                    }  
                    else 
                    {if (!RHS && !inh)  
                        {yerror("attribute synthesized in the rule\n"); 
                         fprintf(stderr,"%s%d of %s.\n%s     %s   %s\n\n",
                         "  ending near line ", 
                         pSym->lineOfFirstDef,  
                         pSym->fileOfFirstDef,  
                         "  undefined for LHS in this rule.  ",
                         pSym->lexeme,
                         pSym->attrbID[j]  
                               ); 
                        } 
                     if (RHS && inh)  
                        {yerror("attribute inherited in the rule\n"); 
                         fprintf(stderr,"%s%d of %s.\n%s%d%s     %s   %s\n\n",
                         "  ending near line ", 
                         pSym->lineOfFirstDef,  
                         pSym->fileOfFirstDef,  
                         "  undefined for grammar symbol ",
                         i, 
                         " in this rule.", 
                         pSym->lexeme,
                         pSym->attrbID[j]  
                                ); 
                        } 
                    } /* else */  
                } /* for j */ 
           } /* else */ 
       } /* for i */  
  } 


/* put synthesized attributes of tokens in the ready set (queue) */ 
void enqueueSynAttrbsOfTerms(void)  
  {int i,j; 
   struct gSym *pSym; 
   for (i=1; i<prodBufferPointer2; i++) 
       {pSym = &(gSymTable[prodBuffer2[i].gSymTabInd]);  
        for (j=0;j<pSym->nAttrbs;j++) 
            {if ((!(pSym->aaList[j].InotS)) && (pSym->symNum > 0)) 
                {ldgTopSortQ[++ldgTopSortQrear].symPos = i;  
                 ldgTopSortQ[ldgTopSortQrear].attrbNum = j; 
                }  
            } /* for j */  
       } /* for i */ 
  } 
 


/* the static part of the topological sort. 
   outputs semantic actions into the yacc actions or the big nested 
   switch statement in yyySolveAndSignal().
 */
void ldgTopSort(void) 
  {int i,j,nA;
   long nextAN; 
   struct attrbOccr aoDum;
   struct ldgAdjListHeads *pHead; 
   static long entOccrSeqLoc; 
   long entOccrSeqLocHold; 
   struct gSym *pSym; 
/*
fprintf(stderr,"%d: \n",prodNumSeq-1);
 */
   enqueueSynAttrbsOfTerms(); 
   while (ldgTopSortQfront <= ldgTopSortQrear) 
     {aoDum=ldgTopSortQ[ldgTopSortQfront++]; 
      pHead = &(ldgALhead[ldgALheadInd[aoDum.symPos]+aoDum.attrbNum]); 
/*
fprintf(stderr,"    %s.%s\n",
        gSymTable[prodBuffer2[aoDum.symPos].gSymTabInd].lexeme,
        gSymTable[prodBuffer2[aoDum.symPos].gSymTabInd].attrbID[aoDum.attrbNum]
       );
 */
      if (pHead->refTextStart <= pHead->refTextEnd) 
         {translateAndOutputRefSeg(imm,&(pHead->refTextStart),pHead->refTextEnd,
                                   DONTCARE,DONTCARE, 
                                   yoxOutG 
                                  );
         } 
      nextAN = pHead->entList;  
      while (nextAN != 0) 
        {aoDum = ldgNodeStg[nextAN].node; 
         if (!(--(ldgALhead[ldgALheadInd[aoDum.symPos]+
                            aoDum.attrbNum
                           ].eeCount   
                 )
              ) 
            ) 
            ldgTopSortQ[++ldgTopSortQrear] = aoDum;  
         nextAN = ldgNodeStg[nextAN].next;  
        }
     }

/* separate the part above this comment and the part below into two separate
   routines.  The above is truly the sort. 
*/
 
   fprintf(temp1Out,
           "case %d:  /***%d***/\n  switch (yyyws) {\n",
           prodNumSeq-1,prodNumSeq-1
          ); 
   for (i=0; i<prodBufferPointer2; i++) 
     {pSym = &(gSymTable[prodBuffer2[i].gSymTabInd]);   
      nA = pSym->nAttrbs; 
      if (nA) fprintf(temp1Out,"  case %d:  /**/\n",i); 
      if (nA) fprintf(temp1Out,"    switch (yyywa) {\n"); 
      for (j=0;j<nA;j++) 
        {pHead = &(ldgALhead[ldgALheadInd[i]+j]); 
         entOccrSeqLocHold = entOccrSeqLoc;  
         if (pHead->eeCount != 0)
            {  
             nextAN = pHead->entList; 
             while (nextAN != 0) 
               {entOccrSeqLoc++;  
                nextAN = ldgNodeStg[nextAN].next; 
               } 
            } 
         if (pHead->eeCount > 0) 
            {fprintf(temp1Out,"    case %d:\n",j ); 
             if (pHead->refTextStart <= pHead->refTextEnd) 
                {translateAndOutputRefSeg(decor, &(pHead->refTextStart),
                                          pHead->refTextEnd,
                                          i,j,temp1Out  
                                         ); 
                }
             if (entOccrSeqLocHold < entOccrSeqLoc) 
                fprintf(temp1Out,
                        "    yyySignalEnts(yyyRefN,yyyEntL+%ld,yyyEntL+%ld);\n",
                        entOccrSeqLocHold*2, 
                        entOccrSeqLoc*2 
                       );  
             if (pHead->hasYaccPvarRef) 
                {yerror("decoration-time reference to\n     "); 
                 fprintf(stderr,"Yacc pseudo-variable\n");
                }
             if (pHead->inhFellowEnt && pHead->synFellowEnt) 
                {yerror("attribute reference segment with both\n     "); 
                 fprintf(stderr,"inherited and synthesized dependents\n"); 
                }
             fprintf(temp1Out,"    break;\n"); 
            } 
         pHead->synFellowEnt = 0;  
         pHead->inhFellowEnt = 0;  
         pHead->hasYaccPvarRef = 0; 
        } /* for j */  
      if (nA) fprintf(temp1Out,"    }"); 
      if (nA) fprintf(temp1Out,"\n  break;\n"); 
     } /* for i */ 
   fprintf(temp1Out,"  }\nbreak;\n"); 
  } 
 


#if G_DEBUG 
void showDepGraph(FILE *fp)
  {int i,j; 
   long nextAN; 
   struct attrbOccr aoDum;
   int nA;  
   fprintf(fp,"***************dependency graph for production %u:", 
              prodNumSeq-1 
          );  
   for (i=0; i<prodBufferPointer2; i++) 
       {nA = gSymTable[prodBuffer2[i].gSymTabInd].nAttrbs;   
        for (j=0;j<nA;j++) 
            {fprintf(fp,"\n\ndependees  of %u %u:  ",i,j);  
             nextAN = ldgALhead[ldgALheadInd[i]+j].eeList;  
             while (nextAN != 0) 
               {aoDum = ldgNodeStg[nextAN].node; 
                fprintf(fp,"    %u %u",aoDum.symPos,aoDum.attrbNum);  
                nextAN = ldgNodeStg[nextAN].next;  
               }
             fprintf(fp,"\ndependents of %u %u:  ",i,j);  
             nextAN = ldgALhead[ldgALheadInd[i]+j].entList;  
             while (nextAN != 0) 
               {aoDum = ldgNodeStg[nextAN].node; 
                fprintf(fp,"    %u %u",aoDum.symPos,aoDum.attrbNum);  
                nextAN = ldgNodeStg[nextAN].next;  
               }
            } /* for j */  
       } /* for i */ 
   fprintf(fp,"\n\n\n\n"); 
  } 

void flushAll(void) 
  {int i; 
   fflush(yoxOutG);  
   for (i=0;i<nLfiles;i++) fflush(Lfile[i].outPtr);  
   fflush(yoxOutH);  
   fflush(temp1Out);  
   fflush(temp2Out);  
   fflush(temp3Out);  
   fflush(temp4Out);  
   fflush(temp5Out);  
   fflush(temp6Out);  
   fflush(temp7Out);  
   fflush(tempTravOut);  
  } 
#endif 
 
 

int ldgCDcount; 



/* Returns false if pred is true for some attribute occurrence 
 * in the graph closure of AO.  graph is either dependEE or dependENT. 
 * Returns true  if pred is false for every attribute occurrence 
 * in the graph closure of AO.   
 *
 * In other words, returns true exactly when AO's graph closure 
 * is disjoint from the set of occurrences that satisfy pred. 
 *
 * Used for computing pruning conditions.
 */
int ldgClosureDisjoint(struct attrbOccr AO, int (*pred)(struct attrbOccr), enum ENTorEE graph)
  {long nextAN;

   if (ldgCDcount++ > 200) 
      {ywarning("rule probably has cycle.\n");  
       return 0;
      }
   if ((*pred)(AO)) return 0; 
   nextAN = (graph == EE)?
            ldgALhead[ldgALheadInd[AO.symPos] + AO.attrbNum].eeList:
            ldgALhead[ldgALheadInd[AO.symPos] + AO.attrbNum].entList;
   while (nextAN != 0)  
     {if (!ldgClosureDisjoint(ldgNodeStg[nextAN].node,pred,graph)) 
         return 0;   
      nextAN = ldgNodeStg[nextAN].next;  
     } 
   return 1;
  }



int isLHSinh(struct attrbOccr AO)
  {return (!(AO.symPos)
           &&
           (gSymTable[prodBuffer2[AO.symPos].gSymTabInd].
               aaList[AO.attrbNum].InotS
           )
          );
  }



int areAllRHSoccrsSolvable(void)
  {int i,j,nA; 
   struct attrbOccr AO;
   
   ldgCDcount = 0; 
   for (i=1; i<prodBufferPointer2; i++) 
      {AO.symPos = i;
       nA = gSymTable[prodBuffer2[i].gSymTabInd].nAttrbs;   
       for (j=0;j<nA;j++) 
          {AO.attrbNum = j;
           if (!ldgClosureDisjoint(AO,isLHSinh,EE))
              return 0; 
          } /* for j */  
      } /* for i */ 
   return 1; 
  } 



int onRHS(struct attrbOccr AO)
  {return (AO.symPos > 0);
  }



int oneSynEELevel(void) 
  {struct attrbOccr AO;
   struct gSym dumSym;
   int j,nA;

   ldgCDcount = 0; 
   dumSym = gSymTable[prodBuffer2[0].gSymTabInd];   
   nA = dumSym.nAttrbs;   
   AO.symPos = 0; 
   for (j=0;j<nA;j++) 
       {AO.attrbNum = j; 
        if (!(dumSym.aaList[j].InotS))
           if (!ldgClosureDisjoint(AO,isLHSinh,EE)
               &&
               !ldgClosureDisjoint(AO,onRHS,EE)
              )
              return 0;
       } /* for j */  
   return 1;
  }



int allRHSoccrsSolvable;
int singleLevelSynEE;

void doPruneCondChecks(void)
  { 
   allRHSoccrsSolvable = areAllRHSoccrsSolvable();  
   singleLevelSynEE = oneSynEELevel(); 
  }



/* sticks node-generation code in yacc actions */
void printIntNodeGenInvocation(void) 
  {fprintf(yoxOutG,
           "\nyyyRSU(%d,%d,%d,%d);",
           prodNumSeq-1,prodBufferPointer2-1,
           gSymTable[prodBuffer2[0].gSymTabInd].nAttrbs,
           gSymTable[prodBuffer2[0].gSymTabInd].typeNum   
          );
   fprintf(yoxOutG,"\nyyyGenIntNode();");  
  } 



/* output the dependee-count initialization lists and indices into 
   the space of those lists
 */
void conveyLocalDepGraphInfo(void) 
  {int i,j,nA,nextAN; 
   long eeCountInitCurrentProd;  
   struct ldgAdjListHeads *pHead;  
   static int k1; 
   static int k2; 
   static long GSoccrSeq;   
   static long entOccrSeqHold;  
   eeCountInitCurrentProd = eeCountInitSeq; 
   indToEntsByProdNum[prodNumSeq-1] = GSoccrSeq; 
   pHead = ldgALhead; 
   for (i=0; i<prodBufferPointer2; i++) 
       {fprintf(temp5Out,"%ld,",attrbOccrSeq); k1++;  
        if (!(k1%12)) fprintf(temp5Out,"\n");   
        GSoccrSeq++;   
        nA = ldgALheadInd[i+1] - ldgALheadInd[i];   
        for (j=0;j<nA;j++) 
            {if (pHead->eeCount > RCmax) RCmax = pHead->eeCount; 
             if (pHead->eeCount != -1)  
                {if (i == 0) 
                    {if (permitCtimeCircTest) 
                        {fprintf(temp2Out,"yyyR,"); 
                         fprintf(temp2Out,"%u,%u, ",j,pHead->eeCount);
                         eeCountInitSeq++; 
                        } 
                    } 
                    else     
                    {fprintf(temp2Out,"%u,",i-1); 
                     fprintf(temp2Out,"%u,%u, ",j,pHead->eeCount);
                     eeCountInitSeq++; 
                    } 
                 if (!(eeCountInitSeq%6)) fprintf(temp2Out,"\n");  
                 pHead->eeList = 0; 
                }
             entOccrSeqHold = entOccrSeq; 
             fprintf(temp4Out,"%ld,",entOccrSeq*2); k2++;   
             if (!(k2%8)) fprintf(temp4Out,"\n");  
             attrbOccrSeq++; 
             if (pHead->eeCount != 0) 
                {nextAN = pHead->entList; 
                 while (nextAN != 0) 
                   {fprintf(temp3Out,"%u,%u,",ldgNodeStg[nextAN].node.symPos, 
                                              ldgNodeStg[nextAN].node.attrbNum 
                           );   
                    entOccrSeq++;  
                    if (!(entOccrSeq%8)) fprintf(temp3Out,"\n");  
                    nextAN = ldgNodeStg[nextAN].next; 
                   } 
                }  
             fprintf(temp6Out,"%ld,",entOccrSeq-entOccrSeqHold);  
             if (!(k2%8)) fprintf(temp6Out,"\n");  
             pHead->eeCount = -1;  
             pHead->entList = 0; 
             pHead++; 
            } /* for j */  
       } /* for i */ 
   fprintf(yoxOutG,"yyyAdjustINRC(yyyRCIL+%ld,yyyRCIL+%ld);",
           eeCountInitCurrentProd*3,
           eeCountInitSeq*3 
          );
   {int prune = 0; 
    if ((prodBufferPointer2 > 1) 
        &&
        allRHSoccrsSolvable  
        &&
        singleLevelSynEE
       )  
       {if (!suppressPruning) 
           {fprintf(yoxOutG,"yyyPrune();");  
            prune = 1; 
           } 
           else 
           {if (pruneRequested) 
               {fprintf(yoxOutG,"yyyPrune();");  
                prune = 1; 
               } 
               else 
               fprintf(yoxOutG,"/*yyyPrune();*/");  
           } 
       }
       else 
       {if (pruneRequested) 
           ywarning("warning: pruning requested but not permitted. ignored.\n");
       } 
    if (!prune && decorationRequested) 
       fprintf(yoxOutG,"yyyDecorate();");  
   }
  } 
 

/* writes the list of solved synthesized attributes of each token */
void installLeafRefCountInitList(void) 
  {struct gSym *pSym; 
   int i,leafRCIShold; 
   
   if (!permitCtimeCircTest) return;  
   leafRCIShold = leafRefCountInitSeq;  
   pSym = &(gSymTable[prodBuffer2[0].gSymTabInd]);  
   if (leafRefCountInitIndex[pSym->symNum] == -1)  
      {leafRefCountInitIndex[pSym->symNum] = leafRefCountInitSeq;    
       for (i=0; i<pSym->nAttrbs; i++)
           { 
            if (!pSym->aaList[i].InotS) 
               {
                fprintf(temp7Out,"%u,",i); 
                leafRefCountInitSeq++; 
                if (!(leafRefCountInitSeq%20)) fprintf(temp7Out,"\n");  
               }  
           }   
       leafRefCountListLength[pSym->symNum] = 
                   leafRefCountInitSeq - leafRCIShold;    
      }
   fprintf(Lfile[currentLfile].outPtr,
           "yyyGenLeaf(%d,%d,yyyLRCIL+%ld,yyyLRCIL+%ld);", 
           gSymTable[prodBuffer2[0].gSymTabInd].nAttrbs,   
           gSymTable[prodBuffer2[0].gSymTabInd].typeNum, 
           leafRefCountInitIndex[pSym->symNum], 
           leafRefCountInitIndex[pSym->symNum] +  
              leafRefCountListLength[pSym->symNum]
          );   
  } 


