/* $Id: outMgr.c 2625 2022-08-02 02:07:14Z teshields $ */ 

/* The information in this file is the intellectual property of 
 * Kurt M. Bischoff.  It may not be used, copied, or transferred 
 * without permission from the author. 
 *
 * Copyright (c) 1992, Kurt M. Bischoff.  All rights reserved.
 *
 * Beta Test Release--not for general distribution.
 */ 

#if defined (__cplusplus)
#include <cstdio>
#else
#include <stdio.h>
#endif

#include "glob.h" 
#include "opts.h" 
#include "scan.h" 
#include "YYLENG_INT.h" 

extern FILE *yoxOutG; 
extern FILE *yoxOutH; 
extern int *plineno; 

extern int yy1input(void); 
extern int yy2input(void); 
extern int yy3input(void); 
extern int yy4input(void); 
extern int whichScanner; 

extern struct gSym *gSymTable; 
extern char **structTDarr; 
extern int prodNumSeq; 

extern void debCom(void); 
extern void debString(void); 


extern void putCharInMacroDefBody(char); 

extern int line; 




#define LAbuffer1ArraySize 1024 
int LAbuffer1Pointer = 0;  
char LAbuffer1Array[LAbuffer1ArraySize]; 
FILE *LAbuffer1File; 
const char *LAbuffer1FileStr = "LAbuf1.temp"; 

#define LAbuffer2ArraySize 1024 
int LAbuffer2Pointer = 0;  
char LAbuffer2Array[LAbuffer2ArraySize]; 
FILE *LAbuffer2File; 
const char *LAbuffer2FileStr = "LAbuf2.temp"; 

long refSegBufferSize = 16384;   
long refSegBufferPointer;  
long refSegBufferPointerHold;  
char *refSegBuffer; 


#define S_DEBUG 0                   
#define DFAACCEPT 0 



int eatCom = 0; 
int eatStr = 0; 



char yyinput(void) 
  {switch(whichScanner) 
     {case 1: return yy1input(); 
      case 2: return yy2input(); 
      case 3: return yy3input(); 
      case 4: return yy4input(); 
     } 
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  } 



void putCharInLAbuffer(char c, int whichBuf)
  {switch(whichBuf)  
    {case LA1:   
     if (LAbuffer1Pointer < LAbuffer1ArraySize) 
        {LAbuffer1Array[LAbuffer1Pointer++] = c; 
         return; 
        } 
     if (LAbuffer1Pointer == LAbuffer1ArraySize) 
        LAbuffer1File = fopen(LAbuffer1FileStr,"w"); 
     putc(c,LAbuffer1File);  
     LAbuffer1Pointer++; 
     break;  
     case LA2: 
     if (LAbuffer2Pointer < LAbuffer2ArraySize) 
        {LAbuffer2Array[LAbuffer2Pointer++] = c; 
         return; 
        } 
     if (LAbuffer2Pointer == LAbuffer2ArraySize) 
        LAbuffer2File = fopen(LAbuffer2FileStr,"w"); 
     putc(c,LAbuffer2File);  
     LAbuffer2Pointer++; 
     break;  
    } 
  } 



void putCharInRefSegBuffer(char c)
  {if (refSegBufferPointer < refSegBufferSize) 
      refSegBuffer[refSegBufferPointer++] = c; 
      else 
      handleOverflow(RSBS); 
   if (c == '#') 
      {if (refSegBufferPointer < refSegBufferSize) 
          refSegBuffer[refSegBufferPointer++] = c; 
          else 
          handleOverflow(RSBS); 
      } 
   return;  
  } 



void putAttrbNumPairInRefSegBuffer(int symPos, int attrbNum)
  {if (refSegBufferPointer+16 <= refSegBufferSize) 
      {sprintf(refSegBuffer+refSegBufferPointer,"#%7d",symPos); 
       sprintf(refSegBuffer+(refSegBufferPointer += 8),",%7d",attrbNum); 
       refSegBufferPointer += 8; 
      } 
      else 
      handleOverflow(RSBS); 
  } 



void genImmCrefForAttrbRef(int symPos, int attrbNum, int GnotL, FILE *fp)
  {struct gSym *pSym; 
   if ((symPos == -1) || (attrbNum == -1)) return; 
   pSym = &(gSymTable[prodBuffer2[symPos].gSymTabInd]);  
   if (!symPos) 
      fprintf(fp,"(((%s)%s)->%s)",structTDarr[pSym->typeNum], 
                                   (GnotL?"yyySTsn":
                                          "(yyyTermBuffer.snBufPtr)"),
                                   pSym->attrbID[attrbNum] 
             ); 
      else 
      fprintf(fp,"(((%s)(((char *)((yyySTN->cL)[%u]))+yyyGNSz))->%s)", 
              structTDarr[pSym->typeNum], 
              symPos-1, 
              pSym->attrbID[attrbNum] 
             );  
  }


 

void genTravCrefForAttrbRef(int symPos, int attrbNum, FILE *outFile)
  {struct gSym *pSym; 
   if ((symPos == -1) || (attrbNum == -1)) return; 
   pSym = &(gSymTable[prodBuffer2[symPos].gSymTabInd]);  
   if (!symPos) 
      fprintf(outFile,"(((%s)(%s))->%s)",structTDarr[pSym->typeNum], 
                                  "((char *)yyyTSTn)+yyyGNSz",
                                  pSym->attrbID[attrbNum] 
             ); 
      else 
      fprintf(outFile,"(((%s)(((char *)((yyyTSTn->cL)[%u]))+yyyGNSz))->%s)", 
              structTDarr[pSym->typeNum], 
              symPos-1, 
              pSym->attrbID[attrbNum] 
             );  
  }



void genDecorCrefForAttrbRef(int symPos,
			     int attrbNum,
			     int targetSymPos,
			     int targetAttrbNum,
			     FILE *fp)
  {struct gSym *pSym; 
   if ((symPos == -1) || (attrbNum == -1)) return;  
   pSym = &(gSymTable[prodBuffer2[symPos].gSymTabInd]);  
   if ((symPos == targetSymPos) && (attrbNum == targetAttrbNum)) 
      {fprintf(fp,"(((%s)(((char *)yyyRSTopN)+yyyGNSz))->%s)",
                  structTDarr[pSym->typeNum],
                  pSym->attrbID[attrbNum] 
              ); 
       return; 
      } 
   if (!symPos) 
      fprintf(fp,"(((%s)(((char *)yyyRefN)+yyyGNSz))->%s)",
                 structTDarr[pSym->typeNum], 
                 pSym->attrbID[attrbNum] 
             ); 
      else  
      fprintf(fp,"(((%s)(((char *)((yyyRefN->cL)[%u]))+yyyGNSz))->%s)", 
              structTDarr[pSym->typeNum], 
              symPos-1, 
              pSym->attrbID[attrbNum] 
             );  
  }
 


void translateAndOutputRefSeg(enum instRefType refType,
			      long *pFirst,
			      long last,
                              int targetSymPos,
			      int targetAttrbNum,
			      FILE *fp 
                             ) 
  {char c; 
   int symPos,attrbNum;  
   int state = 1; 
   FILE *temp;

   if (neutralizeSemActions) 
      {temp = fp; 
       fp = tempTrashOut;
      }  
   while (*pFirst <= last) 
     {c = refSegBuffer[(*pFirst)++];  
      switch (state)   
        {case 1: switch(c) 
                   {case '#' : state = 2; break;   
                    default :  putc(c,fp); 
                   } 
                 break;  
         case 2: switch(c) 
                   {case '#' : putc('#',fp); state = 1; break;  
                    default  : sscanf(refSegBuffer+(*pFirst)-2,"#%7d",&symPos);
                               (*pFirst) += 6;  
                               sscanf(refSegBuffer+(*pFirst),",%7d",&attrbNum); 
                               (*pFirst) += 8; 
                               switch (refType)
                                 {case imm:
                                    genImmCrefForAttrbRef(symPos,attrbNum,1,fp);
                                    break;
                                  case decor:
                                    genDecorCrefForAttrbRef(symPos,
                                                            attrbNum, 
                                                            targetSymPos, 
                                                            targetAttrbNum, 
                                                            fp 
                                                           ); 
                                    break;
                                  case trav:
                                    genTravCrefForAttrbRef(symPos,
                                                           attrbNum,
                                                           fp
                                                          );
                                    break;
                                 }
                               state = 1;  
                   }
                 break; 
         default : ;   
        } 
     }  /* while */ 
   if (neutralizeSemActions) 
      {fp = temp;
      }  
  }  



void charOut(char c, int whereTo) 
  {switch (whereTo) 
     {case outG  : putc(c,yoxOutG); break; 
      case outL  : putc(c,Lfile[currentLfile].outPtr); break; 
      case outGandL  : putc(c,yoxOutH); break; 
      case outT1 : putc(c,temp1Out); break; 
      case rsB   : putCharInRefSegBuffer(c); break;  
      case outMB : putCharInMacroDefBody(c); break;  
      case outTrv: putc(c,tempTravOut); break; 
      case nowhere: break;
      default    : putCharInLAbuffer(c,whereTo); break;   
     }  
  } 



void stringOut(char *str, int whereTo)
  {while (*str != '\0') charOut(*str++,whereTo);
  }



void flushLAbuffer(int whichBuf, int whereTo) 
  {FILE *echoOut; 
   int i=0; 
   int count; 
   switch (whereTo) 
     {case outG: echoOut = yoxOutG; break; 
      case outL: echoOut = Lfile[currentLfile].outPtr; break; 
      case outT1: echoOut = temp1Out; break; 
     }  
   switch (whichBuf)  
     {case LA1: 
      count=(LAbuffer1ArraySize<LAbuffer1Pointer)?
            LAbuffer1ArraySize:LAbuffer1Pointer;
      while (i < count) charOut(LAbuffer1Array[i++],whereTo);    
      if (LAbuffer1Pointer > LAbuffer1ArraySize) 
         {fclose(LAbuffer1File); 
          LAbuffer1File = fopen(LAbuffer1FileStr,"r"); 
          do 
            {LAbuffer1Pointer -= LAbuffer1ArraySize;  
             count=(LAbuffer1ArraySize<LAbuffer1Pointer)?
                   LAbuffer1ArraySize:LAbuffer1Pointer;
             fread(LAbuffer1Array,sizeof(char),count,LAbuffer1File); 
             fwrite(LAbuffer1Array,sizeof(char),count,echoOut); 
            } 
          while (LAbuffer1Pointer > LAbuffer1ArraySize);   
         } 
      LAbuffer1Pointer = 0; 
      break; 
      case LA2: 
      count=(LAbuffer2ArraySize<LAbuffer2Pointer)?
             LAbuffer2ArraySize:LAbuffer2Pointer;
      while (i < count) charOut(LAbuffer2Array[i++],whereTo);    
      if (LAbuffer2Pointer > LAbuffer2ArraySize) 
         {fclose(LAbuffer2File); 
          LAbuffer2File = fopen(LAbuffer2FileStr,"r"); 
          do 
            {LAbuffer2Pointer -= LAbuffer2ArraySize;  
             count=(LAbuffer2ArraySize<LAbuffer2Pointer)?
                   LAbuffer2ArraySize:LAbuffer2Pointer;
             fread(LAbuffer2Array,sizeof(char),count,LAbuffer2File); 
             fwrite(LAbuffer2Array,sizeof(char),count,echoOut); 
            } 
          while (LAbuffer2Pointer > LAbuffer2ArraySize);   
         } 
      LAbuffer2Pointer = 0; 
      break;
     }  
  } 



void eatCommentAndEcho(int whereTo, int echoStart) 
  {char c; 
   int state = 1; 
   eatCom = 1;  
#if S_DEBUG 
   debCom(); 
#endif
   line = *plineno; 
   if (echoStart)
      {charOut('/',whereTo); 
       charOut('*',whereTo); 
      }
   while (state != DFAACCEPT)  
     {c = yyinput();  
      charOut(c,whereTo); 
      switch (state)   
        {case 1: switch(c) 
                   {case '*' : state = 2; break;  
                    default : ;
                   } 
                 break;  
         case 2: switch(c) 
                   {case '*' : break;  
                    case '/' : state = DFAACCEPT; 
                               continue; 
                    default  : state = 1;  
                   }
                 break; 
         default : ;   
        } 
     }  /* while */ 
   eatCom = 0;  
  } 



void eatStringAndEcho(int whereTo) 
  {char c; 
   int state = 1; 
   eatStr = 1;  
#if S_DEBUG 
   debString(); 
#endif
   line = *plineno; 
   charOut('\"',whereTo); 
   while (state != DFAACCEPT)  
     {c = yyinput();  
      charOut(c,whereTo); 
      switch (state)   
        {case 1: switch(c) 
                   {case '\\' : state = 2; break;  
                    case '\"' : state = DFAACCEPT; break;  
                    default : ;  
                   }
                 break;  
         case 2: state = 1; 
                 break; 
         default : ;   
        } 
     }  /* while */
   eatStr = 0;  
  } 





/* S_DEBUG is for debugging the scanner and/or the parser.  The person 
 * doing the debugging sets S_DEBUG to 1 and places comments in his test 
 * file at places at which he wants to know the start condition of the 
 * scanner.  The scanner will print a message to stderr giving the line 
 * number and the current start condition while reading that comment.  
 * Comments can appear about anywhere, so this can give a good picture of 
 * what is happening. 
*/  


/* The lookahead buffer is a fixed size main memory array LAbufferArray 
 * (that is supposed to be fast and supposed to almost always be big enough 
 * to hold the lookahead text) and file buffer LAbufferFile (that is probably  
 * slow, but of unlimited size).  putCharInLAbuffer(), 
 * copyLexemeToLAbuffer(), and flushLAbuffer() act on these data structures. 
 * Flushing the lookahead buffer when empty has no effect.  The main 
 * memory array part of the buffer is used as a buffer for flushing 
 * the file part of the buffer. 
*/ 

