/* To demonstrate the use of an attribute grammar for the printing of parse
 *  trees of a simple expression grammar.
 */

%{
#if defined (__cplusplus)
#include <cstring>
#include <cstdio>
#else
#include <string.h>
#include <stdio.h>
#endif

#if YYBTYACC && !YYBYACC
/* the `btyacc' 3.0 skeleton omits the following */
#include "parser.h"
#endif

char *temp,*image;
long i,iWidth,sHeight,size;
int quiet = 0;

#define max(i1,i2) (((i1) >= (i2)) ? (i1) : (i2))

extern int yylex(void);

#if YYBTYACC && !YYBYACC
void yyerror(const char *, ...);
#else
void yyerror(const char *);
#endif
%}

%token CONST IDENT

%start s

@attributes {short height,width,hanger,line,column; @void print;/*synthesized*/
            }  s

@attributes {short height,width,hanger; @void print; /* synthesized */
             @void prntSeq;  short line,column;      /* inherited */
	    } e t f

%%
s	: 	e {} 	
	 @{ @i @s.line@ = 0;
	    @i @s.column@ = 0;
	    @i sHeight = @s.height@ = @e.height@+2;
	    @i iWidth = (1 + (@s.width@ = @e.width@));
	    @i @s.hanger@ = @e.hanger@;
	    @e s.print: s.height s.width s.hanger
		   ; {
 	      size = iWidth*@s.height@;
	      if ((image = (char *)malloc(size))
		  ==
		  ((char *)NULL)
		 )
		{fprintf(stderr,"error in mallocing image storage\n");
	 	 exit(-1);
		}
	      for (i=0;i<size;i++)
                {image[i] = ' '; }
	      for (i=1; i <= @s.height@; i++)
                image[(i*iWidth)-1] = '\n';
	      image[@s.hanger@] = 's';
	      image[iWidth+@s.hanger@] = '|';
		       }
	    @i @e.line@ = @s.line@ + 2;
	    @i @e.column@ = @s.column@;
	    @e e.prntSeq: s.print; {}
	@}
	;

e	:	e 	'+'	t{} 	
	 @{ @i @e.1.line@ = @e.0.line@ + 2;
            @i @e.1.column@ = @e.0.column@;
	    @i @t.line@ = @e.0.line@ + 2;
            @i @t.column@ = @e.1.column@ + @e.1.width@ + 3;
	    @i @e.0.height@ = max(@e.1.height@,@t.height@) + 2;
	    @i @e.0.width@ = @e.1.width@ + 3 + @t.width@;
            @i @e.0.hanger@ = (@e.1.hanger@ + @e.1.width@ +3+ @t.hanger@) /2;
	    @e e.0.print: e.0.line e.0.column e.0.hanger
		       e.1.hanger e.1.width
		       t.hanger
		       e.0.prntSeq
		     ;                    {
 	      temp = image+(iWidth*@e.0.line@)+@e.0.column@;
 	      *(temp + @e.0.hanger@) = 'e';
 	      temp += iWidth;
	      *(temp + (@e.0.hanger@+@e.1.hanger@)/2) = '/';
	      i=(@e.0.hanger@+@e.1.width@+1)/2;
 	      *(temp + @e.0.hanger@) = '|';
	      *(temp+(@e.0.hanger@+@e.1.width@+3+@t.hanger@)/2) = '\\';            					
 	      *(temp + iWidth + @e.0.hanger@) = '+';
		     }
	   @e e.1.prntSeq t.prntSeq: e.0.print; {}
        @}

	|	t  {}
	 @{ @i @e.height@ = @t.height@ + 2;
	    @i @e.width@ = @t.width@;
	    @i @e.hanger@ = @t.hanger@;
	    @i @t.line@ = @e.line@ + 2;
	    @i @t.column@ = @e.column@;
	    @e e.print: e.line e.column e.hanger
		     e.prntSeq
		   ;        {
	      *(image+(iWidth*@e.line@)+@e.column@+@e.hanger@) = 'e';
	      *(image+(iWidth*(@e.line@+1))+@e.column@+@e.hanger@) = '|';
			 }
	    @e t.prntSeq: e.print; {}
        @}
	;

t	:	t	'*'	f{} 	
	 @{ @i @t.1.line@ = @t.0.line@ + 2;
            @i @t.1.column@ = @t.0.column@;
	    @i @f.line@ = @t.0.line@ + 2;
            @i @f.column@ = @t.1.column@ + @t.1.width@ + 3;
	    @i @t.0.height@ = max(@t.1.height@,@f.height@) + 2;
	    @i @t.0.width@ = @t.1.width@ + 3 + @f.width@;
            @i @t.0.hanger@ = (@t.1.hanger@ + @t.1.width@ +3+ @f.hanger@) /2;
	    @e t.0.print: t.0.line t.0.column t.0.hanger
		       t.1.hanger t.1.width
		       f.hanger
		       t.0.prntSeq
		     ;                      {
      	      temp = image+(iWidth*@t.0.line@)+@t.0.column@;
	      *(temp+@t.0.hanger@) = 't';
	      temp += iWidth;
	      *(temp+(@t.0.hanger@+@t.1.hanger@)/2) = '/';  						
	      *(temp+@t.0.hanger@) = '|';
              i=(@t.0.hanger@+@t.1.width@+1)/2;
	      *(temp+(@t.0.hanger@+@t.1.width@+3+@f.hanger@)/2) = '\\';
	      *(temp+iWidth+@t.0.hanger@) = '*';
			 }
	    @e f.prntSeq t.1.prntSeq: t.0.print; {}
        @}

	| 	f	{} 	
	 @{ @i @t.height@ = @f.height@ + 2;
	    @i @t.width@ = @f.width@;
	    @i @t.hanger@ = @f.hanger@;
	    @i @f.line@ = @t.line@ + 2;
	    @i @f.column@ = @t.column@;
	    @e t.print: t.line t.column t.hanger
		     t.prntSeq
		   ;   {
	      temp = image+(iWidth*@t.line@)+@t.column@+@t.hanger@;
	      *temp = 't';
	      *(temp+iWidth) = '|';
			  }
	    @e f.prntSeq: t.print; {}
         @}
	;

f	:	'('	e	')' 	{}
         @{ @i @f.height@ = @e.height@ + 2;
	    @i @f.width@ = 1 + @e.width@ + 1;
	    @i @f.hanger@ = @e.hanger@ + 1;
	    @i @e.line@ = @f.line@ + 2;
            @i @e.column@ = @f.column@ + 1;
            @e f.print: f.line f.column f.hanger f.width
		     e.line e.column e.hanger e.width
                     f.prntSeq
		   ;   {
              temp = image+(iWidth*@f.line@)+@f.column@+@f.hanger@;
              *temp = 'f';
              temp += iWidth;
              *temp = '|';
              *(temp-2) = '/';
              *(temp+2) = '\\';
              temp += iWidth;
              *(temp-3) = '(';
              *(temp+3) = ')';
			   }
	    @e e.prntSeq: f.print; {}
	 @}

	|	IDENT 	{} 	
	 @{ @i @f.height@ = 3;
	    @i @f.width@ = 7;
	    @i @f.hanger@ = 3;
	    @e f.print : f.line f.column f.hanger
		      f.prntSeq
		    ;  {
	      temp = image+(iWidth*@f.line@)+@f.column@+@f.hanger@;
              *temp = 'f';
              *(temp+iWidth) = '|';
              strncpy(temp+(2*iWidth)-2,"IDENT",5);
						  }
	 @}

	|	CONST  	{}
	 @{ @i @f.height@ = 3;
	    @i @f.width@ = 7;
	    @i @f.hanger@ = 3;
	    @e f.print : f.line f.column f.hanger
		      f.prntSeq
		    ;  {
        /*      fprintf(stderr,"prod 8 print\n"); */
	      temp = image+(iWidth*@f.line@)+@f.column@+@f.hanger@;
              *temp = 'f';
              *(temp+iWidth) = '|';
              strncpy(temp+(2*iWidth)-2,"CONST",5);
			 }
	 @}
	;
%%

#if YYBTYACC && !YYBYACC
void yyerror(const char *s, ...)
{extern int yylineno;
 fprintf(stderr,"Near line %d: %s\n",yylineno,s);
 yyyabort();
}
#else
void yyerror(const char *s)
{fprintf(stderr,"%s\n",s);
 yyyabort();
}
#endif

int yyparse(void);

int main(int argc, char *argv[])
  {
   if ((argc > 1) && (argv[1][1] == 'q')) quiet = 1;
   yyparse();
{int j;
 if (!quiet)
    {for (i=0;i<sHeight;i++)
         {for (j=0;j<iWidth;j++) fputc(image[i*iWidth + j],stderr);
          fputc('\n',stderr);
         }
    }
}
  }
